#ifndef FORMULA_H_
#define FORMULA_H_

#include <iostream>
#include <vector>
#include <math.h>

using namespace std;

enum Op { CONSTANT, PARAMETER,
	PLUS, LOGPLUS, MINUS, TIMES, LOGTIMES, DIVIDE, LOGDIVIDE, NEG, EXP, LOG,
	POWER, LOGPOWER, SQRT, TANH, RELU, TRUNC_RELU, LEAKY_RELU, LOGISTIC, GD, ERF, MAX,
	SAMPLE_MULTINOMIAL, PROBABILISTIC_AVERAGING };

class Model;
typedef unsigned int PARAM_ID;

/**
 * Class to represent a formula node in a formula DAG, which represents a function.
 * Leaf nodes in the DAG are either PARAMETER nodes which correspond to model
 * parameters or CONSTANT nodes which correspond to constants, and non-leaf nodes are
 * operations, such as TIMES, PLUS, DIVIDE, etc. Every node of the formula DAG is
 * an instance of a Formula object. To implement your own operator, add it to the
 * enum above and implement a case for it in evaluate() and getValueToBackprop().
 * Some operators can be made more efficient by using additional state; to implement
 * MAX, for example, we added a variable called m_maxChildIndex to store the index of
 * the child node with maximum value. This variable can be used to store arbitrary
 * integer state for your operator, or you can create new variables here and reference
 * them in evaluate() and getValueToBackprop(). You can also extend the Formula class
 * for operators that require a lot of additional state.
 *
 * Kevin Gimpel
 * 2008-2014
 */
class Formula {
	public:
		// Constructors:
		Formula();
		Formula(const Formula& f);
		Formula(Op o);
		Formula(double v);
		Formula(PARAM_ID index);
		~Formula();
		Formula& operator=(const Formula& f);

		// Reset functions:
		void reset(Op o, Model* model, int numChildren = 0);
		void reset(double v, Model* model);
		void reset(PARAM_ID index, Model* model);

		void setOwner(Model* m);

		// Returns parameter index of this node; only meaningful if it's a PARAMETER node.
		int getIndex() const;
		// Returns current gradient value at this node.
		double getGradient() const;
		// Returns the number of parents of this node.
		int getNumParents() const;
		// Returns the number of children of this node.
		int getNumChildren() const;

		// Adds the given Formula node as a child of this Formula node.
		void addChild(Formula* f);
		// Calls reserve(numChildren) on the children vector of this Formula node.
		void reserveChildren(int numChildren);

		// Resets flags indicating that function and gradient values have been
		// computed. Recurs on children if argument is true. This should be called when parameter values change.
		void changedParamValues(bool recurOnChildren);

		// Computes the value of the Formula with given values for the parameters,
		// returning the computed value.
		double evaluate(int syncValue = 1);

		// Computes the derivative of the root of the FDAG with respect to this Formula node.
		// Must be called on a PARAMETER node.
		// Returns computed derivative of the root of the FDAG with respect to this PARAMETER node.
		// Adds the return value to the derivative stored for the parameter in the Model pointed to by m_owner.
		double computeDerivative(int syncValue = 1);

		// Pulls gradient from parent nodes, then computes gradient at current node,
		// and returns gradient update for given index in the child node list.
		double pull(int childIndex, int syncValue = 1);

		// Returns value to be back-propagated to the given child node.
		double getValueToBackprop(int childIndex, int syncValue = 1);

		// Sets m_value to given newValue. Useful for reusing constant Formula nodes.
		void setValue(double newValue);

		// Recursively prints FDAG starting at this Formula node.
		void printFormula(int numtabs = 0) const;

		string toString() const;

		// Returns string representation of operator, used only for printing FDAGs.
		static string getStringForOp(Op op) {
			switch (op) {
			case LOGTIMES:
				return "log*";
			case TIMES:
				return "*";
			case LOGPLUS:
				return "log+";
			case PLUS:
				return "+";
			case MINUS:
				return "-";
			case LOGDIVIDE:
				return "log/";
			case DIVIDE:
				return "/";
			case NEG:
				return "-";
			case EXP:
				return "e^";
			case LOG:
				return "log";
			case PARAMETER:
				return "param";
			case CONSTANT:
				return "const";
			case POWER:
				return "^";
			case LOGPOWER:
				return "log^";
			case SQRT:
				return "sqrt";
			case TANH:
				return "tanh";
			case RELU:
				return "relu";
			case TRUNC_RELU:
				return "trunc_relu";
			case LEAKY_RELU:
				return "leaky_relu";
			case LOGISTIC:
				return "logistic";
			case GD:
				return "gd";
			case ERF:
				return "erf";
			case MAX:
				return "max";
			case SAMPLE_MULTINOMIAL:
				return "sample_multinomial";
			case PROBABILISTIC_AVERAGING:
				return "probabilistic_averaging";
			}
			return "NULL";
		}

	protected:
		// Internal initialization and reset functions:
		void initCommon();
		void resetCommon(Model* model, int numChildren = 0);

		// protected member variables

		// The operation to be performed by this formula node.
		enum Op m_operation;
		// The parents of this formula in the FDAG.
		// The first entry in the pair is the pointer to the parent node and the
		// second entry is the index of the current Formula in the child list in the parent.
		vector<pair<Formula*, int> > m_parents;
		// The children of this formula in the FDAG.
		vector<Formula*> m_children;
		// The owner of this formula node.
		Model* m_owner;
		// The computed value (u), if m_valueComputed == true.
		double m_value;
		// Indicates whether the value of this formula has been computed.
		//bool m_valueComputed;
		int m_valueComputedSyncValue;
		// The accumulated gradient stored at this formula node.
		double m_gradient;
		// Indicates whether the gradient of this formula has been computed.
		//bool m_gradientComputed;
		int m_gradientComputedSyncValue;
		// Index of the parameter that this node represents, if it's a PARAMETER node.
		// Changed this from PARAM_ID to int because we want to be able to set it to -1 to indicate
		// that it's unset, and also we want to use this variable for other auxiliary data needed by
		// Formula nodes, like MAX.
		int m_index;
		// We used to separately hold the index of the child argument that holds the maximum value
		// for MAX nodes in the variable below (m_maxChildIndex).
		// Now, to save memory, we use m_index for this, since MAX nodes don't otherwise use m_index.
		//int m_maxChildIndex;
};

#endif /*FORMULA_H_*/
