#include "model.h"
#include "formula.h"

Model::Model() {
	initCommon(10000, 10000, 10000, 10000);
}

Model::Model(int formulaCapacity) {
	initCommon(formulaCapacity, formulaCapacity, formulaCapacity, formulaCapacity);
}

Model::Model(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity) {
	initCommon(unaryFormulaCapacity, binaryFormulaCapacity, fertileFormulaCapacity, barrenFormulaCapacity);
}

void Model::initCommon(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity) {
	m_numParams = 0;
	m_unaryFormulaPool.initialize(unaryFormulaCapacity);//2000000);//formulaCapacity);
	m_binaryFormulaPool.initialize(binaryFormulaCapacity);//2000000);//formulaCapacity);
	m_fertileFormulaPool.initialize(fertileFormulaCapacity);//10000);//formulaCapacity);
	m_barrenFormulaPool.initialize(barrenFormulaCapacity);//60000);//formulaCapacity);
}

Model::~Model() {}

PARAM_ID Model::initializeNewParameter(double value, bool useCache) {
	PARAM_ID ret = m_V.size();
	setValue(ret, value);
	setGradient(ret, 0.0);
	if (useCache) m_parameterFormulas.push_back(NULL);
	return ret;
}

void Model::initializeParameterMatrix(vector<vector<PARAM_ID> >& m, unsigned int numRows, unsigned int numCols, double scalingFactor) {
	m.clear();
	m.reserve(numRows);
	for (unsigned int i = 0; i < numRows; ++i) {
		vector<PARAM_ID> rowi;
		rowi.reserve(numCols);
		for (unsigned int j = 0; j < numCols; ++j) {
			// model.initializeNewParameter() returns the PARAM_ID (parameter index) of the new parameter.
			rowi.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
		}
		m.push_back(rowi);
	}
}

void Model::initializeParameterMatrixGivenRowLengths(vector<vector<PARAM_ID> >& m, vector<int>& numRows, double scalingFactor) {
	m.clear();
	m.reserve(numRows.size());
	for (unsigned int i = 0; i < numRows.size(); ++i) {
		vector<PARAM_ID> rowi;
		rowi.reserve(numRows[i]);
		for (unsigned int j = 0; j < numRows[i]; ++j) {
			// model.initializeNewParameter() returns the PARAM_ID (parameter index) of the new parameter.
			rowi.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
		}
		m.push_back(rowi);
	}
}

void Model::initializeParameterMatrices(vector<vector<vector<PARAM_ID> > >& m, const vector<int>& numRows, const vector<int>& numCols, double scalingFactor) {
	m.clear();
	m.reserve(numRows.size());
	for (unsigned int i = 0; i < numRows.size(); ++i) {
		// Initialize a parameter matrix of dimension numRows[i] x numCols[i].
		vector<vector<PARAM_ID> > currentM;
		initializeParameterMatrix(currentM, numRows[i], numCols[i], scalingFactor);
		m.push_back(currentM);
	}
}

void Model::initializeParameterVector(vector<PARAM_ID>& v, unsigned int numEntries, double scalingFactor) {
	v.clear();
	v.reserve(numEntries);
	for (unsigned int i = 0; i < numEntries; ++i) {
		v.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
	}
}


// Sets parameter value corresponding to given index to given value.
void Model::setValue(PARAM_ID index, double newValue) {
	if (index >= m_V.size()) {
		// Add zeroes to the parameter vector until we reach the desired index.
		while (index > m_V.size()) {
			m_V.push_back(0.0);
		}
		// Insert the specified parameter value for the desired index.
		m_V.push_back(newValue);
		m_numParams = m_V.size();
	} else {
		m_V[index] = newValue;
	}
}

// Gets gradient value corresponding to given index.
double Model::getGradient(PARAM_ID index) { return m_G.at(index); }

// Sets gradient value corresponding to given index to given value.
void Model::setGradient(PARAM_ID index, double newGradient) {
	if (index >= m_G.size()) {
		// Add zeroes to the gradient vector until we reach the desired index.
		while (index > m_G.size()) {
			m_G.push_back(0.0);
		}
		// Insert the specified gradient value for the desired index.
		m_G.push_back(newGradient);
	} else {
		m_G[index] = newGradient;
	}
}

void Model::addGradient(PARAM_ID index, double gradientIncrement) {
	if (index >= m_G.size()) {
		// Add zeroes to the gradient vector until we reach the desired index.
		while (index > m_G.size()) {
			m_G.push_back(0.0);
		}
		// Insert the specified gradient value for the desired index.
		m_G.push_back(gradientIncrement);
	} else {
		m_G[index] += gradientIncrement;
	}
}

int Model::getNumParams() {
	return m_numParams;
}

Formula* Model::getFormulaObject(Op o, int numChildren) {
	//Formula* f = m_formulaPool.getFormula();
	if (numChildren == 1) {
		Formula* f = m_unaryFormulaPool.getFormula();
		f->reset(o, this);
		return f;
	} else if (numChildren == 2) {
		Formula* f = m_binaryFormulaPool.getFormula();
		f->reset(o, this);
		return f;
	} else {
		Formula* f = m_fertileFormulaPool.getFormula();
		f->reset(o, this, numChildren);
		return f;
	}
}
Formula* Model::getConstantFormulaObject(double v) {
	//Formula* f = m_formulaPool.getFormula();
	Formula* f = m_barrenFormulaPool.getFormula();
	f->reset(v, this);
	return f;
}
Formula* Model::getCachedParameterFormula(PARAM_ID id) {
	if (id >= m_parameterFormulas.size()) {
		cerr << "Warning: you called getCachedParameterFormula(" << id << ") but the cache was not properly initialized or reset." << endl;
		resetCachedParameterFormulas(m_parameterFormulas);
	}
	Formula* f = m_parameterFormulas.at(id);
	if (f == NULL) {
		f = getParameterFormulaObject(id);
		m_parameterFormulas[id] = f;
	}
	return f;
}
// The following function will use the provided cache rather than the internal cache.
Formula* Model::getCachedParameterFormula(PARAM_ID id, vector<Formula*>& parameterFormulas) {
	if (id >= parameterFormulas.size()) {
		cerr << "Warning: you called getCachedParameterFormula(" << id << ") but the cache was not properly initialized or reset." << endl;
		resetCachedParameterFormulas(parameterFormulas);
	}
	Formula* f = parameterFormulas.at(id);
	if (f == NULL) {
		f = getParameterFormulaObject(id);
		parameterFormulas[id] = f;
	}
	return f;
}
Formula* Model::getParameterFormulaObject(PARAM_ID id) {
	//Formula* f = m_formulaPool.getFormula();
	Formula* f = m_barrenFormulaPool.getFormula();
	f->reset(id, this);
	return f;
}

void Model::resetFormulaObjects(bool resetParameterFormulaCache) {
	//m_formulaPool.freeAllFormulas();
	m_unaryFormulaPool.freeAllFormulas();
	m_binaryFormulaPool.freeAllFormulas();
	m_fertileFormulaPool.freeAllFormulas();
	m_barrenFormulaPool.freeAllFormulas();
	if (resetParameterFormulaCache) resetCachedParameterFormulas(m_parameterFormulas);
}

void Model::resetCachedParameterFormulas(vector<Formula*>& parameterFormulas) {
	if (parameterFormulas.size() == m_V.size()) {
		for (unsigned int i = 0; i < m_V.size(); ++i) {
			parameterFormulas[i] = NULL;
		}
	} else {
		parameterFormulas.clear();
		parameterFormulas.reserve(m_V.size());
		for (unsigned int i = 0; i < m_V.size(); ++i) {
			parameterFormulas.push_back(NULL);
		}
	}
}
void Model::resetFormulaComputedFlags() {
	//m_formulaPool.resetAllFormulaComputedFlags();
	m_unaryFormulaPool.resetAllFormulaComputedFlags();
	m_binaryFormulaPool.resetAllFormulaComputedFlags();
	m_fertileFormulaPool.resetAllFormulaComputedFlags();
	m_barrenFormulaPool.resetAllFormulaComputedFlags();
}

void Model::getNumberOfFormulasUsedInPool(int& unary, int& binary, int& fertile, int& barren) const {
	//return m_formulaPool.getNumberOfFormulasUsed();
	unary = m_unaryFormulaPool.getNumberOfFormulasUsed();
	binary = m_binaryFormulaPool.getNumberOfFormulasUsed();
	fertile = m_fertileFormulaPool.getNumberOfFormulasUsed();
	barren = m_barrenFormulaPool.getNumberOfFormulasUsed();
}

void Model::resetGradientVector() {
	for (unsigned int i = 0; i < m_G.size(); ++i) {
		m_G[i] = 0.0;
	}
}

double Model::computeParameterVectorNorm() { return computeVectorNorm(m_V); }

double Model::computeGradientVectorNorm() { return computeVectorNorm(m_G); }
// Multiplies each entry in the gradient vector by the given constant c.
void Model::scaleGradientVectorByConstant(double c) {
	for (unsigned int i = 0; i < m_G.size(); ++i) {
		m_G[i] *= c;
	}
}

double Model::evalAndComputeDerivatives(Formula* f, int exampleNumber, bool doFiniteDifferenceChecking, bool alwaysPrint) {
	//f->printFormula();
	double result = f->evaluate();
	//f->printFormula();

	for (unsigned int i = 0; i < m_parameterFormulas.size(); ++i) {
		Formula* currParamFormula = m_parameterFormulas[i];
		if (currParamFormula != NULL) {
			double deriv = currParamFormula->computeDerivative();
			if (deriv != deriv) {
				;//cerr << "Error: nan derivative computed for parameter " << currParamFormula->getIndex() << " on example " << exampleNumber << endl;
				//f->printFormula();
			}
			if (doFiniteDifferenceChecking) finiteDifferenceCheck(f, i, result, deriv, 0.000001, exampleNumber, alwaysPrint);
		}
	}
	//cerr << endl << "Printing formula DAG again after computing gradients:" << endl << endl;
	//f->printFormula();
	return result;
}

double Model::evalAndComputeDerivatives(Formula* f, vector<Formula*>& parameterFormulas, int syncValue, int exampleNumber, bool doFiniteDifferenceChecking, bool alwaysPrint) {
	//f->printFormula();
	double result = f->evaluate(syncValue);
	//f->printFormula();

	for (unsigned int i = 0; i < parameterFormulas.size(); ++i) {
		Formula* currParamFormula = parameterFormulas[i];
		if (currParamFormula != NULL) {
			double deriv = currParamFormula->computeDerivative(syncValue);
			if (deriv != deriv) {
				;//cerr << "Error: nan derivative computed for parameter " << currParamFormula->getIndex() << " on example " << exampleNumber << endl;
				//f->printFormula();
			}
			if (doFiniteDifferenceChecking) finiteDifferenceCheck(f, i, result, deriv, 0.000001, exampleNumber, alwaysPrint);
		}
	}
	//cerr << endl << "Printing formula DAG again after computing gradients:" << endl << endl;
	//f->printFormula();
	return result;
}

void Model::finiteDifferenceCheck(Formula* f, PARAM_ID paramInd, double result,
		double deriv, double delta, int exampleNumber, bool alwaysPrint) {
	if (delta < 0.0000000001) {
		cerr << "Warning: delta passed to finiteDifferenceCheck was zero or too small. Using 0.000001." << endl;
		delta = 0.000001;
	}
	bool recurOnChildrenTrue = true;
	// Estimate partial derivatives numerically using a simple finite difference method.
	// Save the original parameter value.
	double originalValue = getValue(paramInd);
	// Decrease the parameter value by delta.
	setValue(paramInd, originalValue - delta);
	// Reset the FDAG since we changed a parameter.
	f->changedParamValues(recurOnChildrenTrue);
	// Compute the function.
	double newResult = f->evaluate();
	// Increase the parameter value by delta.
	setValue(paramInd, originalValue + delta);
	// Reset the FDAG since we changed a parameter.
	f->changedParamValues(recurOnChildrenTrue);
	// Compute the function again.
	double newResult2 = f->evaluate();
	// Estimate the derivative as the difference of the two function values divided by 2*delta.
	double numericalDeriv = ((newResult2 - newResult)/(2*delta));
	//cerr << "Numer. Derivative for " << paramInd // << " (" << currParamFormula->getName() << ")"
	//	 << " = \t" << numericalDeriv << "\tdelta = " << delta << endl;
	double absDifference = fabs(deriv - numericalDeriv);
	if (absDifference > delta) cerr << "Error ";
	if (absDifference > delta || alwaysPrint) {
		cerr << "(example " << exampleNumber << ", param " << paramInd << "): for delta " << delta
			 << ", abs(deriv.(" << deriv << ") - numerical deriv.(" << numericalDeriv << ")) = " << absDifference << endl;
	}
	// Set the parameter value to its original value.
	setValue(paramInd, originalValue);
	// Reset the formula tree and evaluate the function again.
	f->changedParamValues(recurOnChildrenTrue);
	f->evaluate();
}

void Model::doSGDUpdate(double stepsize) {
	for (unsigned int i = 0; i < m_V.size(); ++i) {
		double oldValue = getValue(i);
		double update = -1.0 * getGradient(i) * stepsize;
		setValue(i, oldValue + update);
	}
}

double Model::doSGDMomentumUpdate(vector<double>& previousUpdates, double stepsize, double momentumCoefficient) {
	double gradSquaredNorm = 0.0f;
	for (unsigned int i = 0; i < m_V.size(); ++i) {
		double oldValue = getValue(i);
		double grad = getGradient(i);
		double gradSquared = grad * grad;
		if (gradSquared > 0.0) {
			gradSquaredNorm += gradSquared;

			double currentUpdate = -1.0 * grad * stepsize;
			double update = (previousUpdates[i] * momentumCoefficient) + currentUpdate;
			previousUpdates[i] = update;
			setValue(i, oldValue + update);
		} else {
			previousUpdates[i] = (previousUpdates[i] * momentumCoefficient);
		}
	}
	return gradSquaredNorm;
}

double Model::doAdaGradUpdate(vector<double>& sumOfSquaredGradients, double stepsize) {
	double gradSquaredNorm = 0.0f;
	int numZeroGrad = 0;
	for (unsigned int i = 0; i < m_V.size(); ++i) {
		double grad = getGradient(i);
		if (grad != grad) {
			;//cerr << "Error: in doAdaGradUpdate(), nan grad for parameter " << i << endl;
		} else {
			double gradSquared = grad * grad;
			if (gradSquared > 0.0) {
				gradSquaredNorm += gradSquared;
				sumOfSquaredGradients[i] += gradSquared;
				double adaGradStepsize = stepsize / (sqrt(sumOfSquaredGradients[i]));
				double oldValue = getValue(i);
				double update = -1.0 * grad * adaGradStepsize;
				double newValue = oldValue + update;
				if (newValue != newValue) {
					cerr << "Error: in doAdaGradUpdate(), nan value for parameter " << i << ". oldValue = " << oldValue << ", grad = " << grad << ", stepsize = " << adaGradStepsize << endl;
				} else if (newValue == neg_inf) {
					cerr << "Warning: in doAdaGradUpdate(), -inf for parameter " << i << ". oldValue = " << oldValue << ", grad = " << grad << ", stepsize = " << adaGradStepsize << endl;
				} else if (newValue == pos_inf) {
					cerr << "Warning: in doAdaGradUpdate(), inf for parameter " << i << ". oldValue = " << oldValue << ", grad = " << grad << ", stepsize = " << adaGradStepsize << endl;
				}
				setValue(i, newValue);
			} else {
				++numZeroGrad;
			}
		}
	}
	return gradSquaredNorm;
}

// Performs AdaDelta update on parameters using current gradient vector given sum of squared
// gradients, sum of squared updates, and the two hyperparameters epsilon and rho.
double Model::doAdaDeltaUpdate(vector<double>& sumOfSquaredGradients, vector<double>& sumOfSquaredUpdates, double epsilon, double rho) {
	double gradSquaredNorm = 0.0f;
	int numZeroGrad = 0;
	for (unsigned int i = 0; i < m_V.size(); ++i) {
		double grad = getGradient(i);
		if (grad != grad) {
			;//cerr << "Error: in doAdaDeltaUpdate(), nan grad for parameter " << i << endl;
		} else {
			double gradSquared = grad * grad;
			if (gradSquared > 0.0) {
				gradSquaredNorm += gradSquared;

				//double previousSquaredGradient = sumOfSquaredGradients[i];
				// Update the sum of squared gradients for parameter i.
				sumOfSquaredGradients[i] = rho * sumOfSquaredGradients[i] + (1.0 - rho) * gradSquared;

				double adaDeltaStepsize = (sqrt(sumOfSquaredUpdates[i] + epsilon)) / (sqrt(sumOfSquaredGradients[i] + epsilon));
				double update = -1.0 * grad * adaDeltaStepsize;

				double oldValue = getValue(i);
				double newValue = oldValue + update;
				if (newValue != newValue) {
					cerr << "Error: in doAdaDeltaUpdate(), nan value for parameter " << i << ". oldValue = " << oldValue << ", grad = " << grad << ", stepsize = " << adaDeltaStepsize << endl;
				} else if (newValue == neg_inf) {
					cerr << "Warning: in doAdaDeltaUpdate(), -inf for parameter " << i << ". oldValue = " << oldValue << ", grad = " << grad << ", stepsize = " << adaDeltaStepsize << endl;
				} else if (newValue == pos_inf) {
					cerr << "Warning: in doAdaDeltaUpdate(), inf for parameter " << i << ". oldValue = " << oldValue << ", grad = " << grad << ", stepsize = " << adaDeltaStepsize << endl;
				}
				setValue(i, newValue);

				// Update the sum of squared updates for parameter i.
				sumOfSquaredUpdates[i] = rho * sumOfSquaredUpdates[i] + (1.0 - rho) * (update * update);
			} else {
				++numZeroGrad;
			}
		}
	}
	return gradSquaredNorm;
}
