#include "classifier_utils.h"
#include <iostream>
#include <cstdlib>	// for atoi()
#include <cmath>	// for abs()
#include <sys/time.h> // for gettimeofday()
#include "../fdag/formula.h"
#include "../fdag/model.h"

SequenceDatum::~SequenceDatum() {}

WordSequenceDatum::~WordSequenceDatum() {}

string WordSequenceDatum::toString() const {
	string s;
	for (unsigned int i = 0; i < m_words.size(); ++i) {
		if (i > 0) s += " ";
		s += m_words[i];
	}
	return s;
}

string WordSequenceDatum::toStringInsertStringAfterWordIndex(const string& toInsert, int pos) const {
	string s;
	for (unsigned int i = 0; i < m_words.size(); ++i) {
		if (i > 0) s += " ";
		s += m_words[i];
		if (i == pos) s += (" " + toInsert);
	}
	return s;
}

string WordSequenceDatum::getWordsInRange(int startIndex, int endIndex) const {
	string ret = "";
	int endIndexToUse = endIndex;
	if (endIndexToUse >= m_words.size()) endIndexToUse = ((int)(m_words.size())) - 1;
	for (int i = startIndex; i <= endIndexToUse; ++i) {
		if (i > startIndex) ret += " ";
		ret += m_words.at(i);
	}
	return ret;
}

string WordSequenceDatum::getWordsInRangePadded(int startIndex, int endIndex) const {
	string ret = "";
	for (int i = startIndex; i <= endIndex; ++i) {
		if (i > startIndex) ret += " ";
		if (i < m_words.size()) {
			ret += m_words[i];
		} else {
			ret += "<EOS>";
		}
	}
	return ret;
}

void WordSequenceDatum::readInWords(const string& s, const map<string, vector<double> >& wordVectors, map<string, int>* unkCounts) {
	m_words.clear();
	tokenize(s, ' ', m_words);
	m_sequenceLength = (int)(m_words.size());
	for (unsigned int j = 0; j < m_words.size(); ++j) {
		bool unk = false;
		const vector<double>* vect = getWordVector(wordVectors, m_words[j], unk);
		if (unk) {
			//cerr << "In " << trainDevTest << ", found unknown word " << j << ": " << d.m_words[j] << endl;
			if (unkCounts != NULL) {
				map<string, int>& unkCountsReference = *unkCounts;
				unkCountsReference[m_words[j]] += 1;
				// Check if the word is already contained in the map of unknown word counts.
				//map<string, int>::iterator iter = unkCounts->find(m_words[j]);
				//if (iter == unkCounts->end()) {
					// It wasn't found, so insert it.
				//	unkCounts->insert(make_pair(m_words[j], 1));
				//} else {
					// Add 1 to its count.
				//	int oldCount = iter->second;
				//	unkCounts->insert(make_pair(m_words[j], oldCount + 1));
				//}
			}
		}
		m_vects.push_back(vect);
	}
}

void WordSequenceDatum::insertWordsIntoSet(set<string>& wordSet) const {
	// Go through all words.
	for (unsigned int j = 0; j < m_words.size(); ++j) {
		// Insert current word into given set.
		wordSet.insert(m_words[j]);
	}
}
void WordSequenceDatum::doneProcessing() {
	m_vectFormulas.clear();
}

void WordSequenceDatum::sumVectors() {
	m_vectorSum.clear();
	m_vectorAverage.clear();
	if (m_vects.size() > 0) {
		unsigned int dim = m_vects[0]->size();
		m_vectorSum.reserve(dim);
		m_vectorAverage.reserve(dim);
		double numWords = ((double)m_vects.size());
		for (unsigned int k = 0; k < dim; ++k) {
			double dimK = 0.0;
			for (unsigned int j = 0; j < m_vects.size(); ++j) {
				dimK += m_vects[j]->at(k);
			}
			m_vectorSum.push_back(dimK);
			m_vectorAverage.push_back(dimK / numWords);
		}
	}
}

void WordSequenceDatum::clear() {
	m_label = 0;
	m_words.clear();
	m_sequenceLength = 0;
	m_vects.clear();
	m_vectFormulas.clear();
	m_vectParams.clear();
	m_vectorSum.clear();
	m_vectorAverage.clear();
	m_cachedToString = "";
}

void WordSequencePairDatum::doneProcessing() {
	m_sequences.first.doneProcessing();
	m_sequences.second.doneProcessing();
}

MFCCSequenceDatum::~MFCCSequenceDatum() {}

void MFCCSequenceDatum::doneProcessing() {
	m_vectFormulas.clear();
}

void MFCCSequenceDatum::sumVectors() {
	m_vectorSum.clear();
	m_vectorAverage.clear();
	if (m_frames.size() > 0) {
		unsigned int dim = m_frames[0].size();
		m_vectorSum.reserve(dim);
		m_vectorAverage.reserve(dim);
		double numWords = ((double)m_frames.size());
		for (unsigned int k = 0; k < dim; ++k) {
			double dimK = 0.0;
			for (unsigned int j = 0; j < m_frames.size(); ++j) {
				dimK += m_frames[j].at(k);
			}
			m_vectorSum.push_back(dimK);
			m_vectorAverage.push_back(dimK / numWords);
		}
	}
}

double cosineSimilarity(const vector<double>& m1, const vector<double>& m2) {
	FLOAT ret = 0.0f;
	FLOAT m1Norm = 0.0f;
	FLOAT m2Norm = 0.0f;
	unsigned int len1 = m1.size();
	unsigned int len2 = m2.size();
	if (len1 != len2) cerr << "Error: in cosineSimilarity(), m1.size() = " << len1 << " and m2.size() = " << len2 << endl;
	for (unsigned int i = 0; i < len1; ++i) {
		ret += m1[i] * m2[i];
		m1Norm += m1[i] * m1[i];
		m2Norm += m2[i] * m2[i];
	}
	ret = ret / (sqrt(m1Norm) * sqrt(m2Norm));
	return ret;
}

double dotProduct(const vector<double>& m1, const vector<double>& m2) {
	FLOAT ret = 0.0f;
	unsigned int len1 = m1.size();
	unsigned int len2 = m2.size();
	if (len1 != len2) cerr << "Error: in dotProduct(), m1.size() = " << len1 << " and m2.size() = " << len2 << endl;
	for (unsigned int i = 0; i < len1; ++i) {
		ret += m1[i] * m2[i];
	}
	return ret;
}

int loadData(const string& dataFile, vector<Datum>& data, int* numLabels) {
	vector<string> dataLines;
	readLinesFromFile(dataFile, dataLines);
	data.clear();
	unsigned int dim = 0;
	map<string, int> labelMap;
	set<string> uniqueLabels;
	for (unsigned int i = 0; i < dataLines.size(); ++i) {
		vector<string> tokens;
		tokenize(dataLines[i], ',', tokens);
		Datum d;
		d.m_features.reserve(tokens.size() - 1);
		for (unsigned int j = 0; j < tokens.size() - 1; ++j) {
			d.m_features.push_back(atof(tokens[j].c_str()));
		}
		if (i == 0) {
			dim = d.m_features.size();
		} else {
			if (dim != d.m_features.size()) {
				cerr << "Error: while reading data file " << dataFile << ", on line " << i
					 << ", unexpected number of features found: " << d.m_features.size() << ". Expected " << dim << " features." << endl;
			}
		}
		string label = tokens[tokens.size() - 1];
		int intLabel = -1;
		// Check if the label has been seen.
		set<string>::const_iterator iter = uniqueLabels.find(label);
		if (iter == uniqueLabels.end()) {
			intLabel = uniqueLabels.size();
			// It hasn't been seen yet, so create a new mapping for it.
			labelMap.insert(make_pair(label, intLabel));
			// Add it to the set of unique labels.
			uniqueLabels.insert(label);
		} else {
			// It has been seen.
			map<string, int>::const_iterator iter2 = labelMap.find(label);
			if (iter2 == labelMap.end()) {
				cerr << "Error: should not occur!" << endl;
			} else {
				intLabel = iter2->second;
			}
		}
		d.m_label = intLabel;
		data.push_back(d);
	}
	if (dim == 0) cerr << "Error: while reading in data, no features found! "
		"Expected comma-delimited list of features, followed by label in "
		"final position of comma-delimited list." << endl;
	*numLabels = uniqueLabels.size();
	cerr << "Data read in has input dimension " << dim << " and " << (*numLabels) << " labels." << endl;
	return dim;
}

// Scale input features: divide by the maximum absolute value found in the training data.
void scaleFeatures(vector<Datum>& trainData, vector<Datum>& testData) {
	if (trainData.size() > 0) {
		int inputDimension = trainData[0].m_features.size();
		// Initialize the max feature values to 0.0 (since we are finding the max absolute values).
		vector<double> maxFeatureValues(inputDimension, 0.0);
		// Find the max feature absolute values.
		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			Datum& datum = trainData[exampleIndex];
			for (unsigned int featureIndex = 0; featureIndex < datum.m_features.size(); ++featureIndex) {
				double curr = abs(datum.m_features[featureIndex]);
				if (curr > maxFeatureValues.at(featureIndex)) {
					maxFeatureValues[featureIndex] = curr;
				}
			}
		}
		// Divide by the max feature absolute values if nonzero.
		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			Datum& datum = trainData[exampleIndex];
			for (unsigned int featureIndex = 0; featureIndex < datum.m_features.size(); ++featureIndex) {
				if (maxFeatureValues.at(featureIndex) > 0.0) {
					datum.m_features[featureIndex] /= maxFeatureValues[featureIndex];
				}
			}
		}
		for (unsigned int exampleIndex = 0; exampleIndex < testData.size(); ++exampleIndex) {
			Datum& datum = testData[exampleIndex];
			for (unsigned int featureIndex = 0; featureIndex < datum.m_features.size(); ++featureIndex) {
				if (maxFeatureValues.at(featureIndex) > 0.0) {
					datum.m_features[featureIndex] /= maxFeatureValues[featureIndex];
				}
			}
		}
	}
}


// Fills map<string, vector<double> > m_wordVectors.
void readWordVectorFile(const string& filename, const set<string>& vocabulary,
		map<string, vector<double> >& wordVectors, int& wordVectorDimensionality, vector<double>* unknownWordVector) {
	//m_unknownWordVector = NULL;
	unknownWordVector = NULL;
	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);
	int numRead = 0;
	int numSkipped = 0;
	if (filename != NO_FILE_SPECIFIED_STRING && filename != "none" && filename != "") {
		bool useVocabulary = false;
		if (vocabulary.size() > 0) useVocabulary = true;
		ifstream filestream(filename.c_str());
		if (!filestream) {
			cerr << "Error: Cannot open word vector file " << filename << "!" << endl;
		} else {
			// Read the word vector file.
			string line;
			while (!filestream.eof()) {
				getline(filestream, line);
				if (!filestream.eof()) {
					vector<string> lineTokens;
					tokenize(line, '\t', lineTokens);
					if (lineTokens.size() != 2) {
						cerr << "Error: in word vector file " << filename << ", line without 2 tab-separeted tokens: " << line << endl;
					} else {
						const string& w = lineTokens[0];
						bool readVector = true;
						if (useVocabulary) {
							// If we're using the provided vocabulary to filter word vectors,
							// check if the current word is in the vocabulary.
							set<string>::const_iterator vocabIter = vocabulary.find(w);
							if (vocabIter == vocabulary.end()) {
								// If it's not in the vocabulary, don't read the word vector.
								readVector = false;
							}
						}
						if (readVector) {
							const string& theVector = lineTokens[1];
							vector<string> vectorEntries;
							tokenize(theVector, ' ', vectorEntries);
							vector<double> v;
							for (unsigned int i = 0; i < vectorEntries.size(); ++i) {
								v.push_back(atof(vectorEntries[i].c_str()));
							}
							wordVectorDimensionality = vectorEntries.size();
							wordVectors.insert(make_pair(w, v));
							++numRead;
						} else {
							++numSkipped;
						}
					}
				}
			}
			filestream.close();
			cerr << "Word vectors loaded from file " << filename << ": " << wordVectors.size()
				 << ", dimensionality = " << wordVectorDimensionality
				 << ", num read = " << numRead << ", num skipped = " << numSkipped << endl;
			map<string, vector<double> >::iterator unkIter = wordVectors.find(UNK_STRING);
			if (unkIter != wordVectors.end()) {
				unknownWordVector = &(unkIter->second);
			} else {
				cerr << "Error: no vector found for " << UNK_STRING << "!" << endl;
			}
		}
	}
	gettimeofday(&endTime, 0);
	double readTime = computeTimeDifference(initialTime, endTime);
	cerr << "Time to read word vectors:\t" << readTime << endl;

}

const vector<double>* getWordVector(const map<string, vector<double> >& wordVectors,
		const string& word, bool& unk) {
	map<string, vector<double> >::const_iterator iter = wordVectors.find(word);
	if (iter != wordVectors.end()) {
		unk = false;
		return &(iter->second);
	} else {
		// If it didn't match as-is, try looking up the lower-cased form of the word.
		//string lowercasedWord = lowercase(word);
		//for (unsigned int i = 0; i < word.length(); ++i) {
		//	lowercasedWord[i] = tolower(word[i]);
		//}
		map<string, vector<double> >::const_iterator iter2 = wordVectors.find(lowercase(word));
		if (iter2 != wordVectors.end()) {
			unk = false;
			return &(iter2->second);
			//cerr << "Found lowercase match (" << lowercasedWord << ") for word " << word << ". propertyValue = " << propertyValue << endl;
		} else {
			// Check if it's a PTB token that we need special handling for.
			string modifiedWord;
			bool foundMatch = true;
			if (word == "-LRB-") {
				modifiedWord = "(";
			} else if (word == "-RRB-") {
				modifiedWord = ")";
			} else if (word == "\"") {
				modifiedWord = "''";
			} else if (word == "-LCB-") {
				modifiedWord = "{";
			} else if (word == "-RCB-") {
				modifiedWord = "}";
			} else if (word == "-LSB-") {
				modifiedWord = "[";
			} else if (word == "-RSB-") {
				modifiedWord = "]";
			} else {
				bool foundDigits = false;
				string digitsChanged = changeDigitsToGivenChar(word, '#', &foundDigits);
				if (foundDigits) {
					modifiedWord = digitsChanged;
				} else {
					modifiedWord = UNK_STRING;
					foundMatch = false;
				}
			}

			map<string, vector<double> >::const_iterator iter3 = wordVectors.find(modifiedWord);
			if (iter3 != wordVectors.end()) {
				if (!foundMatch) unk = true;
				return &(iter3->second);
			} else {
				//modifiedWord = UNK_STRING;
				//foundMatch = false;
				map<string, vector<double> >::const_iterator iter4 = wordVectors.find(UNK_STRING);
				//cerr << "Error: unknown word vector not found!" << endl;
				unk = true;
				return &(iter4->second);

				//cerr << "Error: unknown word vector not found!" << endl;
				//unk = true;
				//return NULL;
			}
		}
	}
}

const vector<PARAM_ID>* getWordVectorParams(const map<string, vector<PARAM_ID> >& v,
		const string& word, bool& unk) {
	map<string, vector<PARAM_ID> >::const_iterator iter = v.find(word);
	if (iter != v.end()) {
		unk = false;
		return &(iter->second);
	} else {
		// If it didn't match as-is, try looking up the lower-cased form of the word.
		//string lowercasedWord = word;
		//for (unsigned int i = 0; i < word.length(); ++i) {
		//	lowercasedWord[i] = tolower(word[i]);
		//}
		map<string, vector<PARAM_ID> >::const_iterator iter2 = v.find(lowercase(word));
		if (iter2 != v.end()) {
			unk = false;
			return &(iter2->second);
			//cerr << "Found lowercase match (" << lowercasedWord << ") for word " << word << ". propertyValue = " << propertyValue << endl;
		} else {
			// Check if it's a PTB token that we need special handling for.
			string modifiedWord;
			bool foundMatch = true;
			if (word == "-LRB-") {
				modifiedWord = "(";
			} else if (word == "-RRB-") {
				modifiedWord = ")";
			} else if (word == "\"") {
				modifiedWord = "''";
			} else if (word == "-LCB-") {
				modifiedWord = "{";
			} else if (word == "-RCB-") {
				modifiedWord = "}";
			} else if (word == "-LSB-") {
				modifiedWord = "[";
			} else if (word == "-RSB-") {
				modifiedWord = "]";
			} else {
				bool foundDigits = false;
				string digitsChanged = changeDigitsToGivenChar(word, '#', &foundDigits);
				if (foundDigits) {
					modifiedWord = digitsChanged;
				} else {
					modifiedWord = UNK_STRING;
					foundMatch = false;
				}
			}

			map<string, vector<PARAM_ID> >::const_iterator iter3 = v.find(modifiedWord);
			if (iter3 != v.end()) {
				if (!foundMatch) unk = true;
				return &(iter3->second);
			} else {
				//modifiedWord = UNK_STRING;
				//foundMatch = false;
				map<string, vector<PARAM_ID> >::const_iterator iter4 = v.find(UNK_STRING);
				//cerr << "Error: unknown word vector not found!" << endl;
				unk = true;
				return &(iter4->second);
				//cerr << "Error: unknown word vector params not found!" << endl;
				//return NULL;
			}
		}
	}
}

void copyTrainedWordVectorParamValuesToWordVectorMap(Model* model, const map<string, vector<PARAM_ID> >& v,
		map<string, vector<double> >& wordVectors) {
	// Go through all words in v.
	for (map<string, vector<PARAM_ID> >::const_iterator iter = v.begin(); iter != v.end(); ++iter) {
		const string& w = iter->first;
		const vector<PARAM_ID>& paramVector = iter->second;
		// Look for the word in wordVectors.
		map<string, vector<double> >::iterator wvIter = wordVectors.find(w);
		if (wvIter != wordVectors.end()) {
			// The word was found, so erase it.
			wordVectors.erase(wvIter);
		}

		// Create a new entry for the word and insert it.
		vector<double> wordVector;
		wordVector.reserve(paramVector.size());
		for (unsigned int i = 0; i < paramVector.size(); ++i) {
			wordVector.push_back(model->getValue(paramVector.at(i)));
		}
		wordVectors.insert(make_pair(w, wordVector));
	}
}


void addWordsFromLinesToVocabulary(const vector<string>& dataLines, set<string>& vocabulary) {
	for (unsigned int i = 0; i < dataLines.size(); ++i) {
		vector<string> lineTokens;
		tokenize(dataLines[i], '\t', lineTokens);
		for (unsigned int j = 0; j < lineTokens.size(); ++j) {
			vector<string> lineWords;
			tokenize(lineTokens[j], ' ', lineWords);
			for (unsigned int k = 0; k < lineWords.size(); ++k) {
				vocabulary.insert(lineWords[k]);
				vocabulary.insert(lowercase(lineWords[k]));
				bool foundDigits = false;
				string digitsChanged = changeDigitsToGivenChar(lineWords[k], '#', &foundDigits);
				if (foundDigits) vocabulary.insert(digitsChanged);
			}
		}
	}
	vocabulary.insert(UNK_STRING);
}


void loadWordSequenceData(const vector<string>& dataLines, vector<WordSequenceDatum>& data,
		const map<string, vector<double> >& wordVectors, const string& trainDevTest, map<string, int>& unkCounts) {
	//vector<string> dataLines;
	//readLinesFromFile(dataFile, dataLines);
	data.clear();
	for (unsigned int i = 0; i < dataLines.size(); ++i) {
		vector<string> lineTokens;
		tokenize(dataLines[i], '\t', lineTokens);
		WordSequenceDatum d;
		d.readInWords(lineTokens[0], wordVectors, &unkCounts);
		d.m_label = atoi(lineTokens[1].c_str());
		data.push_back(d);
	}
}

void loadWordSequencePairData(const vector<string>& dataLines, vector<WordSequencePairDatum>& data,
		const map<string, vector<double> >& wordVectors, map<string, int>& unkCounts) {
	//vector<string> dataLines;
	//readLinesFromFile(dataFile, dataLines);
	data.clear();
	for (unsigned int i = 0; i < dataLines.size(); ++i) {
		vector<string> lineTokens;
		tokenize(dataLines[i], '\t', lineTokens);
		WordSequencePairDatum d;
		if (lineTokens.size() < 2) {
			cerr << "Error: when loading sequence pair datum, found " << lineTokens.size()
				 << " tab-delimited tokens. Expected at least 2. Line: " << dataLines[i] << endl;
		} else {
			tokenize(lineTokens[0], ' ', d.m_sequences.first.m_words);
			tokenize(lineTokens[1], ' ', d.m_sequences.second.m_words);
			d.m_sequences.first.m_sequenceLength = (int)(d.m_sequences.first.m_words.size());
			d.m_sequences.second.m_sequenceLength = (int)(d.m_sequences.second.m_words.size());
			for (unsigned int j = 0; j < d.m_sequences.first.m_words.size(); ++j) {
				bool unk = false;
				const vector<double>* vect = getWordVector(wordVectors, d.m_sequences.first.m_words[j], unk);
				if (unk) unkCounts[d.m_sequences.first.m_words[j]] += 1;
				d.m_sequences.first.m_vects.push_back(vect);
			}
			for (unsigned int j = 0; j < d.m_sequences.second.m_words.size(); ++j) {
				bool unk = false;
				const vector<double>* vect = getWordVector(wordVectors, d.m_sequences.second.m_words[j], unk);
				if (unk) unkCounts[d.m_sequences.second.m_words[j]] += 1;
				d.m_sequences.second.m_vects.push_back(vect);
			}
			d.m_sequences.first.sumVectors();
			d.m_sequences.second.sumVectors();
			d.m_sequences.first.m_cachedToString = d.m_sequences.first.toString();
			d.m_sequences.second.m_cachedToString = d.m_sequences.second.toString();
			if (lineTokens.size() > 2) {
				d.m_similarity = atof(lineTokens[2].c_str());
			} else {
				d.m_similarity = -1.0;
			}
			data.push_back(d);
		}
	}
}

void loadMFCCSequenceData(const vector<string>& dataLines, vector<MFCCSequenceDatum>& data, map<string, int>& labels) {
	int frameSize = 39;
	data.clear();
	map<string, int>::iterator iter;
	for (unsigned int i = 0; i < dataLines.size(); ++i) {
		vector<string> lineTokens;
		tokenize(dataLines[i], ' ', lineTokens);
		MFCCSequenceDatum d;
		string label = lineTokens[0];
		int labelInteger = -1;
		// Check the map for the current label string.
		iter = labels.find(label);
		if (iter == labels.end()) {
			// It wasn't there, so create a new mapping for it.
			labelInteger = labels.size();
			labels.insert(make_pair(label, labelInteger));
			cerr << "Creating new label mapping: " << label << " --> " << labelInteger << endl;
		} else {
			// It was there, so just take the integer from the mapping.
			labelInteger = iter->second;
		}
		d.m_labelString = label;
		d.m_label = labelInteger;

		// Read in the MFCC vectors in the frames.
		vector<double> frame;
		frame.reserve(frameSize);
		for (int j = 1; j < lineTokens.size(); ++j) {
			frame.push_back(atof(lineTokens[j].c_str()));
			if (j % frameSize == 0) {
				//cerr << "Adding frame of length " << frame.size() << endl;
				d.m_frames.push_back(frame);
				frame.clear();
				frame.reserve(frameSize);
			}
		}
		d.m_sequenceLength = d.m_frames.size();
		//cerr << "Read in " << d.m_sequenceLength << " frames and label " << d.m_labelString << endl;
		data.push_back(d);
	}
}

void precomputeGivenWordVectorFormulas(Model* model, const vector<const vector<double>* >& datumFixedVectors,
		const vector<const vector<PARAM_ID>* >& datumVectorParams, unsigned int dim, bool updateWordVectors, unsigned int numWords,
		vector< vector< Formula* > >& datumVectorFormulas) {
	datumVectorFormulas.clear();
	datumVectorFormulas.reserve(numWords);
	for (unsigned int i = 0; i < numWords; ++i) {
		vector<Formula*> emptyVector;
		emptyVector.reserve(dim);
		datumVectorFormulas.push_back(emptyVector);
	}
	for (unsigned int i = 0; i < numWords; ++i) {
		for (unsigned int j = 0; j < dim; ++j) {
			if (updateWordVectors) {
				datumVectorFormulas[i].push_back(model->getCachedParameterFormula(datumVectorParams.at(i)->at(j)));
			} else {
				datumVectorFormulas[i].push_back(model->getConstantFormulaObject(datumFixedVectors.at(i)->at(j)));
			}
		}
	}
}

void precomputeWordVectorFormulas(Model* model, WordSequenceDatum& datum, unsigned int inputDimension, bool updateWordVectors) {
	precomputeGivenWordVectorFormulas(model, datum.m_vects, datum.m_vectParams,
		inputDimension, updateWordVectors, datum.m_words.size(), datum.m_vectFormulas);
}

void precomputeWordVectorFormulasWithMapping(Model* model, WordSequenceDatum& datum, unsigned int inputDimension, bool updateWordVectors,
	int mappedDimension, vector<vector<PARAM_ID> >& wordVectorMappingMatrix) {
	int numWords = datum.m_words.size();
	datum.m_vectFormulas.clear();
	datum.m_vectFormulas.reserve(numWords);
	for (unsigned int i = 0; i < numWords; ++i) {
		vector<Formula*> emptyVector;
		emptyVector.reserve(mappedDimension);
		datum.m_vectFormulas.push_back(emptyVector);
	}
	for (unsigned int wordIndex = 0; wordIndex < numWords; ++wordIndex) {
		for (unsigned int i = 0; i < mappedDimension; ++i) {
			// Create the sum node for dimension i in the mapped vector.
			Formula* dimi = model->getFormulaObject(PLUS, inputDimension);
			for (unsigned int j = 0; j < inputDimension; ++j) {
				Formula* prod = model->getFormulaObject(TIMES, 2);
				prod->addChild(model->getCachedParameterFormula(wordVectorMappingMatrix.at(j).at(i)));
				if (updateWordVectors) {
					//prod->addChild(datum.m_vectFormulas.at(t).at(j));
					prod->addChild(model->getCachedParameterFormula(datum.m_vectParams.at(wordIndex)->at(j)));
				} else {
					prod->addChild(model->getConstantFormulaObject(datum.m_vects.at(wordIndex)->at(j)));
				}
				dimi->addChild(prod);
			}
			datum.m_vectFormulas[wordIndex].push_back(dimi);
		}
	}
}

void precomputeMFCCFormulas(Model* model, MFCCSequenceDatum& datum, unsigned int inputDimension) {
	datum.m_vectFormulas.clear();
	datum.m_vectFormulas.reserve(datum.m_sequenceLength);
	for (unsigned int i = 0; i < datum.m_sequenceLength; ++i) {
		vector<Formula*> emptyVector;
		emptyVector.reserve(inputDimension);
		datum.m_vectFormulas.push_back(emptyVector);
	}
	for (unsigned int i = 0; i < datum.m_sequenceLength; ++i) {
		for (unsigned int j = 0; j < inputDimension; ++j) {
			datum.m_vectFormulas[i].push_back(model->getConstantFormulaObject(datum.m_frames.at(i).at(j)));
		}
	}
}


void getWordVectorParamsForWordSequenceData(const map<string, vector<PARAM_ID> >& wordVectors, vector<WordSequenceDatum>& data, bool trainData) {
	for (unsigned int i = 0; i < data.size(); ++i) {
		getWordVectorParamsForWordSequenceDatum(wordVectors, data[i], trainData);
	}
}
void getWordVectorParamsForWordSequencePairData(const map<string, vector<PARAM_ID> >& wordVectors, vector<WordSequencePairDatum>& data, bool trainData) {
	for (unsigned int i = 0; i < data.size(); ++i) {
		getWordVectorParamsForWordSequenceDatum(wordVectors, data[i].m_sequences.first, trainData);
		getWordVectorParamsForWordSequenceDatum(wordVectors, data[i].m_sequences.second, trainData);
	}
}

void getWordVectorParamsForWordSequenceDatum(const map<string, vector<PARAM_ID> >& wordVectors, WordSequenceDatum& d, bool trainData) {
	d.m_vectParams.clear();
	d.m_vectParams.reserve(d.m_words.size());
	for (unsigned int j = 0; j < d.m_words.size(); ++j) {
		bool unk = false;
		d.m_vectParams.push_back(getWordVectorParams(wordVectors, d.m_words[j], unk));
		if (unk) {
			if (trainData) {
				cerr << "Error: ";
			} else {
				cerr << "Note: ";
			}
			cerr << "unknown word vector used for word " << d.m_words[j] << endl;
		}
	}
	/*unk = false;
	d.m_vect2Params = getWordVectorParams(params, d.m_words[1], unk);
	if (unk) {
		if (trainData) {
			cerr << "Error: ";
		} else {
			cerr << "Note: ";
		}
		cerr << "unknown word vector used for word " << d.m_words[1] << endl;
	}*/
}


void printUnknownWordCounts(const map<string, int>& unkWordCounts) {
	cerr << "Unknown words with counts:" << endl;
	for (map<string, int>::const_iterator iter = unkWordCounts.begin(); iter != unkWordCounts.end(); ++iter) {
		cerr << iter->first << "\t\t" << iter->second << endl;
	}
}

void insertWordSequenceDataWordsIntoSet(const vector<WordSequenceDatum>& data, set<string>& wordSet) {
	for (unsigned int i = 0; i < data.size(); ++i) {
		data[i].insertWordsIntoSet(wordSet);
	}
}
void insertWordSequencePairDataWordsIntoSet(const vector<WordSequencePairDatum>& data, set<string>& wordSet) {
	for (unsigned int i = 0; i < data.size(); ++i) {
		data[i].m_sequences.first.insertWordsIntoSet(wordSet);
		data[i].m_sequences.second.insertWordsIntoSet(wordSet);
	}
}


void setVerticalMatrixDimensions(int dataInputDimension, vector<int>& inputDimensions, vector<int>& outputDimensions, vector<int>& layerWidths) {
	// Parameters for observation-gate matrices; more generally, these matrices connect variables vertically.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix 0: m_inputDimension x m_layerWidths[0] (because they connect the input to the first hidden layer).
	// Dimensions of matrix i > 0: m_layerWidths[i-1] x m_layerWidths[i].
	inputDimensions.clear();
	outputDimensions.clear();
	inputDimensions.push_back(dataInputDimension);
	int numLayers = ((int)(layerWidths.size()));
	for (int i = 0; i < numLayers - 1; ++i) {
		inputDimensions.push_back(layerWidths[i]);
		outputDimensions.push_back(layerWidths[i]);
	}
	outputDimensions.push_back(layerWidths[numLayers - 1]);
	if (inputDimensions.size() != outputDimensions.size()) {
		cerr << "Error: vertical matrix inputDimensions.size() != vertical matrix outputDimensions.size()!" << endl;
	}
	cerr << "Printing vertical matrix input and output dimensions:" << endl;
	for (unsigned int i = 0; i < inputDimensions.size(); ++i) {
		cerr << "verticalInput[" << i << "] = " << inputDimensions[i] << endl;
		cerr << "verticalOutput[" << i << "] = " << outputDimensions[i] << endl;
	}
}

void setHorizontalMatrixDimensions(int dataInputDimension, vector<int>& inputDimensions, vector<int>& outputDimensions, vector<int>& layerWidths) {
	// Parameters for hidden-gate matrices; more generally, these matrices connect variables horizontally.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_layerWidths[i].
	inputDimensions.clear();
	outputDimensions.clear();
	int numLayers = ((int)(layerWidths.size()));
	for (int i = 0; i < numLayers; ++i) {
		inputDimensions.push_back(layerWidths[i]);
		outputDimensions.push_back(layerWidths[i]);
	}
	if (inputDimensions.size() != outputDimensions.size()) {
		cerr << "Error: horizontal matrix inputDimensions.size() != horizontal matrix outputDimensions.size()!" << endl;
	}
	cerr << "Printing horizontal matrix input and output dimensions:" << endl;
	for (unsigned int i = 0; i < inputDimensions.size(); ++i) {
		cerr << "horizontalInput[" << i << "] = " << inputDimensions[i] << endl;
		cerr << "horizontalOutput[" << i << "] = " << outputDimensions[i] << endl;
	}
}

void setSoftmaxMatrixDimensions(int numLabels, vector<int>& inputDimensions, vector<int>& outputDimensions, vector<int>& layerWidths) {
	// Parameters for softmax weight matrices. Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_numLabels.
	inputDimensions.clear();
	outputDimensions.clear();
	int numLayers = ((int)(layerWidths.size()));
	for (int i = 0; i < numLayers; ++i) {
		inputDimensions.push_back(layerWidths[i]);
	}
	for (int i = 0; i < numLayers; ++i) {
		outputDimensions.push_back(numLabels);
	}
	if (inputDimensions.size() != outputDimensions.size()) {
		cerr << "Error: softmax matrix inputDimensions.size() != softmax matrix outputDimensions.size()!" << endl;
	}
	cerr << "Printing softmax matrix input and output dimensions:" << endl;
	for (unsigned int i = 0; i < inputDimensions.size(); ++i) {
		cerr << "softmaxInput[" << i << "] = " << inputDimensions[i] << endl;
		cerr << "softmaxOutput[" << i << "] = " << outputDimensions[i] << endl;
	}
}

void resetDeepCachedNodes(unsigned int numCachedNodes, vector<int>& layerWidths, vector<vector<vector<Formula*> > >& cachedNodes) {
	// Clear all data structures containing cached formulas from the previous datum, and set them up for the new datum.
	if (cachedNodes.size() == numCachedNodes) {
		for (unsigned int i = 0; i < cachedNodes.size(); ++i) {
			for (unsigned int j = 0; j < layerWidths.size(); ++j) {
				if (cachedNodes[i][j].size() != layerWidths[j]) {
					cerr << "Error: params.m_cachedNodes[" << i << "][" << j << "].size() != layerWidths[" << j << "] (" << layerWidths[j] << ")" << endl;
				} else {
					for (unsigned int k = 0; k < cachedNodes[i][j].size(); ++k) {
						cachedNodes[i][j][k] = NULL;
					}
				}
			}
		}
	} else {
		cachedNodes.clear();
		cachedNodes.reserve(numCachedNodes);
		for (unsigned int i = 0; i < numCachedNodes; ++i) {
			vector<vector<Formula*> > matrixi;
			matrixi.reserve(layerWidths.size());
			for (unsigned int j = 0; j < layerWidths.size(); ++j) {
				vector<Formula*> rowj(layerWidths[j], NULL);
				matrixi.push_back(rowj);
			}
			cachedNodes.push_back(matrixi);
		}
	}
}

/*void resetCachedNodes(unsigned int numCachedNodes, int layerWidth, vector<vector<Formula*> >& cachedNodes) {
	// Clear all data structures containing cached formulas from the previous datum, and set them up for the new datum.
	if (cachedNodes.size() == numCachedNodes) {
		for (unsigned int i = 0; i < cachedNodes.size(); ++i) {
			if (cachedNodes[i].size() != layerWidth) {
				cerr << "Error: params.m_cachedNodes[" << i << "].size() != " << layerWidth << endl;
			} else {
				for (unsigned int j = 0; j < cachedNodes[i].size(); ++j) {
					cachedNodes[i][j] = NULL;
				}
			}
		}
	} else {
		cachedNodes.clear();
		cachedNodes.reserve(numCachedNodes);
		for (unsigned int i = 0; i < numCachedNodes; ++i) {
			vector<Formula*> rowi(layerWidth, NULL);
			cachedNodes.push_back(rowi);
		}
	}
}*/

