// This file contains utilities for working with data consisting of word sequences
// where each word is represented by a vector.
//
// author: Kevin Gimpel
// date: 2014-2015
//
#ifndef CLASSIFIER_UTILS_H_
#define CLASSIFIER_UTILS_H_

#include "../utils/utils.h"
#include <set>
#include <map>

class Formula;
class Model;

typedef float FLOAT;
typedef unsigned int PARAM_ID;

static const string UNK_STRING = "UUUNKKK";
static const string NO_FILE_SPECIFIED_STRING = "|||NO_FILE_SPECIFIED|||";

struct Datum {
	vector<double> m_features;
	int m_label;
};

struct SequenceDatum {
	int m_label;
	int m_sequenceLength;
	// Cached Formula nodes for word vectors in this sequence. Dimensionality is m_sequenceLength x dimension of each item in the sequence.
	vector< vector< Formula* > > m_vectFormulas;
	virtual ~SequenceDatum();
	virtual void doneProcessing() = 0;
	virtual void sumVectors() = 0;
};

struct WordSequenceDatum : SequenceDatum {
	vector<string> m_words;
	//int m_numWords;
	//int m_label;
	vector<const vector<double>* > m_vects;
	vector<const vector<PARAM_ID>* > m_vectParams;
	// TODO: following is just for efficiency. Can be removed to save memory.
	vector<double> m_vectorSum;
	vector<double> m_vectorAverage;
	string m_cachedToString;
	string toString() const;
	string toStringInsertStringAfterWordIndex(const string& toInsert, int pos) const;
	string getWordsInRange(int startIndex, int endIndex) const;
	string getWordsInRangePadded(int startIndex, int endIndex) const;
	// Reads in the word sequence contained in s (assumed to be space-separated), using the word vectors given.
	void readInWords(const string& s, const map<string, vector<double> >& wordVectors, map<string, int>* unkCounts);
	void insertWordsIntoSet(set<string>& wordSet) const;
	virtual ~WordSequenceDatum();
	virtual void doneProcessing();
	virtual void sumVectors();
	void clear();
};

struct WordSequencePairDatum {
	pair<WordSequenceDatum, WordSequenceDatum> m_sequences;
	double m_similarity;
	void doneProcessing();
};

struct MFCCSequenceDatum : SequenceDatum {
	//int m_numFrames;
	string m_labelString;
	//int m_label;
	// MFCC features in each frame. Outer vector ranges over frames, inner vector ranges over MFCC features.
	// I.e., dimensionality is m_numFrames x MFCC feature vector dimension.
	vector<vector<double> > m_frames;
	vector<double> m_vectorSum;
	vector<double> m_vectorAverage;
	virtual ~MFCCSequenceDatum();
	virtual void doneProcessing();
	virtual void sumVectors();
};

double cosineSimilarity(const vector<double>& m1, const vector<double>& m2);
double dotProduct(const vector<double>& m1, const vector<double>& m2);
int loadData(const string& dataFile, vector<Datum>& data, int* numLabels);
void scaleFeatures(vector<Datum>& trainData, vector<Datum>& testData);
void readWordVectorFile(const string& filename, const set<string>& vocabulary,
	map<string, vector<double> >& wordVectors, int& wordVectorDimensionality, vector<double>* unknownWordVector);
const vector<double>* getWordVector(const map<string, vector<double> >& wordVectors, const string& word, bool& unk);
const vector<PARAM_ID>* getWordVectorParams(const map<string, vector<PARAM_ID> >& v, const string& word, bool& unk);

void copyTrainedWordVectorParamValuesToWordVectorMap(Model* model, const map<string, vector<PARAM_ID> >& v,
	map<string, vector<double> >& wordVectors);

// Tokenizes data lines around tab, then tokenizes each tab-delimited string around space, and adds all space-delimited
// tokens to the given vocabulary. So this adds labels as well (in order to handle single-sequence and sequence-pair data),
// which is unusual but should only increase the vocabulary slightly relative to the number of words in the pre-label fields of each line.
void addWordsFromLinesToVocabulary(const vector<string>& dataLines, set<string>& vocabulary);
void loadWordSequenceData(const vector<string>& dataLines, vector<WordSequenceDatum>& data, const map<string, vector<double> >& wordVectors,
		const string& trainDevTest, map<string, int>& unkCounts);
void loadWordSequencePairData(const vector<string>& dataLines, vector<WordSequencePairDatum>& data,
		const map<string, vector<double> >& wordVectors, map<string, int>& unkCounts);
void loadMFCCSequenceData(const vector<string>& dataLines, vector<MFCCSequenceDatum>& data, map<string, int>& labels);

void precomputeGivenWordVectorFormulas(Model* model, const vector<const vector<double>* >& datumFixedVectors,
	const vector<const vector<PARAM_ID>* >& datumVectorParams, unsigned int dim, bool updateWordVectors, unsigned int numWords,
	vector< vector< Formula* > >& datumVectorFormulas);
void precomputeWordVectorFormulas(Model* model, WordSequenceDatum& datum, unsigned int inputDimension, bool updateWordVectors);

void precomputeWordVectorFormulasWithMapping(Model* model, WordSequenceDatum& datum, unsigned int inputDimension, bool updateWordVectors, int mappedDimension, vector<vector<PARAM_ID> >& wordVectorMappingMatrix);

void precomputeMFCCFormulas(Model* model, MFCCSequenceDatum& datum, unsigned int inputDimension);
void getWordVectorParamsForWordSequenceData(const map<string, vector<PARAM_ID> >& wordVectors, vector<WordSequenceDatum>& data, bool trainData);
void getWordVectorParamsForWordSequencePairData(const map<string, vector<PARAM_ID> >& wordVectors, vector<WordSequencePairDatum>& data, bool trainData);
void getWordVectorParamsForWordSequenceDatum(const map<string, vector<PARAM_ID> >& wordVectors, WordSequenceDatum& d, bool trainData);
void printUnknownWordCounts(const map<string, int>& unkWordCounts);
void insertWordSequenceDataWordsIntoSet(const vector<WordSequenceDatum>& data, set<string>& wordSet);
void insertWordSequencePairDataWordsIntoSet(const vector<WordSequencePairDatum>& data, set<string>& wordSet);
void setVerticalMatrixDimensions(int dataInputDimension, vector<int>& inputDimensions, vector<int>& outputDimensions, vector<int>& layerWidths);
void setHorizontalMatrixDimensions(int dataInputDimension, vector<int>& inputDimensions, vector<int>& outputDimensions, vector<int>& layerWidths);
void setSoftmaxMatrixDimensions(int numLabels, vector<int>& inputDimensions, vector<int>& outputDimensions, vector<int>& layerWidths);
void resetDeepCachedNodes(unsigned int numCachedNodes, vector<int>& layerWidths, vector<vector<vector<Formula*> > >& cachedNodes);

#endif /* CLASSIFIER_UTILS_H_ */
