#include <map>
#include <set>
#include <queue>
#include <algorithm>
#include <stdlib.h>
#include <sys/time.h>
#include "../classifier_utils.h"
#include "cnn.h"

// public:
CompositionalConvolutionalSentenceModel::CompositionalConvolutionalSentenceModel(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity) : Model(unaryFormulaCapacity, binaryFormulaCapacity, fertileFormulaCapacity, barrenFormulaCapacity) {
	m_numLabels = 2;
	m_numFilters = 0;
	m_inputDimension = 0;
	m_updateWordVectors = false;
	m_useStochasticPooling = false;
	m_useCosineSimilarityFilters = false;
	m_unknownWordVector = NULL;
	m_maxSequenceLength = -1;
}

CompositionalConvolutionalSentenceModel::~CompositionalConvolutionalSentenceModel() {}

double CompositionalConvolutionalSentenceModel::trainAndTest(const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename,
		const string& filters, const string& algorithm, double stepsize, int minibatchSize, int numEpochs,
		int decodeEveryK, double dropoutRate,
		bool updateWordVectors, bool initWordVectorsRandomly, bool stochasticPooling, double l2NormConstraint,
		bool useCosineSimilarityFilters, bool doFiniteDifferenceChecking) {
	struct timeval initialTime, endTime, time1, time2, time3, time4,
		timeBuildStart, timeBuildEnd, timeEvalStart, timeEvalEnd,
		timeDecodeStart, timeDecodeEnd;
	gettimeofday(&initialTime, 0);
	m_updateWordVectors = updateWordVectors;
	m_useStochasticPooling = stochasticPooling;
	m_useCosineSimilarityFilters = useCosineSimilarityFilters;

	// First, load word vectors corresponding to words in train/dev/test data.
	loadWordVectors(trainLines, devLines, testLines, wordVectorFilename);

	resetGradientVector();
	//double stepsize = 0.001;
	map<string, int> unkWordCounts;
	vector<WordSequenceDatum> trainData, devData, testData;
	loadWordSequenceData(trainLines, trainData, m_wordVectors, "train", unkWordCounts);
	loadWordSequenceData(devLines, devData, m_wordVectors, "dev", unkWordCounts);
	loadWordSequenceData(testLines, testData, m_wordVectors, "test", unkWordCounts);
	//printUnknownWordCounts(unkWordCounts);

	bool trainTrue = true;
	bool trainFalse = false;

	// Initialize all parameters shared across LSTMs/GRUs.
	//initializeSharedParameters(model, params, params.m_wordVectorDimensionality, trainData, devData,
	//	testData, updateWordVectors, initWordVectorsRandomly);
	// Initialize parameters for each LSTM/GRU.
	//initializeNetworkFromConfiguration(model, params, forwardLSTMs, true, true);
	//initializeNetworkFromConfiguration(model, params, backwardLSTMs, true, false);
	//initializeNetworkFromConfiguration(model, params, forwardGRUs, false, true);
	//initializeNetworkFromConfiguration(model, params, backwardGRUs, false, false);

	initializeAllParameters(trainData, devData, testData, initWordVectorsRandomly, filters);
	if (m_updateWordVectors) {
		cerr << "Getting word vector params for training data" << endl;
		getWordVectorParamsForWordSequenceData(m_v, trainData, trainTrue);
		cerr << "Getting word vector params for dev data" << endl;
		getWordVectorParamsForWordSequenceData(m_v, devData, trainFalse);
		cerr << "Getting word vector params for test data" << endl;
		getWordVectorParamsForWordSequenceData(m_v, testData, trainFalse);
	}

	// Print nearest neighbors before any learning.
	//printNearestNeighbors(trainData, trainData, params);
	//printNearestNeighbors(trainData, devData, params);
	double maxDevAcc = -1.0;
	double testAccForMaxDevAcc = -1.0;
	//double maxTestAcc = -1.0;

	FLOAT totalBuildTime = 0.0f;
	FLOAT totalEvalTime = 0.0f;
	FLOAT totalDecodeTime = 0.0f;

	int decodeEveryKToUse = decodeEveryK;
	if (trainData.size() < 11000) {
		decodeEveryKToUse = trainData.size() / 5;
	}

	bool doDropout = true;
	if (dropoutRate <= 0.0) doDropout = false;

	bool scaleWeightsForDecoding = doDropout;
	scaleWeightsForDecoding = false;
	if (scaleWeightsForDecoding) {
		cerr << "Scaling weights for decoding by dropout rate (" << dropoutRate << ")." << endl;
	}

	Algorithm alg = getAlgorithm(algorithm);
	// Initialize vector of previous updates for SGD w/ momentum.
	vector<double> sgdMomentumUpdates;//(getNumParams(), 0.0);
	double momentumCoefficient = 0.95;
	// Initialize sum of squared gradients for AdaGrad and AdaDelta.
	vector<double> sumOfSquaredGradients(getNumParams(), 0.0);
	// Initialize sum of squared updates for AdaDelta.
	vector<double> sumOfSquaredUpdates(getNumParams(), 0.0);
	double epsilon = 0.000001;
	double rho = 0.95;
	if (alg != ADAGRAD) {
		cerr << "Error: only AdaGrad is currently supported!" << endl;
	}
	// Initialize sum of squared gradients for AdaGrad.
	//vector<double> sumOfSquaredGradients(getNumParams(), 0.0);
	//int maxNumFormulasUsed = getNumberOfFormulasUsedInPool();
	int maxUnaryFormulasUsed = 0;
	int maxBinaryFormulasUsed = 0;
	int maxFertileFormulasUsed = 0;
	int maxBarrenFormulasUsed = 0;
	int unaryFormulasUsed = 0;
	int binaryFormulasUsed = 0;
	int fertileFormulasUsed = 0;
	int barrenFormulasUsed = 0;
	getNumberOfFormulasUsedInPool(maxUnaryFormulasUsed, maxBinaryFormulasUsed, maxFertileFormulasUsed, maxBarrenFormulasUsed);

	// Train for numEpochs iterations of training.
	for (unsigned int iter = 1; iter <= numEpochs; ++iter) {
		double iterLoss = 0.0;
		double iterGradientNorm = 0.0;

		//int iterNumCorrect = 0;
		gettimeofday(&time1, 0);
		gettimeofday(&time3, 0);

		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			WordSequenceDatum& datum = trainData[exampleIndex];
			// Sample a data point.
			//int exampleIndex = i;//rand() % (trainData.size());
			// We are processing a new example, so reset Formula data structures in the model.
			resetForNewExample(datum);

			//cerr << "Computing gradient for sentence \"" << sentenceTokens[0] << " " << sentenceTokens[1]
			//	 << "\" with true label " << trueLabel << endl;

			// Create the formula containing the loss for the current example.
			gettimeofday(&timeBuildStart, 0);
			Formula* datumLossFDAG = createLossFDAGForWordSequenceDatum(datum, exampleIndex, doDropout, dropoutRate);
			gettimeofday(&timeBuildEnd, 0);
			totalBuildTime += computeTimeDifference(timeBuildStart, timeBuildEnd);
			getNumberOfFormulasUsedInPool(unaryFormulasUsed, binaryFormulasUsed, fertileFormulasUsed, barrenFormulasUsed);
			if (unaryFormulasUsed > maxUnaryFormulasUsed) {
				maxUnaryFormulasUsed = unaryFormulasUsed;
				cerr << "Info: New unary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << unaryFormulasUsed << endl;
			}
			if (binaryFormulasUsed > maxBinaryFormulasUsed) {
				maxBinaryFormulasUsed = binaryFormulasUsed;
				cerr << "Info: New binary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << binaryFormulasUsed << endl;
			}
			if (fertileFormulasUsed > maxFertileFormulasUsed) {
				maxFertileFormulasUsed = fertileFormulasUsed;
				cerr << "Info: New fertile formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << fertileFormulasUsed << endl;
			}
			if (barrenFormulasUsed > maxBarrenFormulasUsed) {
				maxBarrenFormulasUsed = barrenFormulasUsed;
				cerr << "Info: New barren formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << barrenFormulasUsed << endl;
			}
			//int currNumFormulasUsed = getNumberOfFormulasUsedInPool();
			//if (currNumFormulasUsed > maxNumFormulasUsed) {
			//	cerr << "Info: New formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << currNumFormulasUsed << endl;
			//	maxNumFormulasUsed = currNumFormulasUsed;
			//}
			// Compute gradient and loss for current example.
			gettimeofday(&timeEvalStart, 0);
			double loss = evalAndComputeDerivatives(datumLossFDAG, exampleIndex, doFiniteDifferenceChecking);
			if (loss != loss) {
				//cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
				cerr << "Error: nan loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].toString() <<
					"\" with true label " << trainData[exampleIndex].m_label << endl;
			} else {
				iterLoss += loss;
			}
			gettimeofday(&timeEvalEnd, 0);
			totalEvalTime += computeTimeDifference(timeEvalStart, timeEvalEnd);
			datum.doneProcessing();

			// If we're at the end of a mini-batch or at the end of the training data, make an update.
			bool doUpdate = false;
			if ((exampleIndex + 1) % minibatchSize == 0 || trainData.size() <= 1 || exampleIndex == trainData.size() - 1) doUpdate = true;
			if (doUpdate) {
				double gradientSquaredNorm = 0.0;
				if (alg == SGD_MOMENTUM) {
					gradientSquaredNorm = doSGDMomentumUpdate(sgdMomentumUpdates, stepsize, momentumCoefficient);
				} else if (alg == ADAGRAD) {
					gradientSquaredNorm = doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				} else if (alg == ADADELTA){
					gradientSquaredNorm = doAdaDeltaUpdate(sumOfSquaredGradients, sumOfSquaredUpdates, epsilon, rho);
				}
				double l2norm = sqrt(gradientSquaredNorm);
				iterGradientNorm += l2norm;
				if (l2NormConstraint > 0.0) {// && l2norm > l2NormConstraint) {
					scaleWeightsToMatchL2Norm(l2NormConstraint);
				}
				resetGradientVector();

				//double gradientSquaredNorm = doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				//double l2norm = sqrt(gradientSquaredNorm);
				//iterGradientNorm += l2norm;
				//if (l2NormConstraint > 0.0) {// && l2norm > l2NormConstraint) {
				//	scaleWeightsToMatchL2Norm(params, l2NormConstraint);
				//}
				//resetGradientVector();
			}

			if ((exampleIndex + 1) % decodeEveryKToUse == 0) {

				gettimeofday(&timeDecodeStart, 0);
				gettimeofday(&time4, 0);
				FLOAT timek = computeTimeDifference(time3, time4);

				// Decode dev data every decodeEveryK examples.
				if (devData.size() > 0) {
					// If we have dev data, decode the dev data.
					double devAccuracy = doDecoding(devData, scaleWeightsForDecoding, dropoutRate);
					// If it's the best dev accuracy we've seen so far, decode the test data.
					if (devAccuracy > maxDevAcc) {
						maxDevAcc = devAccuracy;
						testAccForMaxDevAcc = doDecoding(testData, scaleWeightsForDecoding, dropoutRate);
					}
					//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
					cerr << "it " << iter << ", " << exampleIndex << "\tdacc\t" << devAccuracy << "\tmaxd\t" << maxDevAcc
						 << "\ttformaxd\t" << testAccForMaxDevAcc; // << "\tmaxt\t" << maxTestAcc;
				} else {
					// Here, we have no dev data, so just decode the test data.
					testAccForMaxDevAcc = doDecoding(testData, scaleWeightsForDecoding, dropoutRate);
					cerr << "it " << iter << ", " << exampleIndex << "\ttacc\t" << testAccForMaxDevAcc;
				}

				cerr << "\ttimek\t" << timek << endl;
				gettimeofday(&time3, 0);
				gettimeofday(&timeDecodeEnd, 0);
				totalDecodeTime += computeTimeDifference(timeDecodeStart, timeDecodeEnd);
				cerr << "times:\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
					 << "\tdec\t" << totalDecodeTime << endl;
				//printFilterNearestNeighbors(trainData, params);

				//cerr << "times:\treset\t" << totalResetTime << "\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
				//	 << "\tupdate\t" << totalUpdateTime << "\tdec\t" << totalDecodeTime << endl;
			}
		}
		//if (iter % 10 == 0) cerr << "it " << iter << "\tloss\t" << iterLoss <<
		//		"\tparam_norm\t" << computeParameterVectorNorm() <<
		//		"\tgrad_norm\t" << iterGradientNorm << endl;
		cerr << "it " << iter << "\tloss\t" << iterLoss << "\tparam_norm\t" << computeParameterVectorNorm() <<
				"\tgrad_norm\t" << iterGradientNorm;
		//double acc = ((double)iterNumCorrect) / ((double)(trainData.size()));
		//cerr << "\ttracc\t" << acc;
		//double trainAccuracy = doDecoding(trainData, m, w, dim, numLabels);
		//cerr << "\ttacc:\t" << trainAccuracy;


		if (devData.size() > 0) {
			// If we have dev data, decode the dev data.
			double devAccuracy = doDecoding(devData, scaleWeightsForDecoding, dropoutRate);
			// If it's the best dev accuracy we've seen so far, decode the test data.
			if (devAccuracy > maxDevAcc) {
				maxDevAcc = devAccuracy;
				testAccForMaxDevAcc = doDecoding(testData, scaleWeightsForDecoding, dropoutRate);
			}
			//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
			cerr << "\tdacc\t" << devAccuracy << "\tmaxd\t" << maxDevAcc
			<< "\ttformaxd\t" << testAccForMaxDevAcc;// << "\tmaxt\t" << maxTestAcc;
		} else {
			// Here, we have no dev data, so just decode the test data.
			testAccForMaxDevAcc = doDecoding(testData, scaleWeightsForDecoding, dropoutRate);
			cerr << "\ttacc\t" << testAccForMaxDevAcc;// << "\tmaxt\t" << maxTestAcc;
		}
		gettimeofday(&time2, 0);
		FLOAT iterTime = computeTimeDifference(time1, time2);
		cerr << "\t\ttime\t" << iterTime;
		cerr << endl;

		//if (iter % 5 == 0)
			printFilterNearestNeighbors(trainData);
	}
	// Print nearest neighbors for dev bigrams.
	//printNearestNeighbors(trainData, devData, params);

	gettimeofday(&endTime, 0);
	FLOAT cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
	cerr << "Final test accuracy:\t" << testAccForMaxDevAcc << endl;
	return testAccForMaxDevAcc;
}

void CompositionalConvolutionalSentenceModel::loadWordVectors(const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename) {
	// Build the relevant word vector vocabulary from the data lines.
	set<string> vocab;
	addWordsFromLinesToVocabulary(trainLines, vocab);
	addWordsFromLinesToVocabulary(devLines, vocab);
	addWordsFromLinesToVocabulary(testLines, vocab);
	// Read in word vectors corresponding to the entries in the vocabulary.
	readWordVectorFile(wordVectorFilename, vocab, m_wordVectors, m_inputDimension, m_unknownWordVector);
}

void CompositionalConvolutionalSentenceModel::initializeAllParameters(const vector<WordSequenceDatum>& trainData,
		const vector<WordSequenceDatum>& devData, const vector<WordSequenceDatum>& testData, bool initWordVectorsRandomly,
		const string& filters) {
	//m_inputDimension = m_wordVectorDimensionality;
	m_numLabels = 2;

	m_nbowBias.clear();
	m_softmaxWeights.clear();
	m_softmaxBias.clear();

	// Go through all sequences and find the maximum length. This will be used as the number of
	// simultaneous sequences in the LSTM and GRU filters.
	m_maxSequenceLength = -1;
	for (unsigned int i = 0; i < trainData.size(); ++i) {
		if (trainData[i].m_sequenceLength > m_maxSequenceLength) m_maxSequenceLength = trainData[i].m_sequenceLength;
	}
	for (unsigned int i = 0; i < devData.size(); ++i) {
		if (devData[i].m_sequenceLength > m_maxSequenceLength) m_maxSequenceLength = devData[i].m_sequenceLength;
	}
	for (unsigned int i = 0; i < testData.size(); ++i) {
		if (testData[i].m_sequenceLength > m_maxSequenceLength) m_maxSequenceLength = testData[i].m_sequenceLength;
	}
	cerr << "Maximum length of any sequence found to be: " << m_maxSequenceLength << endl;

	double scalingFactor = 0.2;

	if (filters != "" && filters != "0") {
		vector<string> filterStrings;
		tokenize(filters, ',', filterStrings);
		for (unsigned int i = 0; i < filterStrings.size(); ++i) {
			vector<string> filterTokens;
			tokenize(filterStrings[i], ':', filterTokens);
			if (filterTokens.size() == 2 || filterTokens.size() == 3) {
				int filterWidth = atoi(filterTokens[0].c_str());
				int numberOfFilters = atoi(filterTokens[1].c_str());
				FILTER_CTYPE ctype = CTYPE_NONE;
				if (filterTokens.size() == 3) {
					if (filterTokens[2] == "n") {
						ctype = CTYPE_NONE;
					} else if (filterTokens[2] == "s") {
						ctype = CTYPE_SUM;
					} else if (filterTokens[2] == "w1") {
						ctype = CTYPE_WSUM1;
					} else if (filterTokens[2] == "w2") {
						ctype = CTYPE_WSUM2;
					} else if (filterTokens[2][0] == 'l') {
						ctype = CTYPE_LSTM;
					} else if (filterTokens[2][0] == 'g') {
						ctype = CTYPE_GRU;
					} else {
						cerr << "Error: unexpected composition type: " << filterTokens[2] << ". Using \"none\"." << endl;
						ctype = CTYPE_NONE;
					}
				}

				bool forward = true;
				string hiddenVectorSizeString = "";
				int hiddenVectorSizeInt = -1;
				bool initSoftmaxFalse = false;
				if (ctype == CTYPE_LSTM || ctype == CTYPE_GRU) {
					if (filterTokens[2].length() == 1) {
						cerr << "Error: only \"" << filterTokens[2] << "\" specified as filter type. Expected direction (\"f\" or \"b\") plus size of hidden vector. Omitting filters of this type." << endl;
					} else {
						if (filterTokens[2][1] == 'f') {
							forward = true;
						} else if (filterTokens[2][1] == 'b') {
							forward = false;
						} else {
							cerr << "Error: unexpected direction character given for LSTM filter type: " << filterTokens[2][1] << endl;
						}
						cerr << "Initializing " << numberOfFilters << " ";
						if (forward)
							cerr << "forward";
						else
							cerr << "backward";
						hiddenVectorSizeString = filterTokens[2].substr(2);
						hiddenVectorSizeInt = atoi(hiddenVectorSizeString.c_str());
						if (ctype == CTYPE_LSTM) cerr << " LSTM ";
						if (ctype == CTYPE_GRU) cerr << " GRU ";
						cerr << "filter with hidden vector of size " << hiddenVectorSizeString << endl;
						if (!forward) cerr << "Error: backward LSTM/GRU not yet implemented for convolutional filter!" << endl;
					}
				}

				for (unsigned int j = 0; j < numberOfFilters; ++j) {
					// Initialize a filter with the given width.
					Filter f;
					if (ctype == CTYPE_NONE) {
						f.m_width = filterWidth;
						int numWeights = filterWidth * m_inputDimension;
						for (unsigned int k = 0; k < numWeights; ++k) {
							f.m_weights.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
						}
						f.m_maxWords = -1;
						f.m_bias = initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5));

					} else if (ctype == CTYPE_SUM || ctype == CTYPE_WSUM1 || ctype == CTYPE_WSUM2){
						// If this is a composition filter, set the filter width to be 1, since it's only
						// the size of a single word vector.
						f.m_width = 1;
						int numWeights = m_inputDimension;
						for (unsigned int k = 0; k < numWeights; ++k) {
							f.m_weights.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
						}

						// TODO if w1 or w2, create extra parameters here.

						f.m_maxWords = filterWidth;
						// Create a bias for each possible length of n-gram.
						f.m_ngramLengthBiases.clear();
						for (unsigned int k = 0; k < f.m_maxWords; ++k) {
							f.m_ngramLengthBiases.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
						}
						f.m_bias = initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5));

					} else if (ctype == CTYPE_LSTM) {
						f.m_width = filterWidth;
						for (unsigned int k = 0; k < hiddenVectorSizeInt; ++k) {
							f.m_weights.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
						}
						f.m_bias = initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5));
						f.m_maxWords = -1;
						f.m_lstmParams.initializeParameters(*this, m_inputDimension, m_numLabels, forward, hiddenVectorSizeString, m_maxSequenceLength, initSoftmaxFalse, true, true, true, false);
					} else if (ctype == CTYPE_GRU) {
						f.m_width = filterWidth;
						for (unsigned int k = 0; k < hiddenVectorSizeInt; ++k) {
							f.m_weights.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
						}
						f.m_bias = initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5));
						f.m_maxWords = -1;
						f.m_gruParams.initializeParameters(*this, m_inputDimension, m_numLabels, forward, hiddenVectorSizeString, m_maxSequenceLength, initSoftmaxFalse, true, true);
					}
					// Set the composition type.
					f.m_compositionType = ctype;

					m_filters.push_back(f);
				}
				cerr << "Created " << numberOfFilters << " filters of width " << filterWidth << " and ctype " << ctype << "." << endl;
			} else {
				cerr << "Error: expected each (comma-delimited) filter to have the format \"width:number\" "
						"or \"width:number:ctype\". Given: " << filterStrings[i] << endl;
			}
		}
	}

	m_numFilters = m_filters.size();
	cerr << "Using " << m_numFilters << " convolutional filters." << endl;
	// We need to create a fully-connected layer after the max-over-time pooling layer.

	if (m_numFilters == 0) {
		for (int j = 0; j < m_inputDimension; ++j) {
			m_nbowBias.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
		}
	}

	int outputDimension = m_inputDimension;
	if (m_numFilters > 0) outputDimension = m_numFilters;
	for (unsigned int i = 0; i < m_numLabels; ++i) {
		vector<PARAM_ID> rowi;
		for (unsigned int j = 0; j < outputDimension; ++j) {
			rowi.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
		}
		m_softmaxWeights.push_back(rowi);
		m_softmaxBias.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
	}

	if (m_updateWordVectors) {
		//params.v;//map<string, vector<PARAM_ID> > v
		// New way: only create parameters for words in training data.
		set<string> trainWords;
		for (unsigned int i = 0; i < trainData.size(); ++i) {
			// Go through all words.
			for (unsigned int j = 0; j < trainData[i].m_words.size(); ++j) {
				trainWords.insert(trainData[i].m_words[j]);
			}
			//trainWords.insert(trainData[i].m_words[1]);
		}
		for (unsigned int i = 0; i < devData.size(); ++i) {
			for (unsigned int j = 0; j < devData[i].m_words.size(); ++j) {
				trainWords.insert(devData[i].m_words[j]);
			}
			//trainWords.insert(devData[i].m_words[1]);
		}
		for (unsigned int i = 0; i < testData.size(); ++i) {
			for (unsigned int j = 0; j < testData[i].m_words.size(); ++j) {
				trainWords.insert(testData[i].m_words[j]);
			}
			//trainWords.insert(devData[i].m_words[1]);
		}
		//trainWords.insert(UNK_STRING);

		for (set<string>::const_iterator iter = trainWords.begin(); iter != trainWords.end(); ++iter) {
			const string& currWord = *iter;
			bool unk = false;
			const vector<double>* currVector = getWordVector(m_wordVectors, currWord, unk);
			vector<PARAM_ID> currVectorParams;
			// Go through all dimensions of the current word vector.
			for (unsigned int i = 0; i < currVector->size(); ++i) {
				// Create a new parameter for the current word vector entry.
				if (initWordVectorsRandomly) {
					currVectorParams.push_back(initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
				} else {
					currVectorParams.push_back(initializeNewParameter(currVector->at(i)));
				}
			}
			m_v.insert(make_pair(currWord, currVectorParams));
		}

		// Go through all word vectors.
		//for (map<string, vector<double> >::const_iterator iter = m_wordVectors.begin(); iter != m_wordVectors.end(); ++iter) {
		//	const string& currWord = iter->first;
		//	const vector<double>& currVector = iter->second;
		//	vector<PARAM_ID> currVectorParams;
		//	// Go through all dimensions of the current word vector.
		//	for (unsigned int i = 0; i < currVector.size(); ++i) {
		//		// Create a new parameter for the current word vector entry.
		//		currVectorParams.push_back(initParam(currVector[i]));
		//	}
		//	params.v.insert(make_pair(currWord, currVectorParams));
		//}
	}
}

// Resets data structures for processing the next example (whether in training or decoding).
void CompositionalConvolutionalSentenceModel::resetForNewExample(WordSequenceDatum& datum) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	bool resetCachedParameterFormulasTrue = true;
	resetFormulaObjects(resetCachedParameterFormulasTrue);

	// With no convolutional layer, the number of ci nodes is the same as the input dimension.
	unsigned int numCiNodes = m_inputDimension;
	if (m_numFilters > 0) numCiNodes = m_numFilters;

	// Clear all data structures containing cached formulas from the previous datum, and set them up for the new datum.
	if (m_ciNodes.size() == numCiNodes) {
		for (unsigned int i = 0; i < m_ciNodes.size(); ++i) {
			m_ciNodes[i] = NULL;
		}
	} else {
		m_ciNodes.clear();
		m_ciNodes.reserve(numCiNodes);
		for (unsigned int i = 0; i < numCiNodes; ++i) {
			m_ciNodes.push_back(NULL);
		}
	}

	// Have to loop thru number of simultaneous sequences in each RNN, each time passing the filter width instead of
	// the number of words in the sequence.
	//unsigned int numCachedNodes = datum.m_words.size();
	int maxSequenceLengthToUse = datum.m_sequenceLength;
	for (unsigned int i = 0; i < m_filters.size(); ++i) {
		if (m_filters[i].m_compositionType == CTYPE_LSTM) {
			for (unsigned int j = 0; j < maxSequenceLengthToUse; ++j) {
				m_filters[i].m_lstmParams.resetForNewExample(j, m_filters[i].m_width);
			}
		} else if (m_filters[i].m_compositionType == CTYPE_GRU) {
			for (unsigned int j = 0; j < maxSequenceLengthToUse; ++j) {
				m_filters[i].m_gruParams.resetForNewExample(j, m_filters[i].m_width);
			}
		}
	}
	precomputeWordVectorFormulas(this, datum, m_inputDimension, m_updateWordVectors);
}

Formula* CompositionalConvolutionalSentenceModel::createLossFDAGForWordSequenceDatum(WordSequenceDatum& datum, int exampleNumber, bool doDropout, double dropoutRate) {
	bool scaleWeightsFalse = false, testTimeFalse = false;
	// Create a term for the correct label's score, which is negated in the loss.
	Formula* negCorrect = getFormulaObject(NEG, 1);
	// Create a sum node to sum over labels.
	Formula* sum = getFormulaObject(PLUS, m_numLabels);
	// Sum over labels.
	for (unsigned int label = 0; label < m_numLabels; ++label) {
		Formula* score = createSoftmaxScoreForLabel(label, datum, doDropout, dropoutRate, scaleWeightsFalse, testTimeFalse);
		// If the current label is the true label, add the score to the negated correct term.
		if (label == datum.m_label) negCorrect->addChild(score);
		// Always add the exponentiated label score to the sum over labels.
		Formula* expScore = getFormulaObject(EXP, 1);
		expScore->addChild(score);
		sum->addChild(expScore);
	}
	// Put a log in front of the sum of exponentiated label scores.
	Formula* logSum = getFormulaObject(LOG, 1);
	logSum->addChild(sum);
	// Add the negated correct score to the log-sum over all exponentiated label scores.
	Formula* negCorrectPlusSum = getFormulaObject(PLUS, 2);
	negCorrectPlusSum->addChild(negCorrect);
	negCorrectPlusSum->addChild(logSum);
	// Return the sum.
	return negCorrectPlusSum;
}

Formula* CompositionalConvolutionalSentenceModel::createSoftmaxScoreForLabel(unsigned int label, WordSequenceDatum& datum,
		bool doDropout, double dropoutRate, bool scaleWeights, bool testTime) {

	// Go through all dimensions of the c vector.
	unsigned int numCiNodes = m_inputDimension;
	bool convolutional = false;
	if (m_numFilters > 0) {
		numCiNodes = m_numFilters;
		convolutional = true;
	}
	Formula* sumOverDimensions = getFormulaObject(PLUS, numCiNodes + 1);
	for (unsigned int i = 0; i < numCiNodes; ++i) {
		bool dropout = false;
		if (doDropout && getRandomDouble() < dropoutRate) dropout = true;
		if (!dropout) {
			Formula* ci = NULL;
			if (convolutional) {
				// Create the Formula node for c_i using the ith convolutional filter.
				ci = createCiNodeConvolutional(i, datum, testTime);
			} else {
				// Create the Formula node for c_i using the neural bag-of-words model of the entire sentence.
				ci = createCiNodeNBOW(i, datum);
			}
			// Multiply it by the weight corresponding to the provided label and the current dimension i.
			Formula* prod = getFormulaObject(TIMES, 2);
			//prod->reserveChildren(2);
			prod->addChild(ci);
			if (scaleWeights) {
				double currWeight = getValue(m_softmaxWeights[label][i]);
				Formula* weightConstant = getConstantFormulaObject(currWeight * dropoutRate);
				prod->addChild(weightConstant);
			} else {
				prod->addChild(getCachedParameterFormula(m_softmaxWeights.at(label).at(i)));
			}
			sumOverDimensions->addChild(prod);
		}
	}
	// Add the softmax bias.
	sumOverDimensions->addChild(getCachedParameterFormula(m_softmaxBias.at(label)));
	return sumOverDimensions;
}

/*Formula* CompositionalConvolutionalSentenceModel::createDiNode(ModelParams& params, unsigned int i, const WordSequenceDatum& datum, bool doDropout, double dropoutThreshold) {
	if (params.diNodes[i] != NULL) return params.diNodes[i];
	Formula* sumOverDimensions = getFormulaObject(PLUS);
	for (unsigned int j = 0; j < params.m_dimOutput; ++j) {
		bool dropout = false;
		if (doDropout) {
			double randDouble = (((double) rand()) / ((double) RAND_MAX));
			if (randDouble < dropoutThreshold) dropout = true;
		}
		if (!dropout) {
			// Create the formula for c_j.
			Formula* cj = createCiNodeSentence(params, j, datum);
			// Multiply it by the h matrix entry corresponding to the given dimension i and the current dimension j.
			Formula* prod = getFormulaObject(TIMES);
			prod->addChild(cj);
			prod->addChild(getCachedParameterFormula(params.h[i][j]));
			sumOverDimensions->addChild(prod);
		}
	}
	//fireFeature(bigplus, params.bias[i], false, false);//, "m_" + stringiComma + itos(j + params.m_dim));
	Formula* activation = getFormulaObject(RELU);
	activation->addChild(sumOverDimensions);
	params.diNodes[i] = activation;
	return activation;
}*/

Formula* CompositionalConvolutionalSentenceModel::createCiNodeConvolutional(unsigned int i, WordSequenceDatum& datum, bool testTime) {
	if (m_ciNodes.at(i) != NULL) return m_ciNodes[i];
	//cerr << "Error: conv" << endl;
	// Node c_i is a max-over-time node for filter i.
	// Get a reference to filter i.
	Filter& filter = m_filters.at(i);
	// Create a max node.
	Formula* poolingNode = NULL;
	int numWords = datum.m_words.size();
	int numChildren = numWords + 1 - filter.m_width;
	if ((filter.m_compositionType == CTYPE_NONE || filter.m_compositionType == CTYPE_LSTM || filter.m_compositionType == CTYPE_GRU) && filter.m_width > numWords) {
		// If it's a simple filter and the filter is wider than the word sequence, return the constant 0.
		poolingNode = getConstantFormulaObject(0.0);
	} else if (m_useStochasticPooling && testTime) {
		poolingNode = getFormulaObject(PROBABILISTIC_AVERAGING, numChildren);
	} else if (m_useStochasticPooling) {
		poolingNode = getFormulaObject(SAMPLE_MULTINOMIAL, numChildren);
	} else {
		poolingNode = getFormulaObject(MAX, numChildren);
	}
	if (datum.m_words.size() != datum.m_vectFormulas.size()) {
		cerr << "Error: datum.m_words.size() = " << datum.m_words.size() << ", datum.m_vectFormulas.size() = " << datum.m_vectFormulas.size() << endl;
		cerr << "sentence: " << datum.toString() << endl;
	}
	//poolingNode->reserveChildren(numWords + 1 - filter.m_width);
	// Go through all possible time positions in the sentence.
	for (int wordIndex = 0; wordIndex < numWords; ++wordIndex) {
		// If it's a non-compositional filter, just apply it.
		if (filter.m_compositionType == CTYPE_NONE) {
			// Only apply the filter if its width fits in the remainder of the sentence.
			if (wordIndex + filter.m_width <= numWords) {
				poolingNode->addChild(createConvolutionalFilterAtWord(filter, datum, wordIndex));
			}
		} else if (filter.m_compositionType == CTYPE_LSTM || filter.m_compositionType == CTYPE_GRU) {
			if (wordIndex + filter.m_width <= numWords) {
				poolingNode->addChild(createCompositionalConvolutionalFilterAtPosition(filter, datum, wordIndex));
			}
		} else {
			// If it's a compositional filter, the length of the filter is dynamic, so do the
			// pooling over all allowable n-grams starting at word j.
			for (int k = 0; k < filter.m_maxWords; ++k) {
				if (wordIndex + k < numWords) {
					poolingNode->addChild(createCompositionalConvolutionalFilterOnWordRange(filter, datum, wordIndex, wordIndex + k));
				}
			}
		}
	}
	//bigplus->addChild(getCachedParameterFormula(params.m_bias[i]));
	//Formula* activation = getFormulaObject(RELU);
	//activation->addChild(maxOverTime);
	m_ciNodes[i] = poolingNode;
	return poolingNode;
}

Formula* CompositionalConvolutionalSentenceModel::createConvolutionalFilterAtWord(Filter& filter, const WordSequenceDatum& datum, int wordIndex) {
	int numWords = datum.m_words.size();
	int wordLength = m_inputDimension;
	Formula* filterSum = getFormulaObject(PLUS, (filter.m_width * wordLength) + 1);
	//Formula* filterSumSquares = NULL;
	//Formula* ngramSumSquares = NULL;
	//if (m_useCosineSimilarityFilters) {
	//	filterSumSquares = getFormulaObject(PLUS);
	//	ngramSumSquares = getFormulaObject(PLUS);
	//}
	//filterSum->reserveChildren((filter.m_width * wordLength) + 1);

	if (filter.m_compositionType == CTYPE_NONE) {
		// Go through the number of words in the filter.
		for (int i = 0; i < filter.m_width; ++i) {
			// If the current desired word position in the sentence is a valid word position,
			if (wordIndex + i < numWords) {
				// Apply the filter to the current word.
				for (int j = 0; j < wordLength; ++j) {
					// Multiply the filter weight by the word parameter.
					Formula* prod = getFormulaObject(TIMES, 2);
					//prod->reserveChildren(2);
					// TODO speed up by avoiding multiplication.
					int filterPosition = (i * wordLength) + j;
					Formula* filterWeightFormula = getCachedParameterFormula(filter.m_weights.at(filterPosition));
					prod->addChild(filterWeightFormula);
					Formula* ngramFormula = datum.m_vectFormulas.at(wordIndex + i).at(j);
					prod->addChild(ngramFormula);

					/*if (m_useCosineSimilarityFilters) {
						Formula* filterSquare = getFormulaObject(TIMES, 2);
						filterSquare->addChild(filterWeightFormula);
						filterSquare->addChild(filterWeightFormula);
						filterSumSquares->addChild(filterSquare);

						Formula* ngramSquare = getFormulaObject(TIMES, 2);
						ngramSquare->addChild(ngramFormula);
						ngramSquare->addChild(ngramFormula);
						ngramSumSquares->addChild(ngramSquare);
					}*/
					filterSum->addChild(prod);
				}
			}
		}
	} else {
		cerr << "Error: called non-compositional filter function on compositional filter!" << endl;
	}
	// filterSum now contains the dot product of the filter and the n-gram.
	/*if (m_useCosineSimilarityFilters) {
		Formula* cosineSimilarity = getFormulaObject(DIVIDE);
		cosineSimilarity->addChild(filterSum);
		Formula* sqrtFilterSumSquares = getFormulaObject(SQRT);
		sqrtFilterSumSquares->addChild(filterSumSquares);
		Formula* sqrtNGramSumSquares = getFormulaObject(SQRT);
		sqrtNGramSumSquares->addChild(ngramSumSquares);
		Formula* normalizationTerms = getFormulaObject(TIMES);
		normalizationTerms->addChild(sqrtFilterSumSquares);
		normalizationTerms->addChild(sqrtNGramSumSquares);
		cosineSimilarity->addChild(normalizationTerms);

		Formula* filterMatchSum = getFormulaObject(PLUS);
		filterMatchSum->addChild(cosineSimilarity);
		filterMatchSum->addChild(getCachedParameterFormula(filter.m_bias));
		Formula* activation = getFormulaObject(LEAKY_RELU, 1);
		activation->addChild(filterMatchSum);
		return activation;
		//return filterMatchSum;
	} else {*/
		filterSum->addChild(getCachedParameterFormula(filter.m_bias));
		Formula* activation = getFormulaObject(LEAKY_RELU, 1);
		activation->addChild(filterSum);
		return activation;
	//}
}

Formula* CompositionalConvolutionalSentenceModel::createCompositionalConvolutionalFilterAtPosition(Filter& filter, WordSequenceDatum& datum, int startIndex) {
	Formula* dotProductSum = NULL;
	//int wordLength = m_inputDimension;
	if (filter.m_compositionType == CTYPE_SUM || filter.m_compositionType == CTYPE_WSUM1 || filter.m_compositionType == CTYPE_WSUM2) {
		cerr << "Error: filter type not yet implemented." << endl;
	} else if (filter.m_compositionType == CTYPE_LSTM) {
		// Create the LSTM hidden vector for the sequence of length filter.m_width that starts at position startIndex.
		// Sum over all dimensions of the final hidden vectors for all LSTMs, as well as the bias weight.
		// We only use the bias weight from the first LSTM.
		DeepLSTM& lstmParams = filter.m_lstmParams;
		int numTermsInSum = 1 + lstmParams.getTotalHiddenVectorSize();
		dotProductSum = getFormulaObject(PLUS, numTermsInSum);

		// Go through all hidden layers for the deep LSTM.
		for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
			// Go through all entries in the current hidden layer for the current deep LSTM.
			for (unsigned int k = 0; k < lstmParams.m_layerWidths[j]; ++k) {
				// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
				int startTime = 0;
				if (lstmParams.m_forward) startTime = (startIndex + filter.m_width) - 1;//datum.m_numWords - 1;
				int startDepth = j;
				// We use the start index as the index for the particular sequence in the LSTM (since each unique sequence-per-datum has to have a unique ID).
				Formula* zj = lstmParams.createHiddenNodeAtTimeAtDepth(*this, datum, startIndex, k, startTime, startDepth, -1.0, -1.0, startIndex);
				// Multiply it by the weight corresponding to the provided label and the current dimension.
				Formula* prod = getFormulaObject(TIMES, 2);
				prod->addChild(zj);
				prod->addChild(getCachedParameterFormula(filter.m_weights.at(k)));
				dotProductSum->addChild(prod);
			}
		}

	} else {
		cerr << "Error: ctype not supported: " << filter.m_compositionType << endl;
	}
	// Add the bias weight for the given filter, regardless of n-gram length.
	dotProductSum->addChild(getCachedParameterFormula(filter.m_bias));
	Formula* activation = getFormulaObject(LEAKY_RELU, 1);
	activation->addChild(dotProductSum);
	return activation;
}

Formula* CompositionalConvolutionalSentenceModel::createCompositionalConvolutionalFilterOnWordRange(Filter& filter, const WordSequenceDatum& datum, int startIndex, int endIndex) {
	Formula* dotProductSum = getFormulaObject(PLUS);
	int wordLength = m_inputDimension;
	if (filter.m_compositionType == CTYPE_SUM || filter.m_compositionType == CTYPE_WSUM1 || filter.m_compositionType == CTYPE_WSUM2) {
		if (startIndex == endIndex) {
			// There is just a single word in the n-gram of interest.
			// Apply the filter to the word directly without any composition.
			for (int filterPosition = 0; filterPosition < wordLength; ++filterPosition) {
				// Multiply the filter weight by the word parameter.
				Formula* prod = getFormulaObject(TIMES, 2);
				prod->addChild(getCachedParameterFormula(filter.m_weights.at(filterPosition)));
				prod->addChild(datum.m_vectFormulas.at(startIndex).at(filterPosition));
				dotProductSum->addChild(prod);
			}
		} else {
			// Here the n-gram is longer than a unigram, so we have to do composition.
			// Create a vector to contain the sum of all word vectors in the n-gram.
			vector<Formula*> ngramSum(wordLength, NULL);
			for (int wordPosition = startIndex; wordPosition <= endIndex; ++wordPosition) {
				// Add the current word to the sum.
				for (int j = 0; j < wordLength; ++j) {
					if (wordPosition == startIndex) {
						// We're on the first word in the n-gram, so we have to initialize the vector
						// entries to PLUS nodes.
						ngramSum[j] = getFormulaObject(PLUS);
					}
					// Add the jth entry in the current word to the jth entry of the n-gram sum vector.
					ngramSum[j]->addChild(datum.m_vectFormulas.at(wordPosition).at(j));
				}
			}
			// Now that we have the n-gram sum, compute the n-gram average.
			vector<Formula*> ngramAverage(wordLength, NULL);
			//double oneOverNumWordsInRange = 1.0 / ((double)((endIndex - startIndex) + 1));
			//Formula* scalingMultiplier = getConstantFormulaObject(oneOverNumWordsInRange);
			Formula* scalingMultiplier = getCachedParameterFormula(filter.m_ngramLengthBiases.at(endIndex - startIndex));
			for (int j = 0; j < wordLength; ++j) {
				ngramAverage[j] = getFormulaObject(TIMES, 2);
				ngramAverage[j]->addChild(scalingMultiplier);
				ngramAverage[j]->addChild(ngramSum[j]);
			}
			// Apply the filter to the average.
			for (int filterPosition = 0; filterPosition < wordLength; ++filterPosition) {
				// Multiply the filter weight by the word parameter.
				Formula* prod = getFormulaObject(TIMES, 2);
				prod->addChild(getCachedParameterFormula(filter.m_weights.at(filterPosition)));
				//prod->addChild(ngramSum.at(filterPosition));
				prod->addChild(ngramAverage.at(filterPosition));
				dotProductSum->addChild(prod);
			}
		}
	} else {
		cerr << "Error: ctype not supported: " << filter.m_compositionType << endl;
	}

	//int ngramLengthBiasIndex = endIndex - startIndex;
	// Add the bias weight for the given n-gram length.
	//dotProductSum->addChild(getCachedParameterFormula(filter.m_ngramLengthBiases.at(ngramLengthBiasIndex)));
	// Add the bias weight for the given filter, regardless of n-gram length.
	dotProductSum->addChild(getCachedParameterFormula(filter.m_bias));
	Formula* activation = getFormulaObject(LEAKY_RELU, 1);
	activation->addChild(dotProductSum);
	return activation;
}

Formula* CompositionalConvolutionalSentenceModel::createCiNodeNBOW(unsigned int i, const WordSequenceDatum& datum) {
	if (m_ciNodes[i] != NULL) return m_ciNodes[i];
	Formula* bigplus = getFormulaObject(PLUS);

	int numWords = datum.m_words.size();
	//if (datum.m_words.size() != datum.m_vectFormulas.size()) {
	//	cerr << "Error: datum.m_words.size() = " << datum.m_words.size() << ", datum.m_vectFormulas.size() = " << datum.m_vectFormulas.size() << endl;
	//	cerr << "sentence: " << datum.toString() << endl;
	//}
	bool useAverage = true;
	// NBoW network:
	if (i < m_inputDimension) {
		for (int k = 0; k < numWords; ++k) {
			bigplus->addChild(datum.m_vectFormulas.at(k).at(i));
		}
	} else {
		cerr << "Error" << endl;
	}
	bool useActivation = false;

	if (useAverage) {
		Formula* avg = getFormulaObject(TIMES, 2);
		avg->addChild(bigplus);
		double doubleOneOverNumWords = 1.0 / ((double)numWords);
		avg->addChild(getConstantFormulaObject(doubleOneOverNumWords));
		Formula* biggerplus = getFormulaObject(PLUS, 2);
		biggerplus->addChild(avg);
		biggerplus->addChild(getCachedParameterFormula(m_nbowBias[i]));

		if (useActivation) {
			Formula* activation = getFormulaObject(LEAKY_RELU, 1);
			activation->addChild(biggerplus);
			m_ciNodes[i] = activation;
			return activation;
		} else {
			m_ciNodes[i] = biggerplus;
			return biggerplus;
		}
	} else {
		bigplus->addChild(getCachedParameterFormula(m_nbowBias[i]));

		if (useActivation) {
			Formula* activation = getFormulaObject(LEAKY_RELU, 1);
			activation->addChild(bigplus);
			m_ciNodes[i] = activation;
			return activation;
		} else {
			m_ciNodes[i] = bigplus;
			return bigplus;
		}
	}

}

void CompositionalConvolutionalSentenceModel::scaleWeightsToMatchL2Norm(double l2NormConstraint) {
	bool considerBias = false;
	// Compute the l2 norm of the relevant weights.
	double sumOfSquaredWeights = 0.0;
	// Go through all weights in filters and final softmax layer.
	for (vector<Filter>::const_iterator iter = m_filters.begin(); iter != m_filters.end(); ++iter) {
		//Filter& filter = *iter;
		const vector<PARAM_ID>& weights = iter->m_weights;
		for (vector<PARAM_ID>::const_iterator iter2 = weights.begin(); iter2 != weights.end(); ++iter2) {
			double currValue = getValue(*iter2);
			sumOfSquaredWeights += (currValue * currValue);
		}
		if (considerBias) {
			double currValue = getValue(iter->m_bias);
			sumOfSquaredWeights += (currValue * currValue);
		}
	}
	int outputDimension = m_inputDimension;
	if (m_numFilters > 0) outputDimension = m_numFilters;
	for (unsigned int i = 0; i < m_numLabels; ++i) {
		for (unsigned int j = 0; j < outputDimension; ++j) {
			double currValue = getValue(m_softmaxWeights.at(i).at(j));
			sumOfSquaredWeights += (currValue * currValue);
		}
		if (considerBias) {
			double currValue = getValue(m_softmaxBias.at(i));
			sumOfSquaredWeights += (currValue * currValue);
		}
		//double currValue = getValue(params.m_softmaxWeights.at(i).at(j));
		//sumOfSquaredWeights += (currValue * currValue);
		//params.m_softmaxBias.push_back(initializeNewParameter(getRandomDouble() - 0.5));
	}

	double l2norm = sqrt(sumOfSquaredWeights);
	if (l2norm > l2NormConstraint) {
		double scalingFactor = l2NormConstraint / l2norm;

		sumOfSquaredWeights = 0.0;
		// Go through all weights in filters and final softmax layer and scale by the scaling factor.
		for (vector<Filter>::const_iterator iter = m_filters.begin(); iter != m_filters.end(); ++iter) {
			const vector<PARAM_ID>& weights = iter->m_weights;
			for (vector<PARAM_ID>::const_iterator iter2 = weights.begin(); iter2 != weights.end(); ++iter2) {
				double currValue = getValue(*iter2);
				double newValue = currValue * scalingFactor;
				setValue(*iter2, newValue);
				sumOfSquaredWeights += (newValue * newValue);
			}
			if (considerBias) {
				double currValue = getValue(iter->m_bias);
				double newValue = currValue * scalingFactor;
				setValue(iter->m_bias, newValue);
				sumOfSquaredWeights += (newValue * newValue);
			}
		}
		for (unsigned int i = 0; i < m_numLabels; ++i) {
			for (unsigned int j = 0; j < outputDimension; ++j) {
				PARAM_ID currParamID = m_softmaxWeights.at(i).at(j);
				double currValue = getValue(currParamID);
				double newValue = currValue * scalingFactor;
				setValue(currParamID, newValue);
				sumOfSquaredWeights += (newValue * newValue);
			}
			if (considerBias) {
				PARAM_ID currParamID = m_softmaxBias.at(i);
				double currValue = getValue(currParamID);
				double newValue = currValue * scalingFactor;
				setValue(currParamID, newValue);
				sumOfSquaredWeights += (newValue * newValue);
			}
		}
		double l2normNew = sqrt(sumOfSquaredWeights);
		cerr << "Original l2 norm:\t" << l2norm << "\t\tnew:\t" << l2normNew << endl;
	}
}

double CompositionalConvolutionalSentenceModel::doDecoding(vector<WordSequenceDatum>& data, bool scaleWeights, double dropoutRate) {
	int numCorrect = 0;
	int total = 0;
	for (unsigned int i = 0; i < data.size(); ++i) {
		WordSequenceDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(datum);
		bool correct = decode(datum, scaleWeights, dropoutRate, i);
		datum.doneProcessing();
		if (correct) ++numCorrect;
		++total;
	}
	return (((double)numCorrect) / ((double)total));
}

bool CompositionalConvolutionalSentenceModel::decode(WordSequenceDatum& datum, bool scaleWeights, double dropoutRate, int exampleNumber) {
	bool doDropoutFalse = false;
	bool testTimeTrue = true;
	double maxScore = neg_inf;
	int maxLabel = -1;
	for (unsigned int i = 0; i < m_numLabels; ++i) {
		Formula* score = createSoftmaxScoreForLabel(i, datum, doDropoutFalse, dropoutRate, scaleWeights, testTimeTrue);
		double currScore = score->evaluate();
		if (currScore > maxScore) {
			maxScore = currScore;
			maxLabel = i;
		}
	}
	if (maxLabel == -1) {
		cerr << "Error: no label found with score > negative infinity! on line #" << exampleNumber << endl;
	} else {
		if (maxLabel == datum.m_label) return true;
	}
	return false;
}

void CompositionalConvolutionalSentenceModel::printFilterNearestNeighbors(vector<WordSequenceDatum>& trainData) {
	// TODO also try with all words in word vector vocabularies.
	// Create a map of word vectors for the training data words.
	/*map<string, vector<double> > updatedWordVectors;
	if (m_updateWordVectors) {
		set<string> trainWords;
		// First get all words in sentences in trainData.
		for (unsigned int j = 0; j < trainData.size(); ++j) {
			WordSequenceDatum& datum = trainData[j];
			for (unsigned int i = 0; i < datum.m_words.size(); ++i) {
				trainWords.insert(datum.m_words[i]);
			}
		}
		for (set<string>::const_iterator iter = trainWords.begin(); iter != trainWords.end(); ++iter) {
			bool unk = false;
			const vector<PARAM_ID>* currVectorParams = getWordVectorParams(params.v, *iter, unk);
			vector<double> wordVector;
			for (unsigned int i = 0; i < currVectorParams->size(); ++i) {
				wordVector.push_back(getValue(currVectorParams->at(i)));
			}
			updatedWordVectors.insert(make_pair(*iter, wordVector));
		}
	}*/

	/*if (m_updateWordVectors) {
		int wordLength = params.m_inputDimension;
		unsigned int numNeighbors = 10;
		cerr << "Printing nearest neighbors to filters:" << endl;
		// Go through all filters.
		for (unsigned int i = 0; i < params.m_filters.size(); ++i) {
			// Go through all words in the current filter.
			Filter& filter = params.m_filters.at(i);
			cerr << "For " << filter.m_width << "-gram filter " << i << ":" << endl;
			vector<vector<string> > wordNeighbors;
			for (unsigned int j = 0; j < numNeighbors; ++j) {
				vector<string> emptyVector;
				wordNeighbors.push_back(emptyVector);
			}
			// Go through the number of words in the filter.
			for (unsigned int j = 0; j < filter.m_width; ++j) {
				// Get the vector for the current word in the current filter.
				vector<double> currentFilterWord;
				for (int k = 0; k < wordLength; ++k) {
					int filterPosition = (j * wordLength) + k;
					currentFilterWord.push_back(getValue(filter.m_weights.at(filterPosition)));
				}

				// Compute the nearest word to the current word in the current filter.
				cerr << "Neighbors of filter word " << j << ":" << endl;
				//cerr << "For unigram \"" << (*iter) << "\", 10-nearest neighbors from training data words:" << endl;
				priority_queue<pair<string, double>, vector<pair<string, double> >, CompareStringDoublePairPriorityQueue> pq;
				for (map<string, vector<double> >::const_iterator iter2 = updatedWordVectors.begin(); iter2 != updatedWordVectors.end(); ++iter2) {
					double cosineSim = cosineSimilarity(currentFilterWord, iter2->second);
					pq.push(make_pair(iter2->first, cosineSim));
				}
				for (unsigned int k = 0; k < numNeighbors; ++k) {
					string currWord = pq.top().first;
					//double cosineSim = pq.top().second;
					//cerr << cosineSim << ":\t" << currWord << endl;
					wordNeighbors[k].push_back(currWord);
					pq.pop();
				}
			}
			// Print the block of words.
			for (unsigned int j = 0; j < numNeighbors; ++j) {
				cerr << "\t";
				for (unsigned int k = 0; k < wordNeighbors[k].size(); ++k) {
					if (k > 0) cerr << " ";
					cerr << wordNeighbors[j][k];
				}
				cerr << endl;
			}
			cerr << endl;
		}
	}*/

	// Go through the training data and find the k highest-scoring filter matches for each filter.
	//int wordLength = m_inputDimension;
	unsigned int numMatches = 25;
	cerr << "Printing top matches for filters:" << endl;
	map<pair<int, double>, vector<string>, CompareIntDoublePairMap > filtersToTopMatches;
	// Go through all filters.
	for (unsigned int i = 0; i < m_filters.size(); ++i) {
		// Get the current filter.
		Filter& filter = m_filters.at(i);
		// Create a priority queue
		//priority_queue<pair<string, double>, vector<pair<string, double> >, CompareStringDoublePairPriorityQueue> pq;
		set<pair<string, double>, CompareStringDoublePairSet > pq;
		// Go through all sentences in trainData.
		for (unsigned int j = 0; j < trainData.size(); ++j) {
			WordSequenceDatum& datum = trainData[j];
			//cerr << "processing " << j << ": " << datum.toString() << endl;
			// We are processing a new example, so reset Formula data structures in the model.
			resetForNewExample(datum);

			// Go through words in current sentence.
			int numWords = datum.m_words.size();

			for (int wordIndex = 0; wordIndex < numWords; ++wordIndex) {
				//cerr << "Processing word " << wordIndex << endl;
				// If it's a non-compositional filter, just apply it.
				if (filter.m_compositionType == CTYPE_NONE) {
					// Only apply the filter if its width fits in the remainder of the sentence.
					if (wordIndex + filter.m_width <= numWords) {
						Formula* filterSumFormula = createConvolutionalFilterAtWord(filter, datum, wordIndex);
						double filterSum = filterSumFormula->evaluate();

						string sentenceWords = datum.getWordsInRangePadded(wordIndex, (wordIndex + filter.m_width) - 1);
						if (filterSum != filterSum || filterSum == neg_inf || filterSum == pos_inf) {
							cerr << "for filter " << i << ", example " << j << ", score = " <<
									filterSum << ", words = " << sentenceWords << endl;
						} else {
							pq.insert(make_pair(sentenceWords, filterSum));
						}
					}
				} else if (filter.m_compositionType == CTYPE_LSTM || filter.m_compositionType == CTYPE_GRU) {
					// Only apply the filter if its width fits in the remainder of the sentence.
					if (wordIndex + filter.m_width <= numWords) {
						Formula* filterSumFormula = createCompositionalConvolutionalFilterAtPosition(filter, datum, wordIndex);
						double filterSum = filterSumFormula->evaluate();

						string sentenceWords = datum.getWordsInRangePadded(wordIndex, (wordIndex + filter.m_width) - 1);
						if (filterSum != filterSum || filterSum == neg_inf || filterSum == pos_inf) {
							cerr << "for filter " << i << ", example " << j << ", score = " <<
									filterSum << ", words = " << sentenceWords << endl;
						} else {
							pq.insert(make_pair(sentenceWords, filterSum));
						}
					}
				} else {
					// If it's a compositional filter, the length of the filter is dynamic, so compute
					// filter matches over all allowable n-grams starting at word j.
					for (int k = 0; k < filter.m_maxWords; ++k) {
						if (wordIndex + k < numWords) {
							Formula* filterSumFormula = createCompositionalConvolutionalFilterOnWordRange(filter, datum, wordIndex, wordIndex + k);
							double filterSum = filterSumFormula->evaluate();
							string sentenceWords = datum.getWordsInRangePadded(wordIndex, wordIndex + k);
							pq.insert(make_pair(sentenceWords, filterSum));
						}
					}
				}
			}

			//for (int wordIndex = 0; wordIndex < numWords; ++wordIndex) {
				// Compute the filter's match score starting at the current word.
				//double filterSum = 0.0;
				//string sentenceWords = "sent. " + itos(j) + ":";
				//string sentenceWords = "";
				// Go through the number of words in the filter.
				//for (int filterWordIndex = 0; filterWordIndex < filter.m_width; ++filterWordIndex) {
				//	if (wordIndex + filterWordIndex < numWords) {
						// Apply the filter to the current word.
				//		for (int jj = 0; jj < wordLength; ++jj) {
							// Multiply the filter weight by the word parameter.
				//			int filterPosition = (filterWordIndex * wordLength) + jj;
				//			double prod = 0.0;
				//			if (m_updateWordVectors) {
				//				prod = getValue(filter.m_weights.at(filterPosition))
				//					* getValue(datum.m_vectParams.at(wordIndex + filterWordIndex)->at(jj));
				//			} else {
				//				prod = getValue(filter.m_weights.at(filterPosition))
				//					* (datum.m_vects.at(wordIndex + filterWordIndex)->at(jj));
				//				//* (datum.m_vectFormulas.at(wordIndex + filterWordIndex).at(jj))->evaluate();
				//			}

							//prod->addChild(datum.m_vectFormulas.at(wordIndex + i).at(j));
				//			filterSum += prod;
				//		}
				//		// Append the current word to the word sequence.
				//		sentenceWords += " " + datum.m_words[wordIndex + filterWordIndex];
				//	}
				//}
				//pq.insert(make_pair(sentenceWords, filterSum));
			//}
		}
		vector<string> topMatches;
		cerr << "For " << filter.m_width << "-gram filter " << i << " (";
		double weightLabel0 = getValue(m_softmaxWeights[0][i]);
		double weightLabel1 = getValue(m_softmaxWeights[1][i]);
		double positivity = weightLabel1 - weightLabel0;
		cerr << weightLabel0 << ", " << weightLabel1 << "; positivity: " << positivity << "):" << endl;
		for (unsigned int k = 0; k < numMatches && pq.size() > 0; ++k) {
			//string currWordSequence = pq.top().first;
			//double filterSum = pq.top().second;
			string currWordSequence = pq.begin()->first;
			double filterSum = pq.begin()->second;
			cerr << filterSum << ":\t" << currWordSequence << endl;
			topMatches.push_back(currWordSequence);
			pq.erase(pq.begin());
			//pq.pop();
		}
		filtersToTopMatches.insert(make_pair(make_pair(i, positivity), topMatches));
	}
	for (map<pair<int, double>, vector<string>, CompareIntDoublePairMap >::const_iterator iter = filtersToTopMatches.begin(); iter != filtersToTopMatches.end(); ++iter) {
		int filterIndex = iter->first.first;
		double positivity = iter->first.second;
		const vector<string>& topMatches = iter->second;
		cerr << "Matches for " << m_filters[filterIndex].m_width << "-gram filter "
			 << filterIndex << " (positivity: " << positivity << "), ";
		// Print biases for different n-gram lengths:
		vector<PARAM_ID>& biases = m_filters[filterIndex].m_ngramLengthBiases;
		cerr << "(biases: ";
		for (unsigned int biasIndex = 0; biasIndex < biases.size(); ++biasIndex) {
			if (biasIndex > 0) cerr << ", ";
			cerr << getValue(biases[biasIndex]);
		}
		cerr << "):" << endl;
		for (vector<string>::const_iterator iter = topMatches.begin(); iter != topMatches.end(); ++iter) {
			cerr << (*iter) << endl;
		}
		cerr << endl;
	}

}

void CompositionalConvolutionalSentenceModel::printNearestNeighbors(vector<WordSequenceDatum>& trainData, vector<WordSequenceDatum>& devData) {
	set<string> trainWords;
	vector<vector<double> > trainBigramVectors;
	// First compute c vectors for all instances in trainData.
	for (unsigned int j = 0; j < trainData.size(); ++j) {
		WordSequenceDatum& datum = trainData[j];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(datum);

		if (m_updateWordVectors) {
			for (unsigned int i = 0; i < datum.m_words.size(); ++i) {
				trainWords.insert(datum.m_words[i]);
			}
			//trainWords.insert(datum.m_words[1]);
		}

		vector<double> bigramVector;
		// Go through all dimensions of the c vector.
		for (unsigned int i = 0; i < m_inputDimension; ++i) {
			// Create the formula for c_i.
			Formula* ci = createCiNodeNBOW(i, datum);
			double ciValue = ci->evaluate();
			bigramVector.push_back(ciValue);
		}
		trainBigramVectors.push_back(bigramVector);
	}
	// Create a map of word vectors for the training data words.
	map<string, vector<double> > updatedWordVectors;
	if (m_updateWordVectors) {
		for (set<string>::const_iterator iter = trainWords.begin(); iter != trainWords.end(); ++iter) {
			bool unk = false;
			const vector<PARAM_ID>* currVectorParams = getWordVectorParams(m_v, *iter, unk);
			vector<double> wordVector;
			for (unsigned int i = 0; i < currVectorParams->size(); ++i) {
				wordVector.push_back(getValue(currVectorParams->at(i)));
			}
			updatedWordVectors.insert(make_pair(*iter, wordVector));
		}
	}

	set<string> devWords;

	// Go through the dev data.
	for (unsigned int j = 0; j < devData.size(); ++j) {
		WordSequenceDatum& datum = devData[j];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(datum);

		if (m_updateWordVectors) {
			for (unsigned int i = 0; i < datum.m_words.size(); ++i) {
				devWords.insert(datum.m_words[i]);
			}
			//devWords.insert(datum.m_words[1]);
		}

		vector<double> bigramVector;
		// Go through all dimensions of the c vector.
		for (unsigned int i = 0; i < m_inputDimension; ++i) {
			// Create the formula for c_i.
			Formula* ci = createCiNodeNBOW(i, datum);
			double ciValue = ci->evaluate();
			bigramVector.push_back(ciValue);
		}
		cerr << "For sentence \"" << datum.toString() << "\", 10-nearest neighbors from training data:" << endl;
		priority_queue<pair<int, double>, vector<pair<int, double> >, CompareIntDoublePairPriorityQueue> pq;
		for (unsigned int k = 0; k < trainBigramVectors.size(); ++k) {
			double cosineSim = dotProduct(bigramVector, trainBigramVectors[k]);
			pq.push(make_pair(k, cosineSim));
		}
		int numNeighbors = 10;
		for (unsigned int k = 0; k < numNeighbors; ++k) {
			int curr = pq.top().first;
			double cosineSim = pq.top().second;
			cerr << cosineSim << ":\t" << trainData[curr].toString() << endl;
			pq.pop();
		}
		// Go through all train bigram vectors and compute cosine similarity to each.
		//trainBigramToVector.insert(make_pair(j, bigramVector));
	}

	if (m_updateWordVectors) {
		cerr << "Printing nearest neighbors for words in dev data now." << endl;
		// Compute nearest neighbors to word vectors for words in dev data.
		// Create a map of word vectors for the dev data words.
		//map<string, vector<double> > devWordVectors;
		for (set<string>::const_iterator iter = devWords.begin(); iter != devWords.end(); ++iter) {
			bool unk = false;
			const vector<PARAM_ID>* currVectorParams = getWordVectorParams(m_v, *iter, unk);
			if (unk) {
				cerr << "Skipping " << UNK_STRING << " for " << (*iter) << endl;
			} else {
				vector<double> wordVector;
				for (unsigned int i = 0; i < currVectorParams->size(); ++i) {
					wordVector.push_back(getValue(currVectorParams->at(i)));
				}
				//devWordVectors.insert(make_pair(*iter, wordVector));
				cerr << "For unigram \"" << (*iter) << "\", 10-nearest neighbors from training data words:" << endl;
				priority_queue<pair<string, double>, vector<pair<string, double> >, CompareStringDoublePairPriorityQueue> pq;
				for (map<string, vector<double> >::const_iterator iter2 = updatedWordVectors.begin(); iter2 != updatedWordVectors.end(); ++iter2) {
					double cosineSim = dotProduct(wordVector, iter2->second);
					pq.push(make_pair(iter2->first, cosineSim));
				}
				int numNeighbors = 10;
				for (unsigned int k = 0; k < numNeighbors; ++k) {
					string currWord = pq.top().first;
					double cosineSim = pq.top().second;
					cerr << cosineSim << ":\t" << currWord << endl;
					pq.pop();
				}
			}
		}
		cerr << "Done printing nearest neighbors for words in dev data." << endl;
	}
}


