#ifndef COMPOSITIONAL_CONVOLUTIONAL_SENTENCE_MODEL_H_
#define COMPOSITIONAL_CONVOLUTIONAL_SENTENCE_MODEL_H_

#include <fstream>
#include <iostream>
#include <vector>
#include <string>
#include <set>
#include <queue>
#include <list>
#include "../../fdag/model.h"
#include "../classifier_utils.h"
#include "../rnn/lstm.h"
#include "../rnn/gru.h"

using namespace std;

enum FILTER_CTYPE { CTYPE_NONE, CTYPE_SUM, CTYPE_WSUM1, CTYPE_WSUM2, CTYPE_MVP, CTYPE_LSTM, CTYPE_GRU };

struct Filter {
	// The width of the filter is the number of words it considers in the sentence.
	int m_width;
	// If original word vectors are k-dimensional, weight vector will have m_width * k entries.
	vector<PARAM_ID> m_weights;
	// The bias weight is a scalar.
	PARAM_ID m_bias;
	// The composition type of the filter. If CTYPE_NONE, it's not a composition filter.
	FILTER_CTYPE m_compositionType;
	// The maximum number of words that should be considered in this composition filter.
	int m_maxWords;
	// For CTYPE_WSUM1, we need additional weights per word in the weighted average:
	// k additional weights per k-gram supported.
	vector<PARAM_ID> m_ngramLengthBiases;
	DeepLSTM m_lstmParams;
	DeepGRU m_gruParams;
};

struct CompareIntDoublePairPriorityQueue {
	bool operator() (const pair<int, double>& a, const pair<int, double>& b) const {
		return a.second < b.second;
	}
};

struct CompareStringDoublePairPriorityQueue {
	bool operator() (const pair<string, double>& a, const pair<string, double>& b) const {
		return a.second < b.second;
	}
};

struct CompareStringDoublePairSet {
	bool operator() (const pair<string, double>& a, const pair<string, double>& b) const {
		return a.second > b.second;
	}
};

struct CompareIntDoublePairMap {
	bool operator() (const pair<int, double>& a, const pair<int, double>& b) const {
		return a.second > b.second;
	}
};

class CompositionalConvolutionalSentenceModel : public Model {
public:
	//************************************** model.cpp:
	CompositionalConvolutionalSentenceModel(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity);
	~CompositionalConvolutionalSentenceModel();
	double trainAndTest(const vector<string>& trainLines, const vector<string>& devLines, const vector<string>& testLines,
		const string& wordVectorFilename, const string& filters, const string& algorithm, double stepsize, int minibatchSize, int numEpochs,
		int decodeEveryK, double dropoutRate,
		bool updateWordVectors, bool initWordVectorsRandomly, bool stochasticPooling, double l2NormConstraint,
		bool useCosineSimilarityFilters, bool doFiniteDifferenceChecking);

private:
	void loadWordVectors(const vector<string>& trainLines, const vector<string>& devLines, const vector<string>& testLines, const string& wordVectorFilename);

	void initializeAllParameters(const vector<WordSequenceDatum>& trainData, const vector<WordSequenceDatum>& devData,
		const vector<WordSequenceDatum>& testData, bool initWordVectorsRandomly, const string& filters);
	void resetForNewExample(WordSequenceDatum& datum);
	// Precomputes the Formula's needed for the words in the datum, where v holds raw word vectors
	// for the words and vp holds the associated parameter IDs (only used if word vectors are being updated).
	// The result is that vf holds pointers to either constant Formula nodes or parameter Formula nodes, depending
	// on whether word vectors are being updated.
	//void precomputeGivenWordVectorFormulas(const vector<const vector<double>* >& v,
	//	const vector<const vector<PARAM_ID>* >& vp, unsigned int dim, bool updateWordVectors, unsigned int numWords,
	//	vector< vector< Formula* > >& vf);
	//void precomputeGivenWordVectorFormulas(const vector<const vector<double>* >& v, const vector<const vector<PARAM_ID>* >& vp,
	//	unsigned int dim, unsigned int numWords, vector< vector< Formula* > >& vf);
	//void precomputeWordVectorFormulas(WordSequenceDatum& datum, unsigned int inputDimension, bool updateWordVectors);
	Formula* createLossFDAGForWordSequenceDatum(WordSequenceDatum& datum, int exampleNumber, bool doDropout, double dropoutRate);

	Formula* createSoftmaxScoreForLabel(unsigned int label, WordSequenceDatum& datum, bool doDropout, double dropoutRate, bool scaleWeights, bool testTime);

	Formula* createCiNodeConvolutional(unsigned int i, WordSequenceDatum& datum, bool testTime);
	Formula* createConvolutionalFilterAtWord(Filter& filter, const WordSequenceDatum& datum, int wordIndex);
	Formula* createCompositionalConvolutionalFilterAtPosition(Filter& filter, WordSequenceDatum& datum, int startIndex);
	Formula* createCompositionalConvolutionalFilterOnWordRange(Filter& filter, const WordSequenceDatum& datum, int startIndex, int endIndex);

	Formula* createCiNodeNBOW(unsigned int i, const WordSequenceDatum& datum);

	void scaleWeightsToMatchL2Norm(double l2NormConstraint);

	double doDecoding(vector<WordSequenceDatum>& data, bool scaleWeights, double dropoutRate);
	bool decode(WordSequenceDatum& datum, bool scaleWeights, double dropoutRate, int exampleNumber);

	void printFilterNearestNeighbors(vector<WordSequenceDatum>& trainData);
	void printNearestNeighbors(vector<WordSequenceDatum>& trainData, vector<WordSequenceDatum>& devData);

	//void printUnknownWordCounts(const map<string, int>& unkWordCounts);
	//void readWordVectorFile(const string& filename, const set<string>& vocabulary);
	//const vector<double>* getWordVector(const string& word, bool& unk) const;
	//const vector<PARAM_ID>* getWordVectorParams(const map<string, vector<PARAM_ID> >& v, const string& word, bool& unk) const;

	map<string, vector<double> > m_wordVectors;
	//int m_wordVectorDimensionality;
	vector<double>* m_unknownWordVector;
	map<string, vector<PARAM_ID> > m_v;	// Word vector parameters, initialized to word vector values from file.

	// The input dimension is the size of the given word vectors.
	int m_inputDimension;
	// The number of class labels.
	unsigned int m_numLabels;
	// The number of filters in the CNN:
	unsigned int m_numFilters;

	vector<PARAM_ID> m_nbowBias;
	vector<vector<PARAM_ID> > m_softmaxWeights;
	vector<PARAM_ID> m_softmaxBias;

	vector<Filter> m_filters;

	vector<Formula*> m_ciNodes;	// Pointers to Formulas for c_i nodes, which are NULL if unset (or dropped out).

	bool m_updateWordVectors;
	bool m_useStochasticPooling;
	bool m_useCosineSimilarityFilters;
	int m_maxSequenceLength;
};

#endif
