#include <iostream>
//#include <stdlib.h>
#include <sys/time.h>
#include "../../utils/cmdparams.h"
#include "../../utils/utils.h"
#include "cnn.h"

int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("Demonstrates a compositional convolutional neural network for sentences using fdag; "
    		"uses AdaGrad to minimize cross-entropy for a binary classification task.");

    params.addConstraint("train_file",
    		"file containing training data, where each line contains the sentence followed by a tab then the "
    		"label (0 or 1)", false); // not optional
    params.addConstraint("dev_file",
    		"file containing dev data to compute accuracy for, same format as train_file", false); // not optional
    params.addConstraint("test_file",
    		"file containing test data to compute accuracy for, same format as train_file", false); // not optional
    params.addConstraint("cv",
    		"if > 0, will do CV on train_file instead of testing on dev/test; this is the number of folds to use with CV (default: 0)", true); // optional
    params.addConstraint("cv_only_do_fold",
    		"if >= 0 and cv > 0, will only do the given CV fold; this can range from 0 to cv - 1 (default: -1)", true); // optional

    params.addConstraint("word_vector_file",
    		"file containing word vectors, where each line contains the word followed by a tab then the "
    		"space-separated vector entries", false); // not optional
    params.addConstraint("update_word_vectors",
    		"if true, will learn update given word vectors during learning (default: true)", true); // optional
    params.addConstraint("init_word_vectors_randomly",
    		"if true, will initialize word vectors randomly before learning (default: false)", true); // optional

    params.addConstraint("filters",
    		"string indicating convolutional filters to use, in format \"width:number:ctype,width:number:ctype,...\".\n"
    		"where \"ctype\" indicates the composition type of the filter, or can be omitted if no composition is desired.\n"
    		"For example, \"3:20,4:50,5:100\" will use 20 trigram filters, 50 4-gram filters, and 100 5-gram filters.\n"
    		"Supported composition types are:\n"
    		"\t\"n\" - none (just n-gram matching),\n"
    		"\t\"s\" - sum words in n-gram,\n"
    		"\t\"w1\" - weighted sum 1,\n"
    		"\t\"w2\" - weighted sum 2,\n"
    		"\t\"lfX\" - forward LSTM RNN, where X is the size of the hidden vector\n"
    		"\t\"lbX\" - backward LSTM RNN,\n"
    		"\t\"gfX\" - forward GRU RNN, and\n"
    		"\t\"gbX\" - backward GRU RNN."
    		"When using something other than \"n\", the \"width\" becomes the max n-gram length to consider when doing composition.\n"
    		"For example, \"3:20:n,3:50:s,5:10:lf10\" will use 20 trigram filters, 50 trigram sum composition filters "
    		"where each filter has the size of a single word and, when matching a trigram, the filter matches the sum "
    		"of the three words in the trigram, and 10 forward LSTM filters that match sequences of length 5 using a hidden vector of dimension 10.\n"
    		"Leave blank or pass \"0\" if no filters are desired (default: 3:100)", true); // optional
    params.addConstraint("stochastic_pooling",
    		"if true, will do stochastic pooling instead of max pooling (default: false)", true); // optional
    params.addConstraint("l2_norm_constraint",
    		"the s value representing desired l2 norm of weights; after updating, will scale weights such that their norm equals"
    		"this value. pass a value <=0 if no constraining is desired (default: -1.0)", true); // optional

    params.addConstraint("algorithm",
    		"optimization algorithm; can be \"sgd_momentum\", \"adagrad\", or \"adadelta\" (case insensitive) (default: adadelta)", true); // optional
    params.addConstraint("stepsize",
    		"stepsize for AdaGrad (default: 0.1)", true); // optional
    params.addConstraint("minibatch_size",
    		"mini-batch size for AdaGrad (default: 10)", true); // optional
    params.addConstraint("num_epochs",
    		"number of epochs of learning (default: 3)", true); // optional
    params.addConstraint("dropout_rate",
    		"probability of setting an output unit to zero; use a value <= 0 for no dropout (default: 0.5)", true); // optional
    params.addConstraint("cosine_sim_filters",
    		"if true, will use cosine similarity filters instead of dot product (default: false)", true); // optional
    params.addConstraint("decode_every_k", "number of examples to decode after processing, if train data is bigger than 11k examples (default: 5000)", true); // optional

   	if (!params.runConstraints("run_cnn")) return 0;

	// Read in command-line parameters.
	string trainFilename = params.asString("train_file");
	string devFilename = params.asString("dev_file");
	string testFilename = params.asString("test_file");
	int cvNumFolds = atoi(params.asString("cv", "0").c_str());
	int cvOnlyDoFold = atoi(params.asString("cv_only_do_fold", "-1").c_str());
	string wordVectorFilename = params.asString("word_vector_file");
	bool updateWordVectors = (params.asString("update_word_vectors", "true") == "true");
	bool initWordVectorsRandomly = (params.asString("init_word_vectors_randomly", "false") == "true");
	string filters = params.asString("filters", "3:100");
	bool stochasticPooling = (params.asString("stochastic_pooling", "false") == "true");
	//int outputDimension = atoi(params.asString("output_dimension", "10").c_str());
	double l2NormConstraint = atof(params.asString("l2_norm_constraint", "-1.0").c_str());
	string algorithm = params.asString("algorithm", "adadelta");
	double stepsize = atof(params.asString("stepsize", "0.1").c_str());
	int minibatchSize = atoi(params.asString("minibatch_size", "10").c_str());
	int numEpochs = atoi(params.asString("num_epochs", "3").c_str());
	double dropoutRate = atof(params.asString("dropout_rate", "0.5").c_str());
	bool useCosineSimilarityFilters = (params.asString("cosine_sim_filters", "false") == "true");
	int decodeEveryK = atoi(params.asString("decode_every_k", "5000").c_str());

	// Change following to true to do finite difference checking of gradients.
	bool doFiniteDifferenceChecking = false;
	srand(1);

	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);

	if (cvNumFolds <= 0) {
		vector<string> trainLines, devLines, testLines;
		readLinesFromFile(trainFilename, trainLines);
		readLinesFromFile(devFilename, devLines);
		readLinesFromFile(testFilename, testLines);

		CompositionalConvolutionalSentenceModel csm(10000, 2000000, 10000, 60000);
		double accuracy = csm.trainAndTest(trainLines, devLines, testLines, wordVectorFilename, filters, algorithm, stepsize, minibatchSize, numEpochs, decodeEveryK, dropoutRate,
			updateWordVectors, initWordVectorsRandomly, stochasticPooling, l2NormConstraint, useCosineSimilarityFilters, doFiniteDifferenceChecking);
		cerr << "Final accuracy: " << accuracy << endl;
	} else {
		// Load data.
		vector<string> data;
		readLinesFromFile(trainFilename, data);
		//int inputDimension = loadData(dataFile, data, &numLabels);
		// Do n-fold cross-validation.
		double accuracySum = 0.0;
		unsigned int dataSize = data.size();
		int foldSize = dataSize / cvNumFolds;
		vector<double> accuracies;
		int cvStartingFoldIndex = 0;
		int cvEndingFoldIndex = cvNumFolds;
		if (cvOnlyDoFold >= 0) {
			cerr << "Only doing CV fold " << cvOnlyDoFold << " instead of full " << cvNumFolds << "-fold CV." << endl;
			cvStartingFoldIndex = cvOnlyDoFold;
			cvEndingFoldIndex = cvOnlyDoFold + 1;
		}
		for (int i = cvStartingFoldIndex; i < cvEndingFoldIndex; ++i) {
			cerr << "For test fold " << i << ", creating test set with examples " << (i*foldSize) << " to " << (((i+1)*foldSize) - 1) << endl;
			// Generate train/test sets for the current fold (approximately based on discretized index cutoffs).
			vector<string> trainData, devData, testData;
			for (unsigned int j = 0; j < dataSize; ++j) {
				if (((int)j) >= (i*foldSize) && ((int)j) < (i+1)*foldSize) {
					// If the current datum falls within the indices of the current test fold, add it to testData.
					testData.push_back(data[j]);
					//cerr << "test index: " << j << endl;
				} else {
					// Otherwise, add 90% of it to the training set and 10% to the dev set
					if (j % 10 == 0) {
						devData.push_back(data[j]);
						//cerr << "dev index: " << j << endl;
					} else {
						trainData.push_back(data[j]);
						//cerr << "train index: " << j << endl;
					}
				}
			}
			cerr << "Size of training set: " << trainData.size() << endl;
			cerr << "Size of dev set: " << devData.size() << endl;
			cerr << "Size of test set: " << testData.size() << endl;
			// Create a model for the current split.
			CompositionalConvolutionalSentenceModel csm(10000, 2000000, 10000, 60000);
			//ConvolutionalSentenceModel csm;
			// Train and test.
			double currAccuracy = csm.trainAndTest(trainData, devData, testData, wordVectorFilename, filters, algorithm, stepsize, minibatchSize, numEpochs, decodeEveryK, dropoutRate,
				updateWordVectors, initWordVectorsRandomly, stochasticPooling, l2NormConstraint, useCosineSimilarityFilters, doFiniteDifferenceChecking);

			accuracies.push_back(currAccuracy);
			accuracySum += currAccuracy;
		}
		double accuracyMean = accuracySum / ((double) cvNumFolds);
		cerr << "CV mean accuracy: \t" << accuracyMean << endl;
		cerr << "CV accuracy stddev: \t" << standardDeviation(accuracies) << endl;
	}
    //params.addConstraint("layer_widths", "widths of hidden layers, comma-delimited. e.g., \"10,10,5\", or use \"0\" for no hidden layers (default: 10,5)", true); // optional
	//string layerWidths = params.asString("layer_widths", "10,5");

	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
    return 0;
}
