#include <iostream>
#include <stdlib.h>
#include <sys/time.h>
#include "paraphrase_model.h"

// public:
ParaphraseModel::ParaphraseModel(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity) : Model(unaryFormulaCapacity, binaryFormulaCapacity, fertileFormulaCapacity, barrenFormulaCapacity) {
	m_updateWordVectors = true;
	m_wordVectorDimensionality = 0;
	m_unknownWordVector = NULL;
}

ParaphraseModel::~ParaphraseModel() {}

double ParaphraseModel::trainAndTest(const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename,
		const string& forwardLSTMs, const string& backwardLSTMs, const string& forwardGRUs, const string& backwardGRUs,
		const string& algorithm, double stepsize, int minibatchSize, int numEpochs, int decodeEveryK, double dropoutRate,
		bool updateWordVectors, bool initWordVectorsRandomly, double gradientNormThreshold, double margin, double lambdaW,
		bool doFiniteDifferenceChecking, const string& phraseFile) {

	struct timeval initialTime, endTime, time1, time2, time3, time4,
		timeBuildStart, timeBuildEnd, timeEvalStart, timeEvalEnd, timeDecodeStart, timeDecodeEnd;
	gettimeofday(&initialTime, 0);

	// First, load word vectors corresponding to words in train/dev/test data.
	// We don't tokenize each line into its two halves when doing this, for simplicity.
	loadWordVectors(trainLines, devLines, testLines, wordVectorFilename);

	resetGradientVector();

	map<string, int> unkWordCounts;
	vector<WordSequencePairDatum> trainData, devData, testData;
	loadWordSequencePairData(trainLines, trainData, m_wordVectors, unkWordCounts);
	loadWordSequencePairData(devLines, devData, m_wordVectors, unkWordCounts);
	loadWordSequencePairData(testLines, testData, m_wordVectors, unkWordCounts);
	printUnknownWordCounts(unkWordCounts);

	bool trainTrue = true;
	bool trainFalse = false;

	// Initialize all parameters shared across LSTMs/GRUs.
	initializeSharedParameters(m_wordVectorDimensionality, trainData, devData, testData, updateWordVectors, initWordVectorsRandomly);
	// Initialize parameters for each LSTM/GRU.
	initializeNetworkFromConfiguration(forwardLSTMs, true, true);
	initializeNetworkFromConfiguration(backwardLSTMs, true, false);
	initializeNetworkFromConfiguration(forwardGRUs, false, true);
	initializeNetworkFromConfiguration(backwardGRUs, false, false);

	if (m_updateWordVectors) {
		cerr << "Getting word vector params for training data" << endl;
		getWordVectorParamsForWordSequencePairData(m_v, trainData, trainTrue);
		cerr << "Getting word vector params for dev data" << endl;
		getWordVectorParamsForWordSequencePairData(m_v, devData, trainFalse);
		cerr << "Getting word vector params for test data" << endl;
		getWordVectorParamsForWordSequencePairData(m_v, testData, trainFalse);
	}

	// Print nearest neighbors before any learning.
	//printNearestNeighbors(trainData, trainData, params);
	//printNearestNeighbors(trainData, devData, params);
	double maxDevAcc = -1.0;
	double testAccForMaxDevAcc = -1.0;
	//double maxTestAcc = -1.0;

	FLOAT totalBuildTime = 0.0f;
	FLOAT totalEvalTime = 0.0f;
	FLOAT totalDecodeTime = 0.0f;

	int decodeEveryKToUse = decodeEveryK;
	if (trainData.size() < 11000) {
		decodeEveryKToUse = trainData.size() / 5;
	}

	bool doDropout = true;
	if (dropoutRate <= 0.0) doDropout = false;

	bool scaleWeightsForDecoding = doDropout;
	scaleWeightsForDecoding = false;
	if (scaleWeightsForDecoding) {
		cerr << "Scaling weights for decoding by dropout rate (" << dropoutRate << ")." << endl;
	}

	Algorithm alg = Model::getAlgorithm(algorithm);
	// Initialize vector of previous updates for SGD w/ momentum.
	vector<double> sgdMomentumUpdates;//(model.getNumParams(), 0.0);
	double momentumCoefficient = 0.95;
	// Initialize sum of squared gradients for AdaGrad and AdaDelta.
	vector<double> sumOfSquaredGradients(getNumParams(), 0.0);
	// Initialize sum of squared updates for AdaDelta.
	vector<double> sumOfSquaredUpdates;//(model.getNumParams(), 0.0);
	//double epsilon = 0.000001;
	//double rho = 0.95;
	if (alg != ADAGRAD) {
		cerr << "Error: only AdaGrad is currently supported!" << endl;
	}

	// Initialize sum of squared gradients for AdaGrad.
	//vector<double> sumOfSquaredGradients(getNumParams(), 0.0);
	//int maxNumFormulasUsed = getNumberOfFormulasUsedInPool();
	int maxUnaryFormulasUsed = 0;
	int maxBinaryFormulasUsed = 0;
	int maxFertileFormulasUsed = 0;
	int maxBarrenFormulasUsed = 0;
	int unaryFormulasUsed = 0;
	int binaryFormulasUsed = 0;
	int fertileFormulasUsed = 0;
	int barrenFormulasUsed = 0;
	getNumberOfFormulasUsedInPool(maxUnaryFormulasUsed, maxBinaryFormulasUsed, maxFertileFormulasUsed, maxBarrenFormulasUsed);

	// Create a vector containing all the training sequences by themselves, for purposes of more efficiently
	// choosing negative examples.
	vector<WordSequenceDatum> allTrainingSequences;
	for (unsigned int i = 0; i < trainData.size(); ++i) {
		allTrainingSequences.push_back(trainData[i].m_sequences.first);
		allTrainingSequences.push_back(trainData[i].m_sequences.second);
	}
	vector<int> subset;
	//double lambdaW = 0.1;

	double sumBaselineDev = doDecodingSumBaseline(devData, false);
	double sumBaselineTest = doDecodingSumBaseline(testData, false);
	cerr << "Sum baselines:\tdacc\t" << sumBaselineDev << "\ttacc\t" << sumBaselineTest << endl;
	double averageBaselineDev = doDecodingSumBaseline(devData, true);
	double averageBaselineTest = doDecodingSumBaseline(testData, true);
	cerr << "Average baselines:\tdacc\t" << averageBaselineDev << "\ttacc\t" << averageBaselineTest << endl;

	int numMinibatches = trainData.size() / minibatchSize;
	set<int> minibatch;
	// Train for numEpochs iterations of training.
	for (unsigned int iter = 1; iter <= numEpochs; ++iter) {
		double iterLoss = 0.0;
		double iterGradientNorm = 0.0;
		int numExamplesProcessed = 0;

		//int iterNumCorrect = 0;
		gettimeofday(&time1, 0);
		gettimeofday(&time3, 0);
		// The number of mini-batches to process per epoch should be equal to the training data size divided by
		for (int minibatchIndex = 0; minibatchIndex < numMinibatches; ++minibatchIndex) {
			// Sample a minibatch.
			ParaphraseModel::sampleMinibatch(trainData.size(), minibatchSize, minibatch);

			// Go through the minibatch.
			for (set<int>::const_iterator minibatchIter = minibatch.begin(); minibatchIter != minibatch.end(); ++minibatchIter) {
				int exampleIndex = *minibatchIter;
				WordSequencePairDatum& datum = trainData[exampleIndex];
				++numExamplesProcessed;
				//for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
				//sampleNegativeExampleSubset(allTrainingSequences, subset);
				// Choose negative examples for the sequences in the example.
				WordSequenceDatum* negativeExample1 = chooseNegativeExample(datum.m_sequences.first, datum.m_sequences.second, trainData, minibatch);
				//SequenceDatum& negativeExample1 = allTrainingSequences.at(negativeExample1Index);
				WordSequenceDatum* negativeExample2 = chooseNegativeExample(datum.m_sequences.second, datum.m_sequences.first, trainData, minibatch);
				//SequenceDatum& negativeExample2 = allTrainingSequences.at(negativeExample2Index);
				// We are processing a new example, so reset Formula data structures in the model.
				resetForNewExample(datum, negativeExample1, negativeExample2);

				//cerr << "Computing gradient for sentence \"" << sentenceTokens[0] << " " << sentenceTokens[1]
				//	 << "\" with true label " << trueLabel << endl;

				// Create the formula containing the loss for the current example.
				gettimeofday(&timeBuildStart, 0);
				Formula* datumLossFDAG = createLossFDAGForWordSequencePairDatum(datum, *negativeExample1, *negativeExample2, margin, doDropout, dropoutRate);
				datumLossFDAG->addChild(createL2RegularizationTerm(trainData.size(), lambdaW));

				gettimeofday(&timeBuildEnd, 0);
				totalBuildTime += computeTimeDifference(timeBuildStart, timeBuildEnd);
				getNumberOfFormulasUsedInPool(unaryFormulasUsed, binaryFormulasUsed, fertileFormulasUsed, barrenFormulasUsed);
				if (unaryFormulasUsed > maxUnaryFormulasUsed) {
					maxUnaryFormulasUsed = unaryFormulasUsed;
					cerr << "Info: New unary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << unaryFormulasUsed << endl;
				}
				if (binaryFormulasUsed > maxBinaryFormulasUsed) {
					maxBinaryFormulasUsed = binaryFormulasUsed;
					cerr << "Info: New binary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << binaryFormulasUsed << endl;
				}
				if (fertileFormulasUsed > maxFertileFormulasUsed) {
					maxFertileFormulasUsed = fertileFormulasUsed;
					cerr << "Info: New fertile formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << fertileFormulasUsed << endl;
				}
				if (barrenFormulasUsed > maxBarrenFormulasUsed) {
					maxBarrenFormulasUsed = barrenFormulasUsed;
					cerr << "Info: New barren formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << barrenFormulasUsed << endl;
				}
				//int currNumFormulasUsed = getNumberOfFormulasUsedInPool();
				//if (currNumFormulasUsed > maxNumFormulasUsed) {
				//	cerr << "Info: New formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << currNumFormulasUsed << endl;
				//	maxNumFormulasUsed = currNumFormulasUsed;
				//}
				// Compute gradient and loss for current example.
				gettimeofday(&timeEvalStart, 0);
				double loss = evalAndComputeDerivatives(datumLossFDAG, exampleIndex, doFiniteDifferenceChecking);
				if (loss != loss) {
					cerr << "Error: nan loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].m_sequences.first.toString() <<
						" ||| " << trainData[exampleIndex].m_sequences.second.toString() << endl;
				} else if (loss == neg_inf) {
					cerr << "Error: -inf loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].m_sequences.first.toString() <<
						" ||| " << trainData[exampleIndex].m_sequences.second.toString() << endl;
				} else if (loss == pos_inf) {
					cerr << "Error: inf loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].m_sequences.first.toString() <<
						" ||| " << trainData[exampleIndex].m_sequences.second.toString() << endl;
				} else {
					iterLoss += loss;
				}
				gettimeofday(&timeEvalEnd, 0);
				totalEvalTime += computeTimeDifference(timeEvalStart, timeEvalEnd);
				datum.doneProcessing();
				negativeExample1->doneProcessing();
				negativeExample2->doneProcessing();
			}
			// We're at the end of a mini-batch, so do an update.
			if (gradientNormThreshold > 0.0) {
				double gradientNorm = computeGradientVectorNorm();
				if (gradientNorm >= gradientNormThreshold) {
					//cerr << "Info: gradientNorm = " << gradientNorm << "; Scaling gradient vector by " << (gradientNormThreshold / gradientNorm) << endl;
					scaleGradientVectorByConstant(gradientNormThreshold / gradientNorm);
					//cerr << "Info: new gradientNorm = " << model.computeGradientVectorNorm() << endl;
				}
			}
			double gradientSquaredNorm = 0.0;
			if (alg == SGD_MOMENTUM) {
				gradientSquaredNorm = doSGDMomentumUpdate(sgdMomentumUpdates, stepsize, momentumCoefficient);
			} else if (alg == ADAGRAD) {
				gradientSquaredNorm = doAdaGradUpdate(sumOfSquaredGradients, stepsize);
			} else if (alg == ADADELTA){
				gradientSquaredNorm = doAdaDeltaUpdate(sumOfSquaredGradients, sumOfSquaredUpdates, stepsize, momentumCoefficient);
			}
			iterGradientNorm += computeGradientVectorNorm();
			resetGradientVector();

			if (numExamplesProcessed % decodeEveryKToUse == 0) {
				gettimeofday(&timeDecodeStart, 0);
				gettimeofday(&time4, 0);
				FLOAT timek = computeTimeDifference(time3, time4);
				cerr << "Decoding dev data..." << endl;
				// Decode dev data every decodeEveryK examples.
				if (devData.size() > 0) {
					// If we have dev data, decode the dev data.
					double devAccuracy = doDecoding(devData);
					// If it's the best dev accuracy we've seen so far, decode the test data.
					if (devAccuracy > maxDevAcc) {
						maxDevAcc = devAccuracy;
						cerr << "Decoding test data..." << endl;
						testAccForMaxDevAcc = doDecoding(testData);
					}
					//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
					cerr << "it " << iter << ", " << numExamplesProcessed << "\tdacc\t" << devAccuracy << "\tmaxd\t" << maxDevAcc
						 << "\ttformaxd\t" << testAccForMaxDevAcc; // << "\tmaxt\t" << maxTestAcc;
				} else {
					// Here, we have no dev data, so just decode the test data.
					cerr << "Decoding test data..." << endl;
					testAccForMaxDevAcc = doDecoding(testData);
					cerr << "it " << iter << ", " << numExamplesProcessed << "\ttacc\t" << testAccForMaxDevAcc;
				}

				cerr << "\ttimek\t" << timek << endl;
				gettimeofday(&time3, 0);
				gettimeofday(&timeDecodeEnd, 0);
				totalDecodeTime += computeTimeDifference(timeDecodeStart, timeDecodeEnd);
				cerr << "times:\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
					 << "\tdec\t" << totalDecodeTime << endl;

				//cerr << "times:\treset\t" << totalResetTime << "\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
				//	 << "\tupdate\t" << totalUpdateTime << "\tdec\t" << totalDecodeTime << endl;
			}
		}
		cerr << "it " << iter << "\tloss\t" << iterLoss << "\tparam_norm\t" << computeParameterVectorNorm() <<
				"\tgrad_norm\t" << iterGradientNorm;

		if (devData.size() > 0) {
			// If we have dev data, decode the dev data.
			double devAccuracy = doDecoding(devData);
			// If it's the best dev accuracy we've seen so far, decode the test data.
			if (devAccuracy > maxDevAcc) {
				maxDevAcc = devAccuracy;
				testAccForMaxDevAcc = doDecoding(testData);
			}
			cerr << "\tdacc\t" << devAccuracy << "\tmaxd\t" << maxDevAcc
			<< "\ttformaxd\t" << testAccForMaxDevAcc;
		} else {
			// Here, we have no dev data, so just decode the test data.
			testAccForMaxDevAcc = doDecoding(testData);
			cerr << "\ttacc\t" << testAccForMaxDevAcc;
		}
		gettimeofday(&time2, 0);
		FLOAT iterTime = computeTimeDifference(time1, time2);
		cerr << "\t\ttime\t" << iterTime;
		cerr << endl;

		if (iter % 10 == 0) printRecurrentNeuralNetworkNearestNeighbors(trainData);
	}
	// Print nearest neighbors for dev bigrams.
	//printNearestNeighbors(trainData, devData, params);

	gettimeofday(&endTime, 0);
	FLOAT cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
	cerr << "Final test accuracy:\t" << testAccForMaxDevAcc << endl;

	if (phraseFile != "" && phraseFile != "NONE") {
		embedPhrasesInFile(phraseFile);
		runDemo();
	}

	return testAccForMaxDevAcc;
}

void ParaphraseModel::embedPhrasesInFile(const string& phraseFile) {
	bool doDropoutFalse = false;
	// Go through phrase file line by line.
	ifstream inputFilestream(phraseFile.c_str());
	if (!inputFilestream) {
		cerr << "Error: Cannot open phrase file " << phraseFile << "!" << endl;
		return;
	}
	string line;
	while (!inputFilestream.eof()) {
		getline(inputFilestream, line);
		// Note: following line should be safer:
		if (!inputFilestream.eof()) {
			// Read the current phrase into a WordSequenceDatum.
			WordSequenceDatum d;
			d.readInWords(line, m_wordVectors, NULL);
			// Reset data structures for processing the new phrase.
			resetSingleSequence(d);
			// Embed the phrase.
			vector<Formula*> embedding;
			embedSequence(d, 0, doDropoutFalse, -1.0, embedding);
			unsigned int embeddingLength = embedding.size();
			// Evaluate each Formula to generate the embedding vector.
			vector<double> embeddingVector;
			embeddingVector.reserve(embeddingLength);
			for (unsigned int i = 0; i < embeddingLength; ++i) {
				embeddingVector.push_back(embedding[i]->evaluate());
			}
			// Insert it into m_embeddedPhrases.
			m_embeddedPhrases.insert(make_pair(line, embeddingVector));
		}
	}
	inputFilestream.close();
}

void ParaphraseModel::demoDisplayHelp() {
	cout << "Enter a phrase and press enter, or enter one of the following:" << endl;
	//cout << "\t--r\treset (start new phrase)" << endl;
	cout << "\t--h\tdisplay this help" << endl;
	cout << "\t--q\tquit" << endl;
}

void ParaphraseModel::runDemo() {
	cout << "Starting demo mode (--h for help). " << endl;

	//WordSequenceDatum d;
	bool doDropoutFalse = false;
	while (1) {
		string responseFull;
		cout << ": ";
		//cin >> responseFull;
		getline(cin, responseFull);
		cout << endl;
		string response = trimString(responseFull);
		if (response == "--q" || response == "--Q") {
			cout << "Quitting..." << endl;
			return;
		//} else if (response == "--R" || response == "--r") {
			// Start a new phrase.
			//currentPhrase = "";
			//d.clear();
		} else if (response == "--H" || response == "--h") {
			demoDisplayHelp();
		} else {
			// Process the given phrase.
			WordSequenceDatum d;
			//d.clear();
			d.readInWords(response, m_wordVectors, NULL);

			// Reset data structures for processing the new phrase.
			resetSingleSequence(d);

			// Embed the phrase.
			vector<Formula*> inputEmbedding;
			embedSequence(d, 0, doDropoutFalse, -1.0, inputEmbedding);
			unsigned int embeddingLength = inputEmbedding.size();
			// Evaluate each Formula to generate the embedding vector.
			vector<double> embeddingVector;
			embeddingVector.reserve(embeddingLength);
			for (unsigned int i = 0; i < embeddingLength; ++i) {
				embeddingVector.push_back(inputEmbedding[i]->evaluate());
			}

			// Go through all phrases in the set and compute cosine similarity to each, retaining the top numNeighbors nearest neighbors.
			unsigned int numNeighbors = 20;

			priority_queue<pair<string, double>, vector<pair<string, double> >, CompareStringDoublePairPriorityQueue> pq;
			for (map<string, vector<double> >::const_iterator iter = m_embeddedPhrases.begin(); iter != m_embeddedPhrases.end(); ++iter) {
				// Compute cosine similarity between embeddings of sequences i and j.
				double currSim = cosineSimilarity(embeddingVector, iter->second);
				pq.push(make_pair(iter->first, currSim));
			}
			// Print the k-nearest neighbors.
			cout << "Printing k-nearest neighbors:" << endl;
			for (unsigned int k = 0; k < numNeighbors; ++k) {
				string currWord = pq.top().first;
				double cosineSim = pq.top().second;
				cout << cosineSim << ":\t" << currWord << endl;
				pq.pop();
			}
			cout << endl;
			/*for (unsigned int i = 0; i < similarities.size(); ++i) {
				for (unsigned int j = 0; j < similarities[i].size(); ++j) {
					double currSim = similarities[i][j];
					// Insert the current similarity into the priority queue.
					pq.insert(make_pair(j, currSim));
				}

				for (unsigned int k = 0; k < numNeighbors && pq.size() > 0; ++k) {
					//string currWordSequence = pq.top().first;
					//double filterSum = pq.top().second;
					int datumIndex = pq.begin()->first;
					double datumSim = pq.begin()->second;
					cerr << datumSim << ":\t" << data[datumIndex].toString() << endl;
					//topMatches.push_back(currWordSequence);
					pq.erase(pq.begin());
					//pq.pop();
				}
			}*/
		}
	}

}

void ParaphraseModel::sampleNegativeExampleSubset(const vector<WordSequenceDatum>& all, vector<int>& subset) {
	subset.clear();
	for (unsigned int i = 0; i < all.size(); ++i) {
		if (getRandomDouble() < 0.05) {
			subset.push_back(i);
		}
	}
	if (subset.size() == 0) {
		subset.push_back(0);
		subset.push_back(1);
	}
}

WordSequenceDatum* ParaphraseModel::chooseNegativeExample(const WordSequenceDatum& x1, const WordSequenceDatum& x2, vector<WordSequencePairDatum>& all,
		const set<int>& indexSubset) const {
	// Go through the given vector of sequences and find the one with the largest dot product when computed with x1, as long as
	// it doesn't equal x2.
	double maxScore = neg_inf;
	WordSequenceDatum* ret = NULL;
	for (set<int>::const_iterator iter = indexSubset.begin(); iter != indexSubset.end(); ++iter) {
		int currIndex = *iter;
		WordSequencePairDatum& curr = all[currIndex];
		// Check if the current sequence equals x2.
		if (curr.m_sequences.first.m_cachedToString != x1.m_cachedToString && curr.m_sequences.first.m_cachedToString != x2.m_cachedToString) {
			//cerr << "Info: For pair \"" << x1.toString() << " ||| " << x2.toString() << "\", skipping equivalent seq \"" << curr.toString() << "\"" << endl;
			// Compute the dot product between the embeddings of x1 and the current entry of all.
			// For efficiency for now, just use the sum of the vectors of the words in each phrase, rather
			// than using the LSTM/GRU.
			double dotP = cosineSimilarity(x1.m_vectorAverage, curr.m_sequences.first.m_vectorAverage);
			if (dotP > maxScore) {
				maxScore = dotP;
				ret = &(curr.m_sequences.first);
			}
		}
		// Check if the current sequence equals x2.
		if (curr.m_sequences.second.m_cachedToString != x1.m_cachedToString && curr.m_sequences.second.m_cachedToString != x2.m_cachedToString) {
			//cerr << "Info: For pair \"" << x1.toString() << " ||| " << x2.toString() << "\", skipping equivalent seq \"" << curr.toString() << "\"" << endl;
			// Compute the dot product between the embeddings of x1 and the current entry of all.
			// For efficiency for now, just use the sum of the vectors of the words in each phrase, rather
			// than using the LSTM/GRU.
			double dotP = cosineSimilarity(x1.m_vectorAverage, curr.m_sequences.second.m_vectorAverage);
			if (dotP > maxScore) {
				maxScore = dotP;
				ret = &(curr.m_sequences.second);
			}
		}
	}
	//cerr << "Info: For sequence \"" << x1.m_cachedToString << "\", using negative example \"" << ret->m_cachedToString << "\"" << endl;

	return ret;
}


void ParaphraseModel::initializeSharedParameters(int inputDimension, const vector<WordSequencePairDatum>& trainData,
		const vector<WordSequencePairDatum>& devData, const vector<WordSequencePairDatum>& testData,
		bool updateWordVectors, bool initWordVectorsRandomly) {
	m_inputDimension = inputDimension;
	m_numLabels = 2;
	m_updateWordVectors = updateWordVectors;

	if (m_updateWordVectors) {
		// Only create parameters for words in training/dev/test data. Words that only appear in dev/test data
		// will not have their parameters updated during training; it's just useful to have parameters for them
		// so that we can do the same procedure during decoding as we do during training.
		set<string> paramWords;
		insertWordSequencePairDataWordsIntoSet(trainData, paramWords);
		insertWordSequencePairDataWordsIntoSet(devData, paramWords);
		insertWordSequencePairDataWordsIntoSet(testData, paramWords);

		for (set<string>::const_iterator iter = paramWords.begin(); iter != paramWords.end(); ++iter) {
			const string& currWord = *iter;
			bool unk = false;
			const vector<double>* currVector = getWordVector(m_wordVectors, currWord, unk);
			vector<PARAM_ID> currVectorParams;
			// Go through all dimensions of the current word vector.
			for (unsigned int i = 0; i < currVector->size(); ++i) {
				// Create a new parameter for the current word vector entry.
				if (initWordVectorsRandomly) {
					currVectorParams.push_back(initializeNewParameter(getRandomDouble() - 0.5));
				} else {
					currVectorParams.push_back(initializeNewParameter(currVector->at(i)));
				}
			}
			m_v.insert(make_pair(currWord, currVectorParams));
		}
	}
	cerr << "Initialized model. # parameters = " << getNumParams() << endl;
}

void ParaphraseModel::resetSingleSequence(WordSequenceDatum& d) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	bool resetCachedParameterFormulasTrue = true;
	resetFormulaObjects(resetCachedParameterFormulasTrue);
	// The number of cached nodes is the length of the sequence times the hidden layer width.
	unsigned int numWords = d.m_words.size();
	for (unsigned int i = 0; i < m_lstmParams.size(); ++i) {
		//cerr << "Resetting caches in LSTM " << (i+1) << " for datum:" << endl;
		m_lstmParams[i].resetForNewExample(0, numWords);
	}
	for (unsigned int i = 0; i < m_gruParams.size(); ++i) {
		m_gruParams[i].resetForNewExample(0, numWords);
	}
	//cerr << "Precomputing word vector formulas:" << endl;
	precomputeWordVectorFormulas(this, d, m_inputDimension, m_updateWordVectors);
}

// Resets data structures for processing the next example (whether in training or decoding).
void ParaphraseModel::resetForNewExample(WordSequencePairDatum& datum, WordSequenceDatum* negativeExample1, WordSequenceDatum* negativeExample2) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	bool resetCachedParameterFormulasTrue = true;
	resetFormulaObjects(resetCachedParameterFormulasTrue);
	// The number of cached nodes is the length of the sequence times the hidden layer width.
	unsigned int numWords1 = datum.m_sequences.first.m_words.size();
	unsigned int numWords2 = datum.m_sequences.second.m_words.size();
	for (unsigned int i = 0; i < m_lstmParams.size(); ++i) {
		//cerr << "Resetting caches in LSTM " << (i+1) << " for datum:" << endl;
		m_lstmParams[i].resetForNewExample(0, numWords1);
		m_lstmParams[i].resetForNewExample(1, numWords2);
		if (negativeExample1 != NULL) m_lstmParams[i].resetForNewExample(2, negativeExample1->m_words.size());
		if (negativeExample2 != NULL) m_lstmParams[i].resetForNewExample(3, negativeExample2->m_words.size());
	}
	for (unsigned int i = 0; i < m_gruParams.size(); ++i) {
		m_gruParams[i].resetForNewExample(0, numWords1);
		m_gruParams[i].resetForNewExample(1, numWords2);
		if (negativeExample1 != NULL) m_gruParams[i].resetForNewExample(2, negativeExample1->m_words.size());
		if (negativeExample2 != NULL) m_gruParams[i].resetForNewExample(3, negativeExample2->m_words.size());
	}
	//cerr << "Precomputing word vector formulas:" << endl;
	precomputeWordVectorFormulas(this, datum.m_sequences.first, m_inputDimension, m_updateWordVectors);
	precomputeWordVectorFormulas(this, datum.m_sequences.second, m_inputDimension, m_updateWordVectors);
	if (negativeExample1 != NULL) precomputeWordVectorFormulas(this, *negativeExample1, m_inputDimension, m_updateWordVectors);
	if (negativeExample2 != NULL) precomputeWordVectorFormulas(this, *negativeExample2, m_inputDimension, m_updateWordVectors);
}

void ParaphraseModel::embedSequence(WordSequenceDatum& datum, unsigned int sequenceIndex,
		bool doDropout, double dropoutRate, vector<Formula*>& embedding) {
	int numTermsInEmbedding = 0;
	for (unsigned int lstmIndex = 0; lstmIndex < m_lstmParams.size(); ++lstmIndex) {
		DeepLSTM& lstmParams = m_lstmParams[lstmIndex];
		// Go through all hidden layers for the current deep LSTM.
		for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
			numTermsInEmbedding += lstmParams.m_layerWidths[j];
		}
	}
	for (unsigned int gruIndex = 0; gruIndex < m_gruParams.size(); ++gruIndex) {
		DeepGRU& gruParams = m_gruParams[gruIndex];
		// Go through all hidden layers for the current deep GRU.
		for (unsigned int j = 0; j < gruParams.m_layerWidths.size(); ++j) {
			numTermsInEmbedding += gruParams.m_layerWidths[j];
		}
	}
	//cerr << "Reserving " << numTermsInEmbedding << " terms in embedding" << endl;
	embedding.clear();
	embedding.reserve(numTermsInEmbedding);
	// Loop over all dimensions of the final hidden vectors for all LSTMs and GRUs, adding an entry
	// to the embedding vector for each dimension.

	for (unsigned int lstmIndex = 0; lstmIndex < m_lstmParams.size(); ++lstmIndex) {
		DeepLSTM& lstmParams = m_lstmParams[lstmIndex];
		// Go through all hidden layers for the current deep LSTM.
		for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
			// Go through all entries in the current hidden layer for the current deep LSTM.
			for (unsigned int k = 0; k < lstmParams.m_layerWidths[j]; ++k) {
				//bool dropout = false;
				//if (doDropout && getRandomDouble() < dropoutRate) dropout = true;
				//if (!dropout) {
					// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (lstmParams.m_forward) startTime = datum.m_sequenceLength - 1;
					int startDepth = j;
					Formula* zj = lstmParams.createHiddenNodeAtTimeAtDepth(*this, datum, sequenceIndex, k, startTime, startDepth, doDropout, dropoutRate);
					embedding.push_back(zj);
				//}
			}
		}
	}
	for (unsigned int gruIndex = 0; gruIndex < m_gruParams.size(); ++gruIndex) {
		DeepGRU& gruParams = m_gruParams[gruIndex];
		// Go through all hidden layers for the current deep GRU.
		for (unsigned int j = 0; j < gruParams.m_layerWidths.size(); ++j) {
			// Go through all entries in the current hidden layer for the current deep GRU.
			for (unsigned int k = 0; k < gruParams.m_layerWidths[j]; ++k) {
				//bool dropout = false;
				//if (doDropout && getRandomDouble() < dropoutRate) dropout = true;
				//if (!dropout) {
					// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (gruParams.m_forward) startTime = datum.m_sequenceLength - 1;
					int startDepth = j;
					Formula* zj = gruParams.createHiddenNodeAtTimeAtDepth(*this, datum, sequenceIndex, k, startTime, startDepth, doDropout, dropoutRate);
					embedding.push_back(zj);
				//}
			}
		}
	}
}

Formula* ParaphraseModel::createDotProduct(vector<Formula*>& v1, vector<Formula*>& v2) {
	unsigned int len = v1.size();
	Formula* dotProductFormula = getFormulaObject(PLUS, len);
	for (unsigned int i = 0; i < len; ++i) {
		Formula* dotProductTerm = getFormulaObject(TIMES, 2);
		dotProductTerm->addChild(v1[i]);
		dotProductTerm->addChild(v2[i]);
		dotProductFormula->addChild(dotProductTerm);
	}
	return dotProductFormula;
}

Formula* ParaphraseModel::createLossFDAGForWordSequencePairDatum(WordSequencePairDatum& datum, WordSequenceDatum& negativeExample1,
		WordSequenceDatum& negativeExample2, double marginDelta, bool doDropout, double dropoutRate) {
	// Build the embedding vectors for the four sequences.
	vector<Formula*> embeddingCorrect1, embeddingCorrect2, embeddingNegative1, embeddingNegative2;
	embedSequence(datum.m_sequences.first, 0, doDropout, dropoutRate, embeddingCorrect1);
	embedSequence(datum.m_sequences.second, 1, doDropout, dropoutRate, embeddingCorrect2);
	//unsigned int embeddingLength = embeddingCorrect1.size();
	embedSequence(negativeExample1, 2, doDropout, dropoutRate, embeddingNegative1);
	embedSequence(negativeExample2, 3, doDropout, dropoutRate, embeddingNegative2);

	Formula* firstSum = getFormulaObject(PLUS, 3);
	// The first term is the margin.
	Formula* marginDeltaFormula = getConstantFormulaObject(marginDelta);
	firstSum->addChild(marginDeltaFormula);
	//sum->addChild(marginDeltaFormula);

	// The second term is the negated dot product of the correct pair.
	Formula* correctDotProduct = createDotProduct(embeddingCorrect1, embeddingCorrect2);
	Formula* negCorrectDotProduct = getFormulaObject(NEG, 1);
	negCorrectDotProduct->addChild(correctDotProduct);
	firstSum->addChild(negCorrectDotProduct);

	// The first sum also contains the dot product of the first incorrect pair.
	Formula* incorrectDotProduct1 = createDotProduct(embeddingCorrect1, embeddingNegative1);
	firstSum->addChild(incorrectDotProduct1);
	// Create the first hinge loss.
	Formula* firstMax = getFormulaObject(MAX, 2);
	firstMax->addChild(getConstantFormulaObject(0.0));
	firstMax->addChild(firstSum);

	// The second hinge loss term contains a similar sum containing the margin and
	// the negated correct dot product.
	Formula* secondSum = getFormulaObject(PLUS, 3);
	secondSum->addChild(marginDeltaFormula);
	secondSum->addChild(negCorrectDotProduct);
	// The second sum also contains the dot product of the second incorrect pair.
	secondSum->addChild(createDotProduct(embeddingCorrect2, embeddingNegative2));

	// Create the second hinge loss.
	Formula* secondMax = getFormulaObject(MAX, 2);
	secondMax->addChild(getConstantFormulaObject(0.0));
	secondMax->addChild(secondSum);

	// Add the two hinge losses and return the sum.
	Formula* outerSum = getFormulaObject(PLUS, 2);
	outerSum->addChild(firstMax);
	outerSum->addChild(secondMax);
	return outerSum;
}

Formula* ParaphraseModel::createL2RegularizationTerm(int numExamples, double lambdaW) {
	// The root node is a product node.
	Formula* ret = getFormulaObject(TIMES, 2);
	// The first term is the constant multiplier.
	Formula* multiplier = getConstantFormulaObject(lambdaW / (1.0 * ((double)numExamples)));
	ret->addChild(multiplier);
	Formula* l2term = getFormulaObject(PLUS, m_V.size());
	// Go through all parameters in the model.
	for (unsigned int i = 0; i < m_V.size(); ++i) {
		// Square the weight for parameter i.
		Formula* paramFormula = getCachedParameterFormula(i);
		Formula* paramSquared = getFormulaObject(TIMES, 2);
		paramSquared->addChild(paramFormula);
		paramSquared->addChild(paramFormula);
		l2term->addChild(paramSquared);
	}
	ret->addChild(l2term);
	return ret;
}

double ParaphraseModel::decode(WordSequencePairDatum& datum) {
	bool doDropoutFalse = false;

	// The second term is the negated dot product of the correct pair.
	vector<Formula*> embedding1;
	embedSequence(datum.m_sequences.first, 0, doDropoutFalse, -1.0, embedding1);
	vector<Formula*> embedding2;
	embedSequence(datum.m_sequences.second, 1, doDropoutFalse, -1.0, embedding2);
	unsigned int embeddingLength = embedding1.size();
	// Evaluate each Formula to generate embedding vectors.
	vector<double> embedding1Vector, embedding2Vector;
	embedding1Vector.reserve(embeddingLength);
	embedding2Vector.reserve(embeddingLength);
	for (unsigned int i = 0; i < embeddingLength; ++i) {
		embedding1Vector.push_back(embedding1[i]->evaluate());
		embedding2Vector.push_back(embedding2[i]->evaluate());
	}
	return cosineSimilarity(embedding1Vector, embedding2Vector);
}

double ParaphraseModel::doDecoding(vector<WordSequencePairDatum>& data) {
	// Use our trained model to assign a continuous score for all examples in the given vector.

	vector<double> modelSimilarities, goldSimilarities;
	modelSimilarities.reserve(data.size());
	goldSimilarities.reserve(data.size());
	for (unsigned int i = 0; i < data.size(); ++i) {
		WordSequencePairDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(datum, NULL, NULL);

		double modelSim = decode(datum);
		modelSimilarities.push_back(modelSim);
		datum.doneProcessing();

		goldSimilarities.push_back(data[i].m_similarity);
	}
	double spearmanRho;
	double d = 0;
	if (modelSimilarities.size() != goldSimilarities.size()) {
		cerr << "Error: size difference! modelSimilarities.size() = " << modelSimilarities.size()
			 << ", goldSimilarities.size() " << goldSimilarities.size() << endl;
	}
	// Compute Spearman's rank correlation between the vector of our scores and the gold similarities.
	spearmanCorrelation(modelSimilarities, goldSimilarities, d, spearmanRho);
	return spearmanRho;
}

double ParaphraseModel::decodeSumBaseline(WordSequencePairDatum& datum, bool useAverage) {
	// Compute the cosine similarity between the averaged or summed vectors of the two sequences.
	if (useAverage) {
		return cosineSimilarity(datum.m_sequences.first.m_vectorAverage, datum.m_sequences.second.m_vectorAverage);
	} else {
		return cosineSimilarity(datum.m_sequences.first.m_vectorSum, datum.m_sequences.second.m_vectorSum);
	}
}

double ParaphraseModel::doDecodingSumBaseline(vector<WordSequencePairDatum>& data, bool useAverage) {
	// Use our trained model to assign a continuous score for all examples in the given vector.

	vector<double> modelSimilarities, goldSimilarities;
	modelSimilarities.reserve(data.size());
	goldSimilarities.reserve(data.size());
	for (unsigned int i = 0; i < data.size(); ++i) {
		WordSequencePairDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(datum, NULL, NULL);

		double modelSim = decodeSumBaseline(datum, useAverage);
		modelSimilarities.push_back(modelSim);
		datum.doneProcessing();

		goldSimilarities.push_back(data[i].m_similarity);
	}
	double spearmanRho;
	double d = 0;
	if (modelSimilarities.size() != goldSimilarities.size()) {
		cerr << "Error: size difference! modelSimilarities.size() = " << modelSimilarities.size()
			 << ", goldSimilarities.size() " << goldSimilarities.size() << endl;
	}
	// Compute Spearman's rank correlation between the vector of our scores and the gold similarities.
	spearmanCorrelation(modelSimilarities, goldSimilarities, d, spearmanRho);
	return spearmanRho;
}


struct CompareIntDoublePairSet {
	bool operator() (const pair<int, double>& a, const pair<int, double>& b) const {
		return a.second > b.second;
	}
};

void ParaphraseModel::printRecurrentNeuralNetworkNearestNeighbors(vector<WordSequencePairDatum>& data) {
/*
	// Go through the data and embed each sequence.
	bool doDropoutFalse = false, scaleWeightsFalse = false, testTimeFalse = false;
	bool forwardTrue = true, forwardFalse = false;

	vector<vector<double> > sequenceEmbeddings;
	int numberToPrint = 200;
	if (numberToPrint > data.size()) numberToPrint = data.size();
	sequenceEmbeddings.reserve(numberToPrint);
	for (unsigned int i = 0; i < numberToPrint; ++i) {
		WordSequenceDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(datum);
		// Get the RNN vector for the sequence.
		vector<double> currEmbedding;
		int sizeToReserve = 0;
		for (unsigned int lstmIndex = 0; lstmIndex < m_lstmParams.size(); ++lstmIndex) {
			DeepLSTMParams& lstmParams = m_lstmParams[lstmIndex];
			// Go through all hidden layers for the current deep LSTM.
			for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
				sizeToReserve += lstmParams.m_layerWidths[j];
			}
		}
		for (unsigned int gruIndex = 0; gruIndex < m_gruParams.size(); ++gruIndex) {
			DeepGRUParams& gruParams = m_gruParams[gruIndex];
			// Go through all hidden layers for the current deep GRU.
			for (unsigned int j = 0; j < gruParams.m_layerWidths.size(); ++j) {
				sizeToReserve += gruParams.m_layerWidths[j];
			}
		}
		currEmbedding.reserve(sizeToReserve);

		for (unsigned int lstmIndex = 0; lstmIndex < m_lstmParams.size(); ++lstmIndex) {
			DeepLSTMParams& lstmParams = m_lstmParams[lstmIndex];
			// Go through all hidden layers for the current deep LSTM.
			for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
				// Go through all entries in the current hidden layer for the current deep LSTM.
				for (unsigned int k = 0; k < lstmParams.m_layerWidths[j]; ++k) {
					// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (lstmParams.m_forward) startTime = datum.m_numWords - 1;
					int startDepth = j;
					Formula* zj = lstmParams.createLSTMHiddenNodeAtTimeAtDepth(*this, datum, k, startTime, startDepth, doDropoutFalse, -1.0, testTimeFalse);
					double currEntry = zj->evaluate();
					currEmbedding.push_back(currEntry);
				}
			}
		}
		for (unsigned int gruIndex = 0; gruIndex < m_gruParams.size(); ++gruIndex) {
			DeepGRUParams& gruParams = m_gruParams[gruIndex];
			// Go through all hidden layers for the current deep LSTM.
			for (unsigned int j = 0; j < gruParams.m_layerWidths.size(); ++j) {
				// Go through all entries in the current hidden layer for the current deep LSTM.
				for (unsigned int k = 0; k < gruParams.m_layerWidths[j]; ++k) {
					// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (gruParams.m_forward) startTime = datum.m_numWords - 1;
					int startDepth = j;
					Formula* zj = gruParams.createGRUHiddenNodeAtTimeAtDepth(*this, datum, j, startTime, startDepth, doDropoutFalse, -1.0, testTimeFalse);
					double currEntry = zj->evaluate();
					currEmbedding.push_back(currEntry);
				}
			}
		}

		sequenceEmbeddings.push_back(currEmbedding);
	}

	// Then go back through embedded sequences and compute all pairwise cosine similarities.
	vector<vector<double> > similarities;
	similarities.reserve(numberToPrint);
	for (unsigned int i = 0; i < numberToPrint; ++i) {
		vector<double> emptySimilarities(numberToPrint, -99.0);
		similarities.push_back(emptySimilarities);
	}
	for (unsigned int i = 0; i < sequenceEmbeddings.size(); ++i) {
		for (unsigned int j = i; j < sequenceEmbeddings.size(); ++j) {
			// Compute cosine similarity between embeddings of sequences i and j.
			double currSim = cosineSimilarity(sequenceEmbeddings[i], sequenceEmbeddings[j]);
			similarities[i][j] = currSim;
			similarities[j][i] = currSim;
		}
	}

	// For each sequence, find its k-nearest neighbors.
	unsigned int numNeighbors = 10;
	cerr << "Printing k-nearest neighbors of sequence embeddings:" << endl;
	for (unsigned int i = 0; i < similarities.size(); ++i) {
		set<pair<int, double>, CompareIntDoublePairSet > pq;
		for (unsigned int j = 0; j < similarities[i].size(); ++j) {
			double currSim = similarities[i][j];
			// Insert the current similarity into the priority queue.
			pq.insert(make_pair(j, currSim));
		}
		cerr << "For sequence:" << endl;
		cerr << data[i].toString() << endl;
		for (unsigned int k = 0; k < numNeighbors && pq.size() > 0; ++k) {
			//string currWordSequence = pq.top().first;
			//double filterSum = pq.top().second;
			int datumIndex = pq.begin()->first;
			double datumSim = pq.begin()->second;
			cerr << datumSim << ":\t" << data[datumIndex].toString() << endl;
			//topMatches.push_back(currWordSequence);
			pq.erase(pq.begin());
			//pq.pop();
		}

	}
*/
}

void ParaphraseModel::loadWordVectors(const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename) {
	// Build the relevant word vector vocabulary from the data lines.
	set<string> vocab;
	addWordsFromLinesToVocabulary(trainLines, vocab);
	addWordsFromLinesToVocabulary(devLines, vocab);
	addWordsFromLinesToVocabulary(testLines, vocab);
	// Read in word vectors corresponding to the entries in the vocabulary.
	readWordVectorFile(wordVectorFilename, vocab, m_wordVectors, m_wordVectorDimensionality, m_unknownWordVector);
}

void ParaphraseModel::initializeNetworkFromConfiguration(const string& cfg, bool lstm, bool forward) {
	if (cfg != "0") {
		vector<string> networks;
		tokenize(cfg, '.', networks);
		for (unsigned int i = 0; i < networks.size(); ++i) {
			// Within each network, tokenize the number of networks from the sequence of hidden layer widths.
			vector<string> toks;
			tokenize(networks[i], ':', toks);
			if (toks.size() != 2) {
				cerr << "Error: no network parameters given in token \"" << networks[i] << "\"." << endl;
			} else {
				int numNetworks = atoi(toks[0].c_str());
				if (lstm) {
					for (int j = 0; j < numNetworks; ++j) {
						cerr << "Initializing deep LSTM " << (j+1) << " with layer widths " << toks[1] << " and for 4 sequences per datum." << endl;
						DeepLSTM lstmParams;
						lstmParams.initializeParameters(*this, m_inputDimension, m_numLabels, forward, toks[1], 4, false, true, true, true, false);
						//initializeDeepLSTMParameters(model, params, lstmParams, forward, toks[1]);
						m_lstmParams.push_back(lstmParams);
					}
				} else {
					for (int j = 0; j < numNetworks; ++j) {
						cerr << "Initializing deep GRU " << (j+1) << " with layer widths " << toks[1] << " and for 4 sequences per datum." << endl;
						DeepGRU gruParams;
						gruParams.initializeParameters(*this, m_inputDimension, m_numLabels, forward, toks[1], 4, false, true, true);
						m_gruParams.push_back(gruParams);
					}
				}
			}
		}
	}
}
