#ifndef PARAPHRASE_MODEL_H_
#define PARAPHRASE_MODEL_H_

#include <fstream>
#include <iostream>
#include <vector>
#include <string>
#include <set>
#include <queue>
#include <list>
#include "../../fdag/model.h"
#include "../classifier_utils.h"
#include "../rnn/lstm.h"
#include "../rnn/gru.h"
//#include "constants.h"

//?

struct CompareStringDoublePairPriorityQueue {
	bool operator() (const pair<string, double>& a, const pair<string, double>& b) const {
		return a.second < b.second;
	}
};

struct CompareStringDoublePairSet {
	bool operator() (const pair<string, double>& a, const pair<string, double>& b) const {
		return a.second > b.second;
	}
};

class ParaphraseModel : public Model {
public:
	ParaphraseModel(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity);
	~ParaphraseModel();
	double trainAndTest(const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename,
		const string& forwardLSTMs, const string& backwardLSTMs, const string& forwardGRUs, const string& backwardGRUs,
		const string& algorithm, double stepsize, int minibatchSize, int numEpochs, int decodeEveryK, double dropoutRate,
		bool updateWordVectors, bool initWordVectorsRandomly, double gradientNormThreshold, double margin, double lambdaW,
		bool doFiniteDifferenceChecking, const string& phraseFile);

private:
	static void sampleMinibatch(unsigned int numTotal, unsigned int numToSample, set<int>& minibatch) {
		minibatch.clear();
		// If the number to sample is larger than the total number of items, just insert them all.
		if (numToSample >= numTotal || numTotal < 1) {
			for (unsigned int i = 0; i < numTotal; ++i) {
				minibatch.insert(i);
			}
		} else {
			while (minibatch.size() < numToSample) {
				minibatch.insert(getRandomInteger(numTotal - 1));
			}
		}
	}

	void embedPhrasesInFile(const string& phraseFile);
	void demoDisplayHelp();
	void runDemo();

	void sampleNegativeExampleSubset(const vector<WordSequenceDatum>& all, vector<int>& subset);
	WordSequenceDatum* chooseNegativeExample(const WordSequenceDatum& x1, const WordSequenceDatum& x2, vector<WordSequencePairDatum>& all, const set<int>& indexSubset) const;
	void initializeSharedParameters(int inputDimension, const vector<WordSequencePairDatum>& trainData,
		const vector<WordSequencePairDatum>& devData, const vector<WordSequencePairDatum>& testData,
		bool updateWordVectors, bool initWordVectorsRandomly);

	void resetSingleSequence(WordSequenceDatum& d);
	void resetForNewExample(WordSequencePairDatum& datum, WordSequenceDatum* negativeExample1, WordSequenceDatum* negativeExample2);
	void embedSequence(WordSequenceDatum& datum, unsigned int sequenceIndex, bool doDropout, double dropoutRate,
		vector<Formula*>& embedding);
	//Formula* createSoftmaxScoreForLabel(unsigned int label, WordSequencePairDatum& datum, bool doDropout, double dropoutRate,
	//	bool testTime);
	//Formula* createLossFDAGForWordSequencePairDatum(WordSequencePairDatum& datum, int exampleNumber, bool doDropout, double dropoutRate);
	Formula* createDotProduct(vector<Formula*>& v1, vector<Formula*>& v2);
	Formula* createLossFDAGForWordSequencePairDatum(WordSequencePairDatum& datum, WordSequenceDatum& negativeExample1,
			WordSequenceDatum& negativeExample2, double marginDelta, bool doDropout, double dropoutRate);
	Formula* createL2RegularizationTerm(int numExamples, double lambdaW);
	double decode(WordSequencePairDatum& datum);
	double doDecoding(vector<WordSequencePairDatum>& data);
	double decodeSumBaseline(WordSequencePairDatum& datum, bool useAverage);
	double doDecodingSumBaseline(vector<WordSequencePairDatum>& data, bool useAverage);
	void printRecurrentNeuralNetworkNearestNeighbors(vector<WordSequencePairDatum>& data);
	void loadWordVectors(const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename);
	void initializeNetworkFromConfiguration(const string& cfg, bool lstm, bool forward);

	// The model can contain multiple LSTM RNNs and multiple GRU (gated recurrent unit) RNNs.
	vector<DeepLSTM> m_lstmParams;
	vector<DeepGRU> m_gruParams;

	// Parameters and flags for word vectors:
	bool m_updateWordVectors;				// If true, will include word vector entries as parameters to be learned.
	map<string, vector<double> > m_wordVectors;	// Word vector entries from word vector file.
	map<string, vector<PARAM_ID> > m_v;		// Word vector parameters, initialized to word vector values from file.
	int m_wordVectorDimensionality;			// Dimensionality of word vectors.
	vector<double>* m_unknownWordVector;	// Pointer to unknown word vector.

	map<string, vector<double> > m_embeddedPhrases;	// Embedded phrases for demo mode.

	// The input dimension is the number of features.
	unsigned int m_inputDimension;
	// The number of labels.
	int m_numLabels;
};

#endif
