#include "bidir_lstm.h"

void DeepBidirectionalLSTM::initializeParameters(Model& model, int dataInputDimension, int numLabels, const string& layerWidths,
		unsigned int numSequencesPerDatum, bool initializeSoftmaxParameters, bool useLookahead) {
	// We initialize all weights to a random double between -0.01 and 0.01.
	double scalingFactor = 0.2;
	m_useLookahead = useLookahead;
	vector<string> layerWidthStrings;
	tokenize(layerWidths, ',', layerWidthStrings);
	if (layerWidths == "0") {
		m_numLayers = 0;
		cerr << "Error: 0 hidden layers desired for LSTM. Must use at least 1 hidden layer." << endl;
	} else {
		m_numLayers = layerWidthStrings.size();
	}
	m_layerWidths.clear();
	m_cachedLayerWidthSum = 0;
	for (unsigned int i = 0; i < layerWidthStrings.size(); ++i) {
		int currentLayerWidth = atoi(layerWidthStrings[i].c_str());
		m_layerWidths.push_back(currentLayerWidth);
		m_cachedLayerWidthSum += currentLayerWidth;
	}
	if (m_numLayers != m_layerWidths.size()) {
		cerr << "Error: DeepBidirectionalLSTMParams m_numLayers (" << m_numLayers << ") != m_layerWidths.size() (" << m_layerWidths.size() << ")!" << endl;
	}

	// Parameters for observation-gate matrices; more generally, these matrices connect variables vertically.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix 0: m_inputDimension x m_layerWidths[0] (because they connect the input to the first hidden layer).
	// Dimensions of matrix i > 0: m_layerWidths[i-1] x m_layerWidths[i].
	setVerticalMatrixDimensions(dataInputDimension, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, m_layerWidths);

	model.initializeParameterMatrices(m_weightsObsInput, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsObsForget, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsObsCell, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsObsOutput, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	if (m_useLookahead) {
		model.initializeParameterMatrices(m_weightsNextObsInput, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
		model.initializeParameterMatrices(m_weightsNextObsForget, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
		model.initializeParameterMatrices(m_weightsNextObsOutput, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	}
	// Parameters for hidden-gate matrices; more generally, these matrices connect variables horizontally.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_layerWidths[i].
	setHorizontalMatrixDimensions(dataInputDimension, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, m_layerWidths);

	model.initializeParameterMatrices(m_weightsHiddenInput, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsHiddenForget, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsHiddenCell, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsHiddenOutput, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);

	// Parameters for diagonal matrices connecting cells and gates.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	model.initializeParameterMatrixGivenRowLengths(m_weightsCellInput, m_layerWidths, scalingFactor);
	model.initializeParameterMatrixGivenRowLengths(m_weightsCellForget, m_layerWidths, scalingFactor);
	model.initializeParameterMatrixGivenRowLengths(m_weightsCellOutput, m_layerWidths, scalingFactor);

	// Parameters for gate and cell bias vectors.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	model.initializeParameterMatrixGivenRowLengths(m_inputGateBias, m_layerWidths, scalingFactor);
	model.initializeParameterMatrixGivenRowLengths(m_forgetGateBias, m_layerWidths, scalingFactor);
	model.initializeParameterMatrixGivenRowLengths(m_cellBias, m_layerWidths, scalingFactor);
	model.initializeParameterMatrixGivenRowLengths(m_outputGateBias, m_layerWidths, scalingFactor);

	if (initializeSoftmaxParameters) {
		// Parameters for softmax weight matrices. Outermost vector ranges over hidden layers.
		// Dimensions of matrix i: m_layerWidths[i] x m_numLabels.
		setSoftmaxMatrixDimensions(numLabels, m_softmaxMatrixInputDimensions, m_softmaxMatrixOutputDimensions, m_layerWidths);
		model.initializeParameterMatrices(m_softmaxWeights, m_softmaxMatrixInputDimensions, m_softmaxMatrixOutputDimensions, scalingFactor);

		// Parameters for softmax bias vectors. Outermost vector ranges over hidden layers.
		// Dimensions of vector i: m_numLabels.
		model.initializeParameterMatrix(m_softmaxBias, m_numLayers, numLabels, scalingFactor);
	}

	// Set up Formula caches for the max number of sequences that will appear per datum.
	m_cachedCellNodes.reserve(numSequencesPerDatum);
	m_cachedHiddenNodes.reserve(numSequencesPerDatum);
	for (unsigned int i = 0; i < numSequencesPerDatum; ++i) {
		vector<vector<vector<Formula*> > > emptyCache;
		m_cachedCellNodes.push_back(emptyCache);
		m_cachedHiddenNodes.push_back(emptyCache);
	}
	cerr << "Initialized LSTM. # parameters = " << model.getNumParams() << endl;
}

int DeepBidirectionalLSTM::getTotalHiddenVectorSize() { return m_cachedLayerWidthSum; }

Formula* DeepBidirectionalLSTM::createGateAtTimeAtDepth(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, unsigned int i, int t,
		unsigned int depth, const vector<vector<vector<PARAM_ID> > >& weightsObsGate, const vector<vector<vector<PARAM_ID> > >& weightsHiddenGate,
		const vector<vector<PARAM_ID> >& weightsCellGate, const vector<vector<PARAM_ID> >& gateBias,
		const vector<vector<vector<PARAM_ID> > >& weightsNextObsGate,
		bool outputGate,
		double inputDropoutRate, double hiddenDropoutRate, int terminationTime) {
	// e.g., equation for input gate is below:
	// \lambda is hidden layer width, d is input dimension
	// i_t = sigmoid(W_{xi} x_t + W_{hi} h_{t-1} + W_{ci} c_{t-1} + b_i)
	// input^{(t)}_i = sigmoid(\sum_{j=1}^{d} W^{obsInput}_{j,i} x^{(t)}_j +
	//                         \sum_{j=1}^{\lambda} W^{hiddenInput}_{j,i} h^{(t-1)}_j +
	//                         W^{cellInput}_i c^{(t-1)}_i +
	//                         b^{input}_i)
	Formula* inputGate = model.getFormulaObject(LOGISTIC, 1);

	bool usePrevNext;
	int prevNextTime;
	setPrevNextTime(usePrevNext, prevNextTime, t, terminationTime, datum.m_sequenceLength);
	bool useLookaheadPrevNext = false;
	int lookaheadPrevNextTime;
	if (m_useLookahead && depth == 0) {
		setLookaheadPrevNextTime(useLookaheadPrevNext, lookaheadPrevNextTime, t, terminationTime, datum.m_sequenceLength);
	}

	int numTerms = m_verticalMatrixInputDimensions[depth] + 1;
	if (usePrevNext) numTerms += m_layerWidths[depth];
	if (useLookaheadPrevNext) numTerms += m_verticalMatrixInputDimensions[depth];
	if (outputGate || usePrevNext) ++numTerms;
	Formula* sumNode = model.getFormulaObject(PLUS, numTerms);

	for (unsigned int j = 0; j < m_verticalMatrixInputDimensions[depth]; ++j) {
		bool dropout = false;
		if (depth == 0) {
			if (inputDropoutRate > 0.0 && getRandomDouble() < inputDropoutRate) dropout = true;
		} else {
			if (hiddenDropoutRate > 0.0 && getRandomDouble() < hiddenDropoutRate) dropout = true;
		}
		if (!dropout) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(weightsObsGate.at(depth).at(j).at(i)));
			if (depth == 0) {
				prod->addChild(datum.m_vectFormulas.at(t).at(j));
			} else {
				prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, t,
					depth - 1, inputDropoutRate, hiddenDropoutRate, terminationTime));
			}
			sumNode->addChild(prod);
		}
	}
	if (useLookaheadPrevNext) {
		if (depth != 0) cerr << "Error: depth (" << depth << ") != 0" << endl;
		for (unsigned int j = 0; j < m_verticalMatrixInputDimensions[depth]; ++j) {
			bool dropout = false;
			if (inputDropoutRate > 0.0 && getRandomDouble() < inputDropoutRate) dropout = true;
			if (!dropout) {
				Formula* prod = model.getFormulaObject(TIMES, 2);
				prod->addChild(model.getCachedParameterFormula(weightsNextObsGate.at(depth).at(j).at(i)));
				prod->addChild(datum.m_vectFormulas.at(lookaheadPrevNextTime).at(j));
				sumNode->addChild(prod);
			}
		}
	}
	if (usePrevNext) {
		//int prevNextTime = t - 1;
		//if (!m_forward) prevNextTime = t + 1;
		for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(weightsHiddenGate.at(depth).at(j).at(i)));
			prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, prevNextTime, depth,
				inputDropoutRate, hiddenDropoutRate, terminationTime));
			sumNode->addChild(prod);
		}
	}
	if (outputGate || usePrevNext) {
		Formula* cellInputProd = model.getFormulaObject(TIMES, 2);
		cellInputProd->addChild(model.getCachedParameterFormula(weightsCellGate.at(depth).at(i)));
		if (outputGate) prevNextTime = t;
		//int prevNextTime = t;
		//if (!outputGate) {
		//	if (m_forward) {
		//		prevNextTime = t - 1;
		//	} else {
		//		prevNextTime = t + 1;
		//	}
		//}
		cellInputProd->addChild(createCellNodeAtTimeAtDepth(model, datum, sequenceIndex, i, prevNextTime,
			depth, inputDropoutRate, hiddenDropoutRate, terminationTime));
		sumNode->addChild(cellInputProd);
	}

	sumNode->addChild(model.getCachedParameterFormula(gateBias.at(depth).at(i)));
	inputGate->addChild(sumNode);
	return inputGate;
}

// prevNextTime is only set if usePrevNext is set to true.
void DeepBidirectionalLSTM::setPrevNextTime(bool& usePrevNext, int& prevNextTime, int t, int terminationTime, int numWords) const {
	usePrevNext = false;
	if ((m_forward && t > terminationTime) || (!m_forward && t < numWords - 1)) usePrevNext = true;
	if (usePrevNext) {
		if (m_forward) {
			prevNextTime = t - 1;
		} else {
			prevNextTime = t + 1;
		}
	}
}

void DeepBidirectionalLSTM::setLookaheadPrevNextTime(bool& useLookaheadPrevNext, int& lookaheadPrevNextTime, int t, int terminationTime, int numWords) const {
	useLookaheadPrevNext = false;
	if ((m_forward && t < numWords - 1) || (!m_forward && t > terminationTime)) useLookaheadPrevNext = true;
	if (useLookaheadPrevNext) {
		if (m_forward) {
			lookaheadPrevNextTime = t + 1;
		} else {
			lookaheadPrevNextTime = t - 1;
		}
	}
}

Formula* DeepBidirectionalLSTM::createCellNodeAtTimeAtDepth(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, unsigned int i,
		int t, unsigned int depth, double inputDropoutRate, double hiddenDropoutRate, int terminationTime) {

	if (m_cachedCellNodes.at(sequenceIndex).at(t - terminationTime).at(depth).at(i) != NULL) return m_cachedCellNodes.at(sequenceIndex).at(t - terminationTime).at(depth).at(i);
	bool usePrevNext;
	int prevNextTime;
	setPrevNextTime(usePrevNext, prevNextTime, t, terminationTime, datum.m_sequenceLength);
	//if ((m_forward && t > 0) || (!m_forward && t < datum.m_numWords - 1)) usePrevNext = true;
	//int prevNextTime = ((int)t) - 1;
	//if (!m_forward) prevNextTime = t + 1;

	Formula* outerSum = model.getFormulaObject(PLUS, 2);
	if (usePrevNext) {
		Formula* prevCellProd = model.getFormulaObject(TIMES, 2);
		prevCellProd->addChild(createCellNodeAtTimeAtDepth(model, datum, sequenceIndex, i, prevNextTime, depth,
			inputDropoutRate, hiddenDropoutRate, terminationTime));
		prevCellProd->addChild(createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
			m_weightsObsForget, m_weightsHiddenForget, m_weightsCellForget, m_forgetGateBias, m_weightsNextObsForget, false,
			inputDropoutRate, hiddenDropoutRate, terminationTime));
		outerSum->addChild(prevCellProd);
	}
	Formula* inputTerm = model.getFormulaObject(TIMES, 2);
	inputTerm->addChild(createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
		m_weightsObsInput, m_weightsHiddenInput, m_weightsCellInput, m_inputGateBias, m_weightsNextObsInput, false,
		inputDropoutRate, hiddenDropoutRate, terminationTime));

	Formula* tanhPart = model.getFormulaObject(TANH, 1);
	int numTerms = m_verticalMatrixInputDimensions[depth] + 1;
	if (usePrevNext) numTerms += m_layerWidths[depth];
	Formula* sumNode = model.getFormulaObject(PLUS, numTerms);
	for (unsigned int j = 0; j < m_verticalMatrixInputDimensions[depth]; ++j) {
		bool dropout = false;
		if (depth == 0) {
			if (inputDropoutRate > 0.0 && getRandomDouble() < inputDropoutRate) dropout = true;
		} else {
			if (hiddenDropoutRate > 0.0 && getRandomDouble() < hiddenDropoutRate) dropout = true;
		}
		if (!dropout) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(m_weightsObsCell.at(depth).at(j).at(i)));
			if (depth == 0) {
				prod->addChild(datum.m_vectFormulas.at(t).at(j));
			} else {
				prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, t,
					depth - 1, inputDropoutRate, hiddenDropoutRate, terminationTime));
			}
			sumNode->addChild(prod);
		}
	}
	if (usePrevNext) {
		for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			PARAM_ID currentWeightID = m_weightsHiddenCell.at(depth).at(j).at(i);
			prod->addChild(model.getCachedParameterFormula(currentWeightID));
			prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, prevNextTime, depth,
				inputDropoutRate, hiddenDropoutRate, terminationTime));
			sumNode->addChild(prod);
		}
	}
	sumNode->addChild(model.getCachedParameterFormula(m_cellBias.at(depth).at(i)));
	tanhPart->addChild(sumNode);
	inputTerm->addChild(tanhPart);
	outerSum->addChild(inputTerm);
	m_cachedCellNodes[sequenceIndex][t - terminationTime][depth][i] = outerSum;
	return outerSum;
}

Formula* DeepBidirectionalLSTM::createHiddenNodeAtTimeAtDepth(Model& model, SequenceDatum& datum, unsigned int sequenceIndex,
		unsigned int i, int t, unsigned int depth, double inputDropoutRate, double hiddenDropoutRate, int terminationTime) {
	//DeepBidirectionalLSTMParams& lstmParams = params.m_lstmParams[lstmIndex];
	if (m_cachedHiddenNodes.at(sequenceIndex).at(t - terminationTime).at(depth).at(i) != NULL) return m_cachedHiddenNodes.at(sequenceIndex).at(t - terminationTime).at(depth).at(i);

	Formula* hiddenNodeProduct = model.getFormulaObject(TIMES, 2);
	hiddenNodeProduct->addChild(createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
		m_weightsObsOutput, m_weightsHiddenOutput, m_weightsCellOutput, m_outputGateBias, m_weightsNextObsOutput, true,
		inputDropoutRate, hiddenDropoutRate, terminationTime));
	Formula* tanhPart = model.getFormulaObject(TANH, 1);
	tanhPart->addChild(createCellNodeAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
		inputDropoutRate, hiddenDropoutRate, terminationTime));
	hiddenNodeProduct->addChild(tanhPart);
	m_cachedHiddenNodes[sequenceIndex][t - terminationTime][depth][i] = hiddenNodeProduct;
	return hiddenNodeProduct;
}

void DeepBidirectionalLSTM::getGateReport(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, int t, unsigned int depth,
	double& inputGateSum, double& forgetGateSum, double& outputGateSum, int terminationTime) {
	for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
		Formula* inputGate = createGateAtTimeAtDepth(model, datum, sequenceIndex, j, t, depth,
			m_weightsObsInput, m_weightsHiddenInput, m_weightsCellInput, m_inputGateBias, m_weightsNextObsInput, false,
			-1.0, -1.0, terminationTime);
		inputGateSum += inputGate->evaluate();

		Formula* forgetGate = createGateAtTimeAtDepth(model, datum, sequenceIndex, j, t, depth,
			m_weightsObsForget, m_weightsHiddenForget, m_weightsCellForget, m_forgetGateBias, m_weightsNextObsForget, false,
			-1.0, -1.0, terminationTime);
		forgetGateSum += forgetGate->evaluate();

		Formula* outputGate = createGateAtTimeAtDepth(model, datum, sequenceIndex, j, t, depth,
			m_weightsObsOutput, m_weightsHiddenOutput, m_weightsCellOutput, m_outputGateBias, m_weightsNextObsOutput, true,
			-1.0, -1.0, terminationTime);
		outputGateSum += outputGate->evaluate();
	}
}

void DeepBidirectionalLSTM::resetForNewExample(unsigned int sequenceIndex, int exampleLength) {
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedCellNodes[sequenceIndex]);
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedHiddenNodes[sequenceIndex]);
}
