#include "gru.h"

void DeepGRU::initializeParameters(Model& model, int dataInputDimension,
		int numLabels, bool forward, const string& layerWidths,
		unsigned int numSequencesPerDatum, bool initializeSoftmaxParameters,
		bool useUpdateGates, bool useResetGates) {
	// We initialize all weights to a random double between -0.01 and 0.01.
	double scalingFactor = 0.2;
	m_forward = forward;
	m_useUpdateGates = useUpdateGates;
	m_useResetGates = useResetGates;
	vector<string> layerWidthStrings;
	tokenize(layerWidths, ',', layerWidthStrings);
	if (layerWidths == "0") {
		m_numLayers = 0;
		cerr << "Error: 0 hidden layers desired for GRU. Must use at least 1 hidden layer." << endl;
	} else {
		m_numLayers = layerWidthStrings.size();
	}
	m_layerWidths.clear();
	m_cachedLayerWidthSum = 0;
	for (unsigned int i = 0; i < layerWidthStrings.size(); ++i) {
		int currentLayerWidth = atoi(layerWidthStrings[i].c_str());
		m_layerWidths.push_back(currentLayerWidth);
		m_cachedLayerWidthSum += currentLayerWidth;
	}
	if (m_numLayers != m_layerWidths.size()) {
		cerr << "Error: DeepGRUParams m_numLayers (" << m_numLayers << ") != m_layerWidths.size() (" << m_layerWidths.size() << ")!" << endl;
	}

	setVerticalMatrixDimensions(dataInputDimension, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, m_layerWidths);
	model.initializeParameterMatrices(m_weightsObsUpdate, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsObsReset, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsObsHidden, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);

	setHorizontalMatrixDimensions(dataInputDimension, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, m_layerWidths);
	model.initializeParameterMatrices(m_weightsHiddenUpdate, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsHiddenReset, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsHiddenHidden, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);

	// Parameters for gate and cell bias vectors.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	model.initializeParameterMatrixGivenRowLengths(m_updateGateBias, m_layerWidths, scalingFactor);
	model.initializeParameterMatrixGivenRowLengths(m_resetGateBias, m_layerWidths, scalingFactor);
	model.initializeParameterMatrixGivenRowLengths(m_hiddenBias, m_layerWidths, scalingFactor);

	if (initializeSoftmaxParameters) {
		setSoftmaxMatrixDimensions(numLabels, m_softmaxMatrixInputDimensions, m_softmaxMatrixOutputDimensions, m_layerWidths);
		model.initializeParameterMatrices(m_softmaxWeights, m_softmaxMatrixInputDimensions, m_softmaxMatrixOutputDimensions, scalingFactor);
		// Parameters for softmax bias vectors. Outermost vector ranges over hidden layers.
		// Dimensions of vector i: m_numLabels.
		model.initializeParameterMatrix(m_softmaxBias, m_numLayers, numLabels, scalingFactor);
	}

	// Set up Formula caches for the max number of sequences that will appear per datum.
	m_cachedGateProducts.reserve(numSequencesPerDatum);
	m_cachedHiddenNodes.reserve(numSequencesPerDatum);
	for (unsigned int i = 0; i < numSequencesPerDatum; ++i) {
		vector<vector<vector<Formula*> > > emptyCache;
		m_cachedGateProducts.push_back(emptyCache);
		m_cachedHiddenNodes.push_back(emptyCache);
	}

	cerr << "Initialized GRURNN. # parameters = " << model.getNumParams() << endl;
}

int DeepGRU::getTotalHiddenVectorSize() { return m_cachedLayerWidthSum; }

Formula* DeepGRU::createGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
		const vector<vector<vector<PARAM_ID> > >& weightsObsGate, const vector<vector<vector<PARAM_ID> > >& weightsHiddenGate,
		const vector<vector<PARAM_ID> >& gateBias, double inputDropoutRate, double hiddenDropoutRate) {
	//DeepGRUParams& gruParams = params.m_gruParams[gruIndex];
	Formula* inputGate = model.getFormulaObject(LOGISTIC, 1);
	bool usePrevNext = false;
	if ((m_forward && t > 0) || (!m_forward && t < datum.m_sequenceLength - 1)) usePrevNext = true;

	int numTerms = m_verticalMatrixInputDimensions[depth] + 1;
	if (usePrevNext) numTerms += m_layerWidths[depth];
	Formula* sumNode = model.getFormulaObject(PLUS, numTerms);

	for (unsigned int j = 0; j < m_verticalMatrixInputDimensions[depth]; ++j) {
		bool dropout = false;
		if (depth == 0) {
			if (inputDropoutRate > 0.0 && getRandomDouble() < inputDropoutRate) dropout = true;
		} else {
			if (hiddenDropoutRate > 0.0 && getRandomDouble() < hiddenDropoutRate) dropout = true;
		}
		if (!dropout) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(weightsObsGate.at(depth).at(j).at(i)));
			if (depth == 0) {
				prod->addChild(datum.m_vectFormulas.at(t).at(j));
			} else {
				prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, t,
					depth - 1, inputDropoutRate, hiddenDropoutRate));
			}
			sumNode->addChild(prod);
		}
	}
	//for (unsigned int j = 0; j < params.m_inputDimension; ++j) {
	//	Formula* prod = model.getFormulaObject(TIMES, 2);
	//	prod->addChild(model.getCachedParameterFormula(weightsObsGate.at(j).at(i)));
	//	prod->addChild(datum.m_vectFormulas.at(t).at(j));
	//	sumNode->addChild(prod);
	//}
	if (usePrevNext) {
		int prevNextTime = ((int)t) - 1;
		if (!m_forward) prevNextTime = t + 1;
		for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(weightsHiddenGate.at(depth).at(j).at(i)));
			prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, prevNextTime, depth, inputDropoutRate, hiddenDropoutRate));
			sumNode->addChild(prod);
		}
	}
	sumNode->addChild(model.getCachedParameterFormula(gateBias.at(depth).at(i)));
	inputGate->addChild(sumNode);
	return inputGate;
}

Formula* DeepGRU::createResetGateProduct(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int prevNextTime, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate) {
	if (m_cachedGateProducts.at(sequenceIndex).at(t).at(depth).at(i) != NULL) return m_cachedGateProducts.at(sequenceIndex).at(t).at(depth).at(i);
	Formula* gateProd = model.getFormulaObject(TIMES, 2);
	if (m_useResetGates) {
		gateProd->addChild(createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
			m_weightsObsReset, m_weightsHiddenReset, m_resetGateBias, inputDropoutRate, hiddenDropoutRate));
	}
	gateProd->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, i,
		prevNextTime, depth, inputDropoutRate, hiddenDropoutRate));

	m_cachedGateProducts[sequenceIndex][t][depth][i] = gateProd;
	return gateProd;
}

Formula* DeepGRU::createHiddenNodeAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth, double inputDropoutRate, double hiddenDropoutRate) {
	//DeepGRUParams& gruParams = params.m_gruParams[gruIndex];
	if (m_cachedHiddenNodes.at(sequenceIndex).at(t).at(depth).at(i) != NULL) return m_cachedHiddenNodes.at(sequenceIndex).at(t).at(depth).at(i);

	bool usePrevNext = false;
	if ((m_forward && t > 0) || (!m_forward && t < datum.m_sequenceLength - 1)) usePrevNext = true;
	int prevNextTime = ((int)t) - 1;
	if (!m_forward) prevNextTime = t + 1;

	// Create the formula for the update gate since it's used in both terms.
	Formula* updateGate = createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
		m_weightsObsUpdate, m_weightsHiddenUpdate, m_updateGateBias, inputDropoutRate, hiddenDropoutRate);

	Formula* outerSum = model.getFormulaObject(PLUS, 2);
	if (usePrevNext) {
		Formula* prevCellProd = model.getFormulaObject(TIMES, 2);
		prevCellProd->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, i,
			prevNextTime, depth, inputDropoutRate, hiddenDropoutRate));
		if (m_useUpdateGates) {
			Formula* oneMinusUpdateGate = model.getFormulaObject(MINUS, 2);
			oneMinusUpdateGate->addChild(model.getConstantFormulaObject(1.0));
			oneMinusUpdateGate->addChild(updateGate);
			prevCellProd->addChild(oneMinusUpdateGate);
		}
		outerSum->addChild(prevCellProd);
	}
	Formula* inputTerm = model.getFormulaObject(TIMES, 2);
	if (m_useUpdateGates) inputTerm->addChild(updateGate);

	// Next, build term for h tilde.
	int numTerms = m_verticalMatrixInputDimensions[depth] + 1;
	if (usePrevNext) numTerms += m_layerWidths[depth];
	Formula* sumNode = model.getFormulaObject(PLUS, numTerms);

	for (unsigned int j = 0; j < m_verticalMatrixInputDimensions[depth]; ++j) {
		bool dropout = false;
		if (depth == 0) {
			if (inputDropoutRate > 0.0 && getRandomDouble() < inputDropoutRate) dropout = true;
		} else {
			if (hiddenDropoutRate > 0.0 && getRandomDouble() < hiddenDropoutRate) dropout = true;
		}
		if (!dropout) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(m_weightsObsHidden.at(depth).at(j).at(i)));
			if (depth == 0) {
				prod->addChild(datum.m_vectFormulas.at(t).at(j));
			} else {
				prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, t,
					depth - 1, inputDropoutRate, hiddenDropoutRate));
			}
			sumNode->addChild(prod);
		}
	}
	//int numTerms = params.m_inputDimension + 1;
	//if (usePrevNext) numTerms += gruParams.m_layerWidth;
	//Formula* sumNode = model.getFormulaObject(PLUS, numTerms);
	//for (unsigned int j = 0; j < params.m_inputDimension; ++j) {
	//	Formula* prod = model.getFormulaObject(TIMES, 2);
		//PARAM_ID currentWeightID = gruParams.m_weightsObsHidden.at(depth).at(j).at(i);
		//prod->addChild(model.getCachedParameterFormula(currentWeightID));
		//prod->addChild(datum.m_vectFormulas.at(t).at(j));
		//sumNode->addChild(prod);
	//}
	if (usePrevNext) {
		for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(m_weightsHiddenHidden.at(depth).at(j).at(i)));
			Formula* gateProd = createResetGateProduct(model, datum, sequenceIndex, j, t, prevNextTime, depth,
				inputDropoutRate, hiddenDropoutRate);

			//model.getFormulaObject(TIMES, 2);
			// Multiply by the reset gate for the current dimension.
			//gateProd->addChild(createGateAtTimeAtDepth(model, datum, j, t, depth,
			//	m_weightsObsReset, m_weightsHiddenReset, m_resetGateBias, doDropout, dropoutRate));
			//gateProd->addChild(createHiddenNodeAtTimeAtDepth(model, datum, j,
			//	prevNextTime, depth, doDropout, dropoutRate));
			prod->addChild(gateProd);
			sumNode->addChild(prod);
		}
	}

	sumNode->addChild(model.getCachedParameterFormula(m_hiddenBias.at(depth).at(i)));
	Formula* tanhPart = model.getFormulaObject(TANH, 1);
	tanhPart->addChild(sumNode);
	inputTerm->addChild(tanhPart);
	outerSum->addChild(inputTerm);
	m_cachedHiddenNodes[sequenceIndex][t][depth][i] = outerSum;
	return outerSum;
}

Formula* DeepGRU::createResetGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
	unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
	double inputDropoutRate, double hiddenDropoutRate) {
	return createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
		m_weightsObsReset, m_weightsHiddenReset, m_resetGateBias, inputDropoutRate, hiddenDropoutRate);
}

Formula* DeepGRU::createUpdateGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
	unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
	double inputDropoutRate, double hiddenDropoutRate) {
	return createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
		m_weightsObsUpdate, m_weightsHiddenUpdate, m_updateGateBias, inputDropoutRate, hiddenDropoutRate);
}

void DeepGRU::getGateReport(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, int t, unsigned int depth,
	double& updateGateSum, double& resetGateSum) {

	for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
		if (m_useUpdateGates) {
			Formula* updateGate = createUpdateGateAtTimeAtDepth(model, datum, sequenceIndex, j, t, depth, -1.0, -1.0);
			updateGateSum += updateGate->evaluate();
		}
		if (m_useResetGates) {
			Formula* resetGate = createResetGateAtTimeAtDepth(model, datum, sequenceIndex, j, t, depth, -1.0, -1.0);
			resetGateSum += resetGate->evaluate();
		}
	}
}

void DeepGRU::resetForNewExample(unsigned int sequenceIndex, int exampleLength) {
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedGateProducts[sequenceIndex]);
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedHiddenNodes[sequenceIndex]);
}
