#ifndef GRU_H_
#define GRU_H_

#include "../classifier_utils.h"
#include "../../fdag/formula.h"
#include "../../fdag/model.h"

class DeepGRU {
public:
	void initializeParameters(Model& model, int dataInputDimension, int numLabels, bool forward, const string& layerWidths,
		unsigned int numSequencesPerDatum, bool initializeSoftmaxParameters, bool useUpdateGates, bool useResetGates);
	int getTotalHiddenVectorSize();
	Formula* createHiddenNodeAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth, double inputDropoutRate, double hiddenDropoutRate);
	Formula* createResetGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate);
	Formula* createUpdateGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate);
	void getGateReport(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, int t, unsigned int depth,
		double& updateGateSum, double& resetGateSum);
	void resetForNewExample(unsigned int sequenceIndex, int exampleLength);

	// If true, this deep GRU RNN goes from left to right in the input sequence.
	bool m_forward;
	bool m_useUpdateGates;
	bool m_useResetGates;
	// The number of hidden layers.
	int m_numLayers;
	// Vector of hidden layer widths.
	vector<int> m_layerWidths;
	int m_cachedLayerWidthSum;
	// Parameters for softmax weight matrices. Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_numLabels.
	vector<vector<vector<PARAM_ID> > > m_softmaxWeights;
	// Parameters for softmax bias vectors. Outermost vector ranges over hidden layers.
	// Dimensions of vector i: m_numLabels.
	vector<vector<PARAM_ID> > m_softmaxBias;

private:
	Formula* createGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
		const vector<vector<vector<PARAM_ID> > >& weightsObsGate,
		const vector<vector<vector<PARAM_ID> > >& weightsHiddenGate,
		const vector<vector<PARAM_ID> >& gateBias, double inputDropoutRate, double hiddenDropoutRate);
	Formula* createResetGateProduct(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int prevNextTime, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate);

	vector<int> m_verticalMatrixInputDimensions;
	vector<int> m_verticalMatrixOutputDimensions;
	vector<int> m_horizontalMatrixInputDimensions;
	vector<int> m_horizontalMatrixOutputDimensions;
	vector<int> m_softmaxMatrixInputDimensions;
	vector<int> m_softmaxMatrixOutputDimensions;

	// Parameters for observation-gate matrices; more generally, these matrices connect variables vertically.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix 0: m_inputDimension x m_layerWidths[0] (because they connect the input to the first hidden layer).
	// Dimensions of matrix i > 0: m_layerWidths[i-1] x m_layerWidths[i].
	vector<vector<vector<PARAM_ID> > > m_weightsObsUpdate, m_weightsObsReset, m_weightsObsHidden;
	// Parameters for hidden-gate matrices; more generally, these matrices connect variables horizontally.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_layerWidths[i].
	vector<vector<vector<PARAM_ID> > > m_weightsHiddenUpdate, m_weightsHiddenReset, m_weightsHiddenHidden;

	// Length of cell-gate weight vectors is m_layerWidth.
	//vector<PARAM_ID> m_weightsCellInput, m_weightsCellForget, m_weightsCellOutput;

	// Parameters for gate and cell bias vectors.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	vector<vector<PARAM_ID> > m_updateGateBias, m_resetGateBias, m_hiddenBias;

	// Pointers to Formulas for cached nodes for current example, which are NULL if unset (or dropped out).
	// Outermost vector ranges over time (and so has size datum.m_words.size()), next ranges over hidden layer, and innermost ranges over units in hidden layer.
	// Dimensions of matrix i: m_numLayers x m_layerWidths[i].
	vector<vector<vector<vector<Formula*> > > > m_cachedHiddenNodes;
	vector<vector<vector<vector<Formula*> > > > m_cachedGateProducts;
};

#endif
