#include "sru.h"

void DeepSRU::initializeParameters(Model& model, int dataInputDimension, int numLabels, bool forward, const string& layerWidths,
		unsigned int numSequencesPerDatum, bool initializeSoftmaxParameters, bool useLookahead, bool oneDimGate) {
	// We initialize all weights to a random double between -0.01 and 0.01.
	double scalingFactor = 0.2;
	m_forward = forward;
	m_oneDimGate = oneDimGate;
	m_useLookahead = useLookahead;

	vector<string> layerWidthStrings;
	tokenize(layerWidths, ',', layerWidthStrings);
	if (layerWidths == "0") {
		m_numLayers = 0;
		cerr << "Error: 0 hidden layers desired for LSTM. Must use at least 1 hidden layer." << endl;
	} else {
		m_numLayers = layerWidthStrings.size();
	}
	m_layerWidths.clear();
	m_cachedLayerWidthSum = 0;
	for (unsigned int i = 0; i < layerWidthStrings.size(); ++i) {
		int currentLayerWidth = atoi(layerWidthStrings[i].c_str());
		m_layerWidths.push_back(currentLayerWidth);
		m_cachedLayerWidthSum += currentLayerWidth;
	}
	if (m_numLayers != m_layerWidths.size()) {
		cerr << "Error: DeepSRUParams m_numLayers (" << m_numLayers << ") != m_layerWidths.size() (" << m_layerWidths.size() << ")!" << endl;
	}

	// Parameters for observation-gate matrices; more generally, these matrices connect variables vertically.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix 0: m_inputDimension x m_layerWidths[0] (because they connect the input to the first hidden layer).
	// Dimensions of matrix i > 0: m_layerWidths[i-1] x m_layerWidths[i].
	setVerticalMatrixDimensions(dataInputDimension, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, m_layerWidths);
	vector<int> allOnes(m_layerWidths.size(), 1);
	if (m_oneDimGate) {
		//model.initializeParameterMatrixGivenRowLengths(vector<vector<PARAM_ID> >& m, vector<int>& numRows, scalingFactor);
		model.initializeParameterMatrices(m_weightsObsReduce, m_verticalMatrixInputDimensions, allOnes, scalingFactor);
	} else {
		model.initializeParameterMatrices(m_weightsObsReduce, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	}
	model.initializeParameterMatrices(m_weightsObsStack, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	if (m_useLookahead) {
		model.initializeParameterMatrices(m_weightsNextObsReduce, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
		//model.initializeParameterMatrices(m_weightsNextObsStack, m_verticalMatrixInputDimensions, m_verticalMatrixOutputDimensions, scalingFactor);
	}
	// Parameters for hidden-gate matrices; more generally, these matrices connect variables horizontally.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_layerWidths[i].
	setHorizontalMatrixDimensions(dataInputDimension, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, m_layerWidths);

	if (m_oneDimGate) {
		model.initializeParameterMatrices(m_weightsStackReduce, m_horizontalMatrixInputDimensions, allOnes, scalingFactor);
		model.initializeParameterMatrices(m_weightsHiddenReduce, m_horizontalMatrixInputDimensions, allOnes, scalingFactor);
	} else {
		model.initializeParameterMatrices(m_weightsStackReduce, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
		model.initializeParameterMatrices(m_weightsHiddenReduce, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
	}
	model.initializeParameterMatrices(m_weightsStackStack, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsStackHidden, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);
	model.initializeParameterMatrices(m_weightsHiddenHidden, m_horizontalMatrixInputDimensions, m_horizontalMatrixOutputDimensions, scalingFactor);

	// Parameters for diagonal matrices connecting cells and gates.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	//model.initializeParameterMatrixGivenRowLengths(m_weightsStackReduce, m_layerWidths, scalingFactor);
	//model.initializeParameterMatrixGivenRowLengths(m_weightsCellForget, m_layerWidths, scalingFactor);
	//model.initializeParameterMatrixGivenRowLengths(m_weightsCellOutput, m_layerWidths, scalingFactor);

	// Parameters for bias vectors.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	if (m_oneDimGate) {
		model.initializeParameterMatrixGivenRowLengths(m_reduceGateBias, allOnes, scalingFactor);
	} else {
		model.initializeParameterMatrixGivenRowLengths(m_reduceGateBias, m_layerWidths, scalingFactor);
	}
	model.initializeParameterMatrixGivenRowLengths(m_stackBias, m_layerWidths, scalingFactor);
	model.initializeParameterMatrixGivenRowLengths(m_hiddenBias, m_layerWidths, scalingFactor);

	if (initializeSoftmaxParameters) {
		// Parameters for softmax weight matrices. Outermost vector ranges over hidden layers.
		// Dimensions of matrix i: m_layerWidths[i] x m_numLabels.
		setSoftmaxMatrixDimensions(numLabels, m_softmaxMatrixInputDimensions, m_softmaxMatrixOutputDimensions, m_layerWidths);
		model.initializeParameterMatrices(m_softmaxWeights, m_softmaxMatrixInputDimensions, m_softmaxMatrixOutputDimensions, scalingFactor);

		// Parameters for softmax bias vectors. Outermost vector ranges over hidden layers.
		// Dimensions of vector i: m_numLabels.
		model.initializeParameterMatrix(m_softmaxBias, m_numLayers, numLabels, scalingFactor);
	}

	// Set up Formula caches for the max number of sequences that will appear per datum.
	m_cachedStackNodes.reserve(numSequencesPerDatum);
	m_cachedHiddenNodes.reserve(numSequencesPerDatum);
	m_cachedReduceGates.reserve(numSequencesPerDatum);
	m_cachedInverseReduceGateProducts.reserve(numSequencesPerDatum);
	m_cachedReduceGateProducts.reserve(numSequencesPerDatum);
	for (unsigned int i = 0; i < numSequencesPerDatum; ++i) {
		vector<vector<vector<Formula*> > > emptyCache;
		m_cachedStackNodes.push_back(emptyCache);
		m_cachedHiddenNodes.push_back(emptyCache);
		m_cachedReduceGates.push_back(emptyCache);
		m_cachedInverseReduceGateProducts.push_back(emptyCache);
		m_cachedReduceGateProducts.push_back(emptyCache);
	}
	cerr << "Initialized SRU. # parameters = " << model.getNumParams() << endl;
}

int DeepSRU::getTotalHiddenVectorSize() { return m_cachedLayerWidthSum; }

Formula* DeepSRU::createGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, int t, unsigned int depth,
		const vector<vector<vector<PARAM_ID> > >& weightsObsGate,
		const vector<vector<vector<PARAM_ID> > >& weightsHiddenGate,
		const vector<vector<vector<PARAM_ID> > >& weightsStackGate,
		const vector<vector<PARAM_ID> >& gateBias,
		double inputDropoutRate, double hiddenDropoutRate, int terminationTime) {
	// e.g., equation for reduce gate is below:
	// r_t = sigmoid(W_{xr} x_t + W_{hr} h_{t-1} + W_{sr} s_t + b_r)
	if (m_cachedReduceGates.at(sequenceIndex).at(t - terminationTime).at(depth).at(i) != NULL)
		return m_cachedReduceGates.at(sequenceIndex).at(t - terminationTime).at(depth).at(i);

	bool usePrevNext;
	int prevNextTime;
	setPrevNextTime(usePrevNext, prevNextTime, t, terminationTime, datum.m_sequenceLength);
	bool useLookaheadPrevNext = false;
	int lookaheadPrevNextTime;
	if (m_useLookahead && depth == 0) {
		setLookaheadPrevNextTime(useLookaheadPrevNext, lookaheadPrevNextTime, t, terminationTime, datum.m_sequenceLength);
	}

	int numTerms = m_verticalMatrixInputDimensions[depth] + m_layerWidths[depth] + 1;
	if (usePrevNext) numTerms += m_layerWidths[depth];
	if (useLookaheadPrevNext) numTerms += m_verticalMatrixInputDimensions[depth];
	Formula* sumNode = model.getFormulaObject(PLUS, numTerms);

	for (unsigned int j = 0; j < m_verticalMatrixInputDimensions[depth]; ++j) {
		bool dropout = false;
		if (depth == 0) {
			if (inputDropoutRate > 0.0 && getRandomDouble() < inputDropoutRate) dropout = true;
		} else {
			if (hiddenDropoutRate > 0.0 && getRandomDouble() < hiddenDropoutRate) dropout = true;
		}
		if (!dropout) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(weightsObsGate.at(depth).at(j).at(i)));
			if (depth == 0) {
				prod->addChild(datum.m_vectFormulas.at(t).at(j));
			} else {
				prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, t,
					depth - 1, inputDropoutRate, hiddenDropoutRate, terminationTime));
			}
			sumNode->addChild(prod);
		}
	}
	// Add terms for the next observation if using lookahead.
	if (useLookaheadPrevNext) {
		if (depth != 0) cerr << "Error: depth (" << depth << ") != 0" << endl;
		for (unsigned int j = 0; j < m_verticalMatrixInputDimensions[depth]; ++j) {
			bool dropout = false;
			if (inputDropoutRate > 0.0 && getRandomDouble() < inputDropoutRate) dropout = true;
			if (!dropout) {
				Formula* prod = model.getFormulaObject(TIMES, 2);
				prod->addChild(model.getCachedParameterFormula(m_weightsNextObsReduce.at(depth).at(j).at(i)));
				prod->addChild(datum.m_vectFormulas.at(lookaheadPrevNextTime).at(j));
				sumNode->addChild(prod);
			}
		}
	}

	// Add the terms for the adjacent hidden vector in the sequence.
	if (usePrevNext) {
		for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(weightsHiddenGate.at(depth).at(j).at(i)));
			prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, prevNextTime, depth,
				inputDropoutRate, hiddenDropoutRate, terminationTime));
			sumNode->addChild(prod);
		}
	}
	// Add the terms for the stack vector at position t in the sequence.
	for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
		Formula* prod = model.getFormulaObject(TIMES, 2);
		prod->addChild(model.getCachedParameterFormula(weightsStackGate.at(depth).at(j).at(i)));
		prod->addChild(createStackNodeAtTimeAtDepth(model, datum, sequenceIndex, j, t, depth,
			inputDropoutRate, hiddenDropoutRate, terminationTime));
		sumNode->addChild(prod);
	}

	sumNode->addChild(model.getCachedParameterFormula(gateBias.at(depth).at(i)));
	Formula* reduceGate = model.getFormulaObject(LOGISTIC, 1);
	reduceGate->addChild(sumNode);
	m_cachedReduceGates[sequenceIndex][t - terminationTime][depth][i] = reduceGate;
	return reduceGate;
}

Formula* DeepSRU::createReduceGateProduct(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate) {
	if (m_cachedReduceGateProducts.at(sequenceIndex).at(t).at(depth).at(i) != NULL)
		return m_cachedReduceGateProducts.at(sequenceIndex).at(t).at(depth).at(i);
	Formula* gateFormula = createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
		m_weightsObsReduce, m_weightsHiddenReduce, m_weightsStackReduce, m_reduceGateBias,
		inputDropoutRate, hiddenDropoutRate);
	Formula* gateProd = model.getFormulaObject(TIMES, 2);
	gateProd->addChild(gateFormula);
	gateProd->addChild(createStackNodeAtTimeAtDepth(model, datum, sequenceIndex, i,
		t, depth, inputDropoutRate, hiddenDropoutRate));

	m_cachedReduceGateProducts[sequenceIndex][t][depth][i] = gateProd;
	return gateProd;
}

Formula* DeepSRU::createInverseReduceGateProduct(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate) {
	if (m_cachedInverseReduceGateProducts.at(sequenceIndex).at(t).at(depth).at(i) != NULL)
		return m_cachedInverseReduceGateProducts.at(sequenceIndex).at(t).at(depth).at(i);
	Formula* gateFormula = createGateAtTimeAtDepth(model, datum, sequenceIndex, i, t, depth,
		m_weightsObsReduce, m_weightsHiddenReduce, m_weightsStackReduce, m_reduceGateBias,
		inputDropoutRate, hiddenDropoutRate);
	Formula* oneMinusGate = model.getFormulaObject(MINUS, 2);
	oneMinusGate->addChild(model.getConstantFormulaObject(1.0));
	oneMinusGate->addChild(gateFormula);
	Formula* gateProd = model.getFormulaObject(TIMES, 2);
	gateProd->addChild(oneMinusGate);
	gateProd->addChild(createStackNodeAtTimeAtDepth(model, datum, sequenceIndex, i,
		t, depth, inputDropoutRate, hiddenDropoutRate));

	m_cachedInverseReduceGateProducts[sequenceIndex][t][depth][i] = gateProd;
	return gateProd;
}

// prevNextTime is only set if usePrevNext is set to true.
void DeepSRU::setPrevNextTime(bool& usePrevNext, int& prevNextTime, int t, int terminationTime, int numWords) const {
	usePrevNext = false;
	if ((m_forward && t > terminationTime) || (!m_forward && t < numWords - 1)) usePrevNext = true;
	if (usePrevNext) {
		if (m_forward) {
			prevNextTime = t - 1;
		} else {
			prevNextTime = t + 1;
		}
	}
}

void DeepSRU::setLookaheadPrevNextTime(bool& useLookaheadPrevNext, int& lookaheadPrevNextTime, int t, int terminationTime, int numWords) const {
	useLookaheadPrevNext = false;
	if ((m_forward && t < numWords - 1) || (!m_forward && t > terminationTime)) useLookaheadPrevNext = true;
	if (useLookaheadPrevNext) {
		if (m_forward) {
			lookaheadPrevNextTime = t + 1;
		} else {
			lookaheadPrevNextTime = t - 1;
		}
	}
}

Formula* DeepSRU::createStackNodeAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate, int terminationTime) {

	if (m_cachedStackNodes.at(sequenceIndex).at(t - terminationTime).at(depth).at(i) != NULL) return m_cachedStackNodes.at(sequenceIndex).at(t - terminationTime).at(depth).at(i);
	bool usePrevNext;
	int prevNextTime;
	setPrevNextTime(usePrevNext, prevNextTime, t, terminationTime, datum.m_sequenceLength);

	// Build sum term.
	int numTerms = m_verticalMatrixInputDimensions[depth] + 1;
	if (usePrevNext) numTerms += m_layerWidths[depth];
	Formula* sumNode = model.getFormulaObject(PLUS, numTerms);

	if (usePrevNext) {
		for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(m_weightsStackStack.at(depth).at(j).at(i)));
			int gateIndexToUse = j;
			if (m_oneDimGate) gateIndexToUse = 0;
			Formula* gateProd = createInverseReduceGateProduct(model, datum, sequenceIndex, gateIndexToUse,
				prevNextTime, depth, inputDropoutRate, hiddenDropoutRate);

			prod->addChild(gateProd);
			sumNode->addChild(prod);
		}
	}
	for (unsigned int j = 0; j < m_verticalMatrixInputDimensions[depth]; ++j) {
		bool dropout = false;
		if (depth == 0) {
			if (inputDropoutRate > 0.0 && getRandomDouble() < inputDropoutRate) dropout = true;
		} else {
			if (hiddenDropoutRate > 0.0 && getRandomDouble() < hiddenDropoutRate) dropout = true;
		}
		if (!dropout) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(m_weightsObsStack.at(depth).at(j).at(i)));
			if (depth == 0) {
				prod->addChild(datum.m_vectFormulas.at(t).at(j));
			} else {
				prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, t,
					depth - 1, inputDropoutRate, hiddenDropoutRate));
			}
			sumNode->addChild(prod);
		}
	}

	sumNode->addChild(model.getCachedParameterFormula(m_stackBias.at(depth).at(i)));
	Formula* tanhSum = model.getFormulaObject(TANH, 1);
	tanhSum->addChild(sumNode);
	m_cachedStackNodes[sequenceIndex][t - terminationTime][depth][i] = tanhSum;
	return tanhSum;
}

Formula* DeepSRU::createHiddenNodeAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate, int terminationTime) {

	if (m_cachedHiddenNodes.at(sequenceIndex).at(t - terminationTime).at(depth).at(i) != NULL) return m_cachedStackNodes.at(sequenceIndex).at(t - terminationTime).at(depth).at(i);
	bool usePrevNext;
	int prevNextTime;
	setPrevNextTime(usePrevNext, prevNextTime, t, terminationTime, datum.m_sequenceLength);

	// Build sum term.
	int numTerms = m_layerWidths[depth] + 1;
	if (usePrevNext) numTerms += m_layerWidths[depth];
	Formula* sumNode = model.getFormulaObject(PLUS, numTerms);

	for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
		Formula* prod = model.getFormulaObject(TIMES, 2);
		prod->addChild(model.getCachedParameterFormula(m_weightsStackHidden.at(depth).at(j).at(i)));
		int gateIndexToUse = j;
		if (m_oneDimGate) gateIndexToUse = 0;
		Formula* gateProd = createReduceGateProduct(model, datum, sequenceIndex, gateIndexToUse,
			t, depth, inputDropoutRate, hiddenDropoutRate);
		prod->addChild(gateProd);
		sumNode->addChild(prod);
	}
	if (usePrevNext) {
		for (unsigned int j = 0; j < m_layerWidths[depth]; ++j) {
			Formula* prod = model.getFormulaObject(TIMES, 2);
			prod->addChild(model.getCachedParameterFormula(m_weightsHiddenHidden.at(depth).at(j).at(i)));
			prod->addChild(createHiddenNodeAtTimeAtDepth(model, datum, sequenceIndex, j, prevNextTime,
				depth, inputDropoutRate, hiddenDropoutRate));
			sumNode->addChild(prod);
		}
	}
	sumNode->addChild(model.getCachedParameterFormula(m_hiddenBias.at(depth).at(i)));
	Formula* tanhSum = model.getFormulaObject(TANH, 1);
	tanhSum->addChild(sumNode);
	m_cachedHiddenNodes[sequenceIndex][t - terminationTime][depth][i] = tanhSum;
	return tanhSum;
}

void DeepSRU::getGateReport(Model& model, SequenceDatum& datum,
	unsigned int sequenceIndex, int t, unsigned int depth,
	vector<double>& reduceGates, double& reduceGateSum, int terminationTime) {
	unsigned int endIndex = m_layerWidths[depth];
	if (m_oneDimGate) endIndex = 1;
	for (unsigned int j = 0; j < endIndex; ++j) {
		Formula* reduceGate = createGateAtTimeAtDepth(model, datum, sequenceIndex, j, t, depth,
			m_weightsObsReduce, m_weightsHiddenReduce, m_weightsStackReduce, m_reduceGateBias,
			-1.0, -1.0, terminationTime);
		double curr = reduceGate->evaluate();
		reduceGates.push_back(curr);
		reduceGateSum += curr;
	}
}

void DeepSRU::resetForNewExample(unsigned int sequenceIndex, int exampleLength) {
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedStackNodes[sequenceIndex]);
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedHiddenNodes[sequenceIndex]);
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedReduceGates[sequenceIndex]);
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedInverseReduceGateProducts[sequenceIndex]);
	resetDeepCachedNodes(exampleLength, m_layerWidths, m_cachedReduceGateProducts[sequenceIndex]);
}

