#ifndef SRU_H_
#define SRU_H_

#include "../classifier_utils.h"
#include "../../fdag/formula.h"
#include "../../fdag/model.h"

// Implementation of a deep shift-reduce-unit (SRU) recurrent neural network.
// This implementation permits multiple sequences to be modeled with the same
// parameters in the same datum.
// There is a single set of parameter matrices for all sequences, but the class
// keeps distinct matrices
// of cached nodes for the different sequences. The functions to build Formulas
// take the index of the sequence
// of interest so that they know which cache to use.
class DeepSRU {
public:
	void initializeParameters(Model& model, int dataInputDimension, int numLabels,
		bool forward, const string& layerWidths,
		unsigned int numSequencesPerDatum, bool initializeSoftmaxParameters, bool useLookahead, bool oneDimGate);
	int getTotalHiddenVectorSize();
	Formula* createHiddenNodeAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i,
		int t, unsigned int depth, double inputDropoutRate, double hiddenDropoutRate,
		int terminationTime = 0);
	void getGateReport(Model& model, SequenceDatum& datum, unsigned int sequenceIndex,
		int t, unsigned int depth, vector<double>& reduceGates, double& reduceGateSum, int terminationTime = 0);
	void resetForNewExample(unsigned int sequenceIndex, int exampleLength);

	// If true, this deep SRU goes from left to right in the input sequence.
	bool m_forward;
	// If true, this deep SRU uses a single dimension for the reduce gate.
	bool m_oneDimGate;
	// The number of hidden layers.
	int m_numLayers;
	// Vector of hidden layer widths.
	vector<int> m_layerWidths;
	int m_cachedLayerWidthSum;
	// Parameters for softmax weight matrices. Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_numLabels.
	vector<vector<vector<PARAM_ID> > > m_softmaxWeights;
	// Parameters for softmax bias vectors. Outermost vector ranges over hidden layers.
	// Dimensions of vector i: m_numLabels.
	vector<vector<PARAM_ID> > m_softmaxBias;

private:
	Formula* createGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, int t, unsigned int depth,
		const vector<vector<vector<PARAM_ID> > >& weightsObsGate,
		const vector<vector<vector<PARAM_ID> > >& weightsHiddenGate,
		const vector<vector<vector<PARAM_ID> > >& weightsStackGate,
		const vector<vector<PARAM_ID> >& gateBias,
		double inputDropoutRate, double hiddenDropoutRate, int terminationTime = 0);
	Formula* createMeanGateAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, int t, unsigned int depth,
		const vector<vector<vector<PARAM_ID> > >& weightsObsGate,
		const vector<vector<vector<PARAM_ID> > >& weightsHiddenGate,
		const vector<vector<vector<PARAM_ID> > >& weightsStackGate,
		const vector<vector<PARAM_ID> >& gateBias,
		double inputDropoutRate, double hiddenDropoutRate, int terminationTime);

	void setPrevNextTime(bool& usePrevNext, int& prevNextTime, int t, int terminationTime, int numWords) const;
	void setLookaheadPrevNextTime(bool& useLookaheadPrevNext, int& lookaheadPrevNextTime, int t, int terminationTime, int numWords) const;
	Formula* createReduceGateProduct(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate);
	Formula* createInverseReduceGateProduct(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, unsigned int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate);

	Formula* createStackNodeAtTimeAtDepth(Model& model, SequenceDatum& datum,
		unsigned int sequenceIndex, unsigned int i, int t, unsigned int depth,
		double inputDropoutRate, double hiddenDropoutRate, int terminationTime = 0);

	vector<int> m_verticalMatrixInputDimensions;
	vector<int> m_verticalMatrixOutputDimensions;
	vector<int> m_horizontalMatrixInputDimensions;
	vector<int> m_horizontalMatrixOutputDimensions;
	vector<int> m_softmaxMatrixInputDimensions;
	vector<int> m_softmaxMatrixOutputDimensions;
	// Parameters for observation-gate matrices;
	// more generally, these matrices connect variables vertically.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix 0: m_inputDimension x m_layerWidths[0] (because they connect the input to the first hidden layer).
	// Dimensions of matrix i > 0: m_layerWidths[i-1] x m_layerWidths[i].
	vector<vector<vector<PARAM_ID> > > m_weightsObsReduce, m_weightsObsStack;
	vector<vector<vector<PARAM_ID> > > m_weightsNextObsReduce;//, m_weightsNextObsStack;
	// Parameters for hidden-gate matrices;
	// more generally, these matrices connect variables horizontally.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_layerWidths[i].
	vector<vector<vector<PARAM_ID> > > m_weightsStackReduce, m_weightsStackStack,
		m_weightsStackHidden, m_weightsHiddenReduce, m_weightsHiddenHidden;

	// Parameters for diagonal matrices connecting stack and reduce gate.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	//vector<vector<PARAM_ID> > m_weightsStackReduce;

	// Parameters for bias vectors.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	vector<vector<PARAM_ID> > m_reduceGateBias, m_stackBias, m_hiddenBias;

	// Pointers to Formulas for cached nodes for current set of sequences,
	// which are NULL if unset (or dropped out).
	// Outermost vector ranges over sequences,
	// then next ranges over time (and so has size datum.m_words.size()),
	// then next ranges over hidden layer, and innermost ranges over units in hidden layer.
	// Dimensions of matrix i: m_numLayers x m_layerWidths[i].
	vector<vector<vector<vector<Formula*> > > > m_cachedStackNodes;
	vector<vector<vector<vector<Formula*> > > > m_cachedHiddenNodes;
	vector<vector<vector<vector<Formula*> > > > m_cachedReduceGates;
	vector<vector<vector<vector<Formula*> > > > m_cachedInverseReduceGateProducts;
	vector<vector<vector<vector<Formula*> > > > m_cachedReduceGateProducts;

	bool m_useLookahead;
};

#endif
