#include <iostream>
#include <stdlib.h>
#include <sys/time.h>
#include "../utils/cmdparams.h"
#include "../fdag/model.h"
#include "classifier_utils.h"

struct ModelParams {
	// The input dimension is the number of features.
	unsigned int m_inputDimension;
	// We store separate vectors of the shared layer input and output dimensions for convenience.
	vector<unsigned int> m_sharedLayerInputSizes;
	vector<unsigned int> m_sharedLayerOutputSizes;
	// The number of labels.
	unsigned int m_numLabels;
	// The number of shared layers. Just a single int.
	unsigned int m_numSharedLayers;
	// The following boolean is false iff m_numSharedLayers is 0.
	bool m_usingSharedLayers;
	// The index to pass as the final layer index when recursively building the FDAGs for all shared layers.
	unsigned int m_finalSharedLayerIndex;

	// The weight matrices for the layers have the following sizes:
	// inputDimension x m_layerWidths[0]
	// m_layerWidths[0] x m_layerWidths[1]
	// ...
	// m_layerWidths[m_numLayers - 2] x m_layerWidths[m_numLayers - 1]
	// m_layerWidths[m_numLayers - 1] x m_numLabels   <-- this is for the final softmax layer
	vector<vector<vector<PARAM_ID> > > m_sharedLayerWeights;
	// The bias vectors for the layers have the following sizes:
	// m_layerWidths[0]
	// m_layerWidths[1]
	// ...
	// m_layerWidths[m_numLayers - 1]
	// m_numLabels   <-- this is for the final softmax layer
	vector<vector<PARAM_ID> > m_sharedLayerBias;


	// The number of class-separated layers; can be different for each label.
	vector<unsigned int> m_numSeparatedLayers;
	// The following boolean is false iff each entry in m_numSeparatedLayers is 0.
	bool m_usingSeparatedLayers;
	// The input sizes of the class-separated layers; outer vector ranges over labels, inner vector ranges over layers.
	vector< vector<unsigned int> > m_separatedLayerInputSizes;
	vector< vector<unsigned int> > m_separatedLayerOutputSizes;
	// The index to pass as the final layer index when recursively building the FDAGs for all class-separated layers.
	// It's a vector since it can be different for each class.
	vector<unsigned int> m_finalSeparatedLayerIndex;

	vector<vector<vector<vector<PARAM_ID> > > > m_separatedLayerWeights;
	vector<vector<vector<PARAM_ID> > > m_separatedLayerBias;
};

struct DatumFormulaCache {
	// Pointers to Formulas for cached nodes for shared layers for current example, which are NULL if unset (or dropped out).
	vector<vector<Formula*> > m_cachedSharedLayerNodes;

	// Pointers to Formulas for cached nodes for separated layers for current example, which are NULL if unset (or dropped out).
	// Outermost vector ranges over labels.
	vector<vector<vector<Formula*> > > m_cachedSeparatedLayerNodes;

	// Pointers to constant Formulas that represent the feature vector for the current example.
	vector<Formula*> m_features;

	// Pointers to parameter Formulas, in order to bypass the Model's parameter Formula cache.
	vector<Formula*> m_parameters;
};

// Parses sharedLayerWidths and separatedLayerWidths strings to obtain sequences of layer widths.
// This function assumes that params.m_inputDimension and params.m_numLabels have already been filled in.
void fillLayerWidths(ModelParams& params, const string& sharedLayerWidths, const string& separatedLayerWidths) {

	if (separatedLayerWidths == "0") {
		params.m_usingSeparatedLayers = false;
	} else {
		params.m_usingSeparatedLayers = true;
	}

	vector<string> sharedLayerWidthStrings;
	tokenize(sharedLayerWidths, ',', sharedLayerWidthStrings);
	if (sharedLayerWidths == "0") {
		params.m_numSharedLayers = 0;
		params.m_usingSharedLayers = false;
	} else {
		params.m_numSharedLayers = sharedLayerWidthStrings.size();
		params.m_usingSharedLayers = true;
	}

	// Fill the vectors of input and output layer widths.
	int lastSharedLayerOutputSize = -1;

	// If we're either using shared hidden layers or not using either shared or separated hidden layers,
	// make the first layer have input size equal to the input feature vector dimension.
	// Using neither type of hidden layer corresponds to logistic regression.
	if (params.m_usingSharedLayers || (params.m_usingSharedLayers == false && params.m_usingSeparatedLayers == false)) {
		params.m_sharedLayerInputSizes.push_back(params.m_inputDimension);
	}

	for (unsigned int i = 0; i < params.m_numSharedLayers; ++i) {
		int currentSharedLayerWidth = atoi(sharedLayerWidthStrings[i].c_str());
		bool addToInputSizes = true;
		// If we're on the final layer, only add the input layer size if we do not have separated layers to add later.
		if (i == (params.m_numSharedLayers - 1) && params.m_usingSeparatedLayers) {
			addToInputSizes = false;
		}
		if (addToInputSizes) params.m_sharedLayerInputSizes.push_back(currentSharedLayerWidth);

		params.m_sharedLayerOutputSizes.push_back(currentSharedLayerWidth);
		lastSharedLayerOutputSize = currentSharedLayerWidth;
	}
	if (params.m_usingSeparatedLayers) {
		// TODO what to do here?
	} else {
		// If we're not using any separated layers,
		params.m_sharedLayerOutputSizes.push_back(params.m_numLabels);
	}

	params.m_finalSharedLayerIndex = params.m_sharedLayerInputSizes.size();

	if (params.m_sharedLayerInputSizes.size() != params.m_sharedLayerOutputSizes.size()) {
		cerr << "Error: params.m_sharedLayerInputSizes.size() (" << params.m_sharedLayerInputSizes.size() << ")"
			 << " != params.m_sharedLayerOutputSizes.size() (" << params.m_sharedLayerOutputSizes.size() << ")" << endl;
	} else {
		for (unsigned int i = 0; i < params.m_sharedLayerInputSizes.size(); ++i) {
			cerr << "m_sharedLayerInputSizes[" << i << "] = " << params.m_sharedLayerInputSizes[i] << endl;
			cerr << "m_sharedLayerOutputSizes[" << i << "] = " << params.m_sharedLayerOutputSizes[i] << endl;
		}
	}
	if (params.m_sharedLayerInputSizes.size() != params.m_numSharedLayers + 1) {
		cerr << "Error: number of layer input sizes (" << params.m_sharedLayerInputSizes.size()
			 << ") not equal to 1 + number of layers (" << (params.m_numSharedLayers + 1) << ")" << endl;
	}
	if (params.m_sharedLayerOutputSizes.size() != params.m_numSharedLayers + 1) {
		cerr << "Error: number of layer output sizes (" << params.m_sharedLayerOutputSizes.size()
			 << ") not equal to 1 + number of layers (" << (params.m_numSharedLayers + 1) << ")" << endl;
	}

	// Tokenize separated layer width string around the colon.
	vector<string> classLayerWidthStrings;
	tokenize(separatedLayerWidths, ':', classLayerWidthStrings);
	if (separatedLayerWidths == "0") {
		params.m_numSeparatedLayers.clear();
		for (unsigned int i = 0; i < params.m_numLabels; ++i) {
			params.m_numSeparatedLayers.push_back(0);
		}
	} else {
		if (params.m_numLabels != classLayerWidthStrings.size()) {
			cerr << "Error: number of labels must equal number of separated layer widths! number of labels = " << params.m_numLabels
				 << ", number of separated layer widths = " << classLayerWidthStrings.size() << " (" << separatedLayerWidths << ")" << endl;
		} else {
			// Go through the layer width strings for the classes.
			params.m_numSeparatedLayers.clear();
			for (unsigned int i = 0; i < classLayerWidthStrings.size(); ++i) {
				// Tokenize the string for the current class around the comma to get individual layer widths.
				vector<string> layerWidthStrings;
				tokenize(classLayerWidthStrings[i], ',', layerWidthStrings);
				int currNumSeparatedLayers = layerWidthStrings.size();
				params.m_numSeparatedLayers.push_back(currNumSeparatedLayers);
				// Fill the vectors of input and output layer widths.
				if (lastSharedLayerOutputSize == -1) {
					// If there were no shared hidden layers,
					lastSharedLayerOutputSize = params.m_inputDimension;
				}
				vector<unsigned int> currInputSizes, currOutputSizes;
				currInputSizes.push_back(lastSharedLayerOutputSize);
				for (unsigned int j = 0; j < currNumSeparatedLayers; ++j) {
					int currentSeparatedLayerWidth = atoi(layerWidthStrings[j].c_str());
					currInputSizes.push_back(currentSeparatedLayerWidth);
					currOutputSizes.push_back(currentSeparatedLayerWidth);
				}
				// The final layer's output dimension is always 1, since the sequence of class-separated
				// hidden layers is simply computing the scalar score of each class.
				currOutputSizes.push_back(1);
				// We've finished building
				params.m_separatedLayerInputSizes.push_back(currInputSizes);
				params.m_separatedLayerOutputSizes.push_back(currOutputSizes);

				params.m_finalSeparatedLayerIndex.push_back(currInputSizes.size());

				//int classIndex = params.m_separatedLayerInputSizes.size() - 1;
				if (currInputSizes.size() != currNumSeparatedLayers + 1) {
					cerr << "Error: number of separated layer input sizes (" << currInputSizes.size()
						 << ") not equal to 1 + number of layers (" << (currNumSeparatedLayers + 1) << ")" << endl;
				}
				if (currOutputSizes.size() != currNumSeparatedLayers + 1) {
					cerr << "Error: number of separated layer output sizes (" << currOutputSizes.size()
						 << ") not equal to 1 + number of layers (" << (currNumSeparatedLayers + 1) << ")" << endl;
				}
			}

			for (unsigned int i = 0; i < params.m_separatedLayerInputSizes.size(); ++i) {
				for (unsigned int j = 0; j < params.m_separatedLayerInputSizes[i].size(); ++j) {
					cerr << "m_separatedLayerInputSizes[" << i << "][" << j << "] = " << params.m_separatedLayerInputSizes[i][j] << endl;
					cerr << "m_separatedLayerOutputSizes[" << i << "][" << j << "] = " << params.m_separatedLayerOutputSizes[i][j] << endl;
				}
			}
			cerr << "params.m_finalSharedLayerIndex = " << params.m_finalSharedLayerIndex << endl;
			for (unsigned int i = 0; i < params.m_finalSeparatedLayerIndex.size(); ++i) {
				cerr << "params.m_finalSeparatedLayerIndex[" << i << "] = " << params.m_finalSeparatedLayerIndex[i] << endl;
			}
		}
	}

}

void initializeAllParameters(Model& model, ModelParams& params, int inputDimension, int numLabels,
		const string& sharedLayerWidths, const string& separatedLayerWidths) {
	params.m_inputDimension = inputDimension;
	params.m_numLabels = numLabels;
	fillLayerWidths(params, sharedLayerWidths, separatedLayerWidths);

	params.m_sharedLayerWeights.clear();
	params.m_sharedLayerBias.clear();

	// Below we initialize all weights to a random double between -0.5 and 0.5.
	// model.initializeNewParameter() returns the PARAM_ID (parameter index) of the new parameter,
	// which we store in the m_weights and m_bias data structures.

	// We create a weight matrix and bias vector for each transition from one layer to the next.
	for (unsigned int k = 0; k < params.m_sharedLayerInputSizes.size(); ++k) {
		// Initialize the current weight matrix to have size params.m_layerInputSizes[k] by params.m_layerOutputSizes[k].
		vector<vector<PARAM_ID> > matk;
		for (unsigned int i = 0; i < params.m_sharedLayerInputSizes[k]; ++i) {
			vector<PARAM_ID> rowi;
			for (unsigned int j = 0; j < params.m_sharedLayerOutputSizes[k]; ++j) {
				//rowi.push_back(model.initializeNewParameter(0.0));
				rowi.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
			}
			matk.push_back(rowi);
		}
		params.m_sharedLayerWeights.push_back(matk);

		// The current bias vector has size params.m_layerOutputSizes[k].
		vector<PARAM_ID> currBias;
		for (unsigned int j = 0; j < params.m_sharedLayerOutputSizes[k]; ++j) {
			//currBias.push_back(model.initializeNewParameter(0.0));
			currBias.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
		}
		params.m_sharedLayerBias.push_back(currBias);
	}

	// We create a weight matrix and bias vector for each class's transitions from each layer to the next.
	for (unsigned int label = 0; label < params.m_separatedLayerInputSizes.size(); ++label) {
		vector<vector<vector<PARAM_ID> > > labelMatrices;
		vector<vector<PARAM_ID> > labelBiases;
		for (unsigned int k = 0; k < params.m_separatedLayerInputSizes[label].size(); ++k) {
			// Initialize the current weight matrix to have size params.m_separatedLayerInputSizes[label][k]
			// by params.m_separatedLayerOutputSizes[label][k].
			vector<vector<PARAM_ID> > matk;
			for (unsigned int i = 0; i < params.m_separatedLayerInputSizes[label][k]; ++i) {
				vector<PARAM_ID> rowi;
				for (unsigned int j = 0; j < params.m_separatedLayerOutputSizes[label][k]; ++j) {
					//rowi.push_back(model.initializeNewParameter(0.0));
					rowi.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
				}
				matk.push_back(rowi);
			}
			labelMatrices.push_back(matk);

			// The current bias vector has size params.m_separatedLayerOutputSizes[label][k].
			vector<PARAM_ID> currBias;
			for (unsigned int j = 0; j < params.m_separatedLayerOutputSizes[label][k]; ++j) {
				//currBias.push_back(model.initializeNewParameter(0.0));
				currBias.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
			}
			labelBiases.push_back(currBias);
		}
		params.m_separatedLayerWeights.push_back(labelMatrices);//[label].push_back(matk);
		params.m_separatedLayerBias.push_back(labelBiases);//[label].push_back(currBias);
	}
	cerr << "Initialized model. # parameters = " << model.getNumParams() << endl;
}

void resetDatumFormulaCache(Model& model, ModelParams& params, DatumFormulaCache& cache) {
	// Clear all data structures containing cached formulas from the previous datum, and set them up for the new datum.
	if (cache.m_cachedSharedLayerNodes.size() == params.m_sharedLayerOutputSizes.size()) {
		// Assume that the dimensions of params.m_cachedSharedLayerNodes are correct.
		for (unsigned int i = 0; i < params.m_sharedLayerOutputSizes.size(); ++i) {
			if (cache.m_cachedSharedLayerNodes[i].size() == params.m_sharedLayerOutputSizes[i]) {
				for (unsigned int j = 0; j < params.m_sharedLayerOutputSizes[i]; ++j) {
					cache.m_cachedSharedLayerNodes[i][j] = NULL;
				}
			} else {
				cerr << "Error: Size of params.m_cachedSharedLayerNodes[" << i << "] (" << cache.m_cachedSharedLayerNodes[i].size() << ") != "
					 << "params.m_sharedLayerOutputSizes[" << i << "] (" << params.m_sharedLayerOutputSizes[i] << ")" << endl;
			}
		}
	} else {
		cache.m_cachedSharedLayerNodes.clear();
		cache.m_cachedSharedLayerNodes.reserve(params.m_sharedLayerOutputSizes.size());
		for (unsigned int i = 0; i < params.m_sharedLayerOutputSizes.size(); ++i) {
			vector<Formula*> rowi(params.m_sharedLayerOutputSizes.at(i), NULL);
			cache.m_cachedSharedLayerNodes.push_back(rowi);
		}
	}
	if (cache.m_cachedSeparatedLayerNodes.size() == params.m_separatedLayerOutputSizes.size()) {
		// Assume that the dimensions of params.m_cachedSeparatedLayerNodes are correct.
		for (unsigned int label = 0; label < params.m_numLabels; ++label) {
			if (cache.m_cachedSeparatedLayerNodes[label].size() == params.m_separatedLayerOutputSizes[label].size()) {
				for (unsigned int i = 0; i < params.m_separatedLayerOutputSizes[label].size(); ++i) {
					if (cache.m_cachedSeparatedLayerNodes[label][i].size() == params.m_separatedLayerOutputSizes[label][i]) {
						for (unsigned int j = 0; j < params.m_separatedLayerOutputSizes[label][i]; ++j) {
							cache.m_cachedSeparatedLayerNodes[label][i][j] = NULL;
						}
					} else {
						cerr << "Error: Size of params.m_cachedSeparatedLayerNodes[" << label << "][" << i << "] ("
							 << cache.m_cachedSeparatedLayerNodes[label][i].size() << ") != "
							 << "params.m_separatedLayerOutputSizes[" << label << "][" << i << "] ("
							 << params.m_separatedLayerOutputSizes[label][i] << ")" << endl;
					}
				}
			}
		}
	} else {
		cache.m_cachedSeparatedLayerNodes.clear();
		cache.m_cachedSeparatedLayerNodes.reserve(params.m_numLabels);
		for (unsigned int label = 0; label < params.m_numLabels; ++label) {
			vector<vector<Formula*> > currLabelCachedNodes;
			currLabelCachedNodes.reserve(params.m_separatedLayerOutputSizes[label].size());
			for (unsigned int i = 0; i < params.m_separatedLayerOutputSizes[label].size(); ++i) {
				vector<Formula*> rowi(params.m_separatedLayerOutputSizes[label][i], NULL);
				currLabelCachedNodes.push_back(rowi);
			}
			cache.m_cachedSeparatedLayerNodes.push_back(currLabelCachedNodes);
		}
	}
	if (cache.m_features.size() == params.m_inputDimension) {
		for (unsigned int i = 0; i < params.m_inputDimension; ++i) {
			cache.m_features[i] = NULL;
		}
	} else {
		cache.m_features.clear();
		cache.m_features.reserve(params.m_inputDimension);
		for (unsigned int i = 0; i < params.m_inputDimension; ++i) {
			cache.m_features.push_back(NULL);
		}
	}
	model.resetCachedParameterFormulas(cache.m_parameters);
}

// Resets data structures for processing the next example (whether in training or decoding).
void resetForNewExample(Model& model, ModelParams& params, DatumFormulaCache& cache) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	//bool resetCachedParameterFormulasTrue = true;
	bool resetCachedParameterFormulasFalse = false;
	model.resetFormulaObjects(resetCachedParameterFormulasFalse);
	resetDatumFormulaCache(model, params, cache);
}

Formula* createLabelScoreFDAGForOutputDimension(Model& model, ModelParams& params, DatumFormulaCache& cache,
		unsigned int label, unsigned int outputDimension, const Datum& datum, bool doDropout, double dropoutRate,
		int layer, bool separated) {
	if (layer == 0) {
		if (separated) {
			return createLabelScoreFDAGForOutputDimension(model, params, cache, label, outputDimension, datum, doDropout,
				dropoutRate, params.m_finalSharedLayerIndex, false);
		} else {
			// Check the feature cache.
			Formula* f = cache.m_features.at(outputDimension);
			if (f != NULL) return f;
			Formula* currentFeatureFormula = model.getConstantFormulaObject(datum.m_features.at(outputDimension));
			cache.m_features[outputDimension] = currentFeatureFormula;
			return currentFeatureFormula;
		}
	} else {
		if (separated) {
			Formula* f = cache.m_cachedSeparatedLayerNodes.at(label).at(layer - 1).at(outputDimension);
			if (f != NULL) return f;
		} else {
			Formula* f = cache.m_cachedSharedLayerNodes.at(layer - 1).at(outputDimension);
			if (f != NULL) return f;
			//if (cache.m_cachedSharedLayerNodes.at(layer - 1).at(outputDimension) != NULL)
			//	return cache.m_cachedSharedLayerNodes.at(layer - 1).at(outputDimension);
		}
		unsigned int numInputDimensions;
		if (separated) {
			numInputDimensions = params.m_separatedLayerInputSizes[label].at(layer - 1);
		} else {
			numInputDimensions = params.m_sharedLayerInputSizes.at(layer - 1);
		}
		Formula* sumOverInputDimensions = model.getFormulaObject(PLUS, numInputDimensions+1);
		// Go through all input dimensions to the current layer.
		for (unsigned int i = 0; i < numInputDimensions; ++i) {
			bool dropout = false;
			// Only do dropout if we are not on the lowest layer (layer != 1).
			if (doDropout && (separated || layer != 1) && getRandomDouble() < dropoutRate) dropout = true;
			//if (doDropout && layer != 1 && getRandomDouble() < dropoutRate) dropout = true;
			// More efficiency can be obtained by simply avoiding construction of the units that will be
			// dropped out. Problems, however, can arise when ALL nodes get dropped out. So there should
			// be some code to either ensure that at least one node is not dropped out or to otherwise
			// add a constant child or something (or to debug the case when all are dropped out).
			//if (!dropout) {
				Formula* nodei = NULL;
				nodei = createLabelScoreFDAGForOutputDimension(model, params, cache, label, i, datum, doDropout, dropoutRate, layer - 1, separated);
				// Multiply it by the w vector corresponding to the provided label and the current dimension i.
				Formula* prod = model.getFormulaObject(TIMES, 2);
				prod->addChild(nodei);
				if (separated) {
					prod->addChild(model.getCachedParameterFormula(params.m_separatedLayerWeights[label].at(layer - 1).at(i).at(outputDimension), cache.m_parameters));
				} else {
					prod->addChild(model.getCachedParameterFormula(params.m_sharedLayerWeights.at(layer - 1).at(i).at(outputDimension), cache.m_parameters));
				}
				if (dropout) prod->addChild(model.getConstantFormulaObject(0.0));
				sumOverInputDimensions->addChild(prod);
			//}
		}
		// Add the bias weight for the current output dimension.
		if (separated) {
			sumOverInputDimensions->addChild(model.getCachedParameterFormula(params.m_separatedLayerBias[label].at(layer - 1).at(outputDimension), cache.m_parameters));
		} else {
			sumOverInputDimensions->addChild(model.getCachedParameterFormula(params.m_sharedLayerBias.at(layer - 1).at(outputDimension), cache.m_parameters));
		}
		// Use a nonlinear activation function for all layers except the final one.
		if (layer != ((int) (params.m_separatedLayerInputSizes[label].size()))) {
			Formula* activation = model.getFormulaObject(RELU);
			activation->addChild(sumOverInputDimensions);
			if (separated) {
				cache.m_cachedSeparatedLayerNodes[label][layer - 1][outputDimension] = activation;
			} else {
				cache.m_cachedSharedLayerNodes[layer - 1][outputDimension] = activation;
			}
			return activation;
		} else {
			//params.m_cachedNodes[layer - 1][outputDimension] = sumOverInputDimensions;
			if (separated) {
				cache.m_cachedSeparatedLayerNodes[label][layer - 1][outputDimension] = sumOverInputDimensions;
			} else {
				cache.m_cachedSharedLayerNodes[layer - 1][outputDimension] = sumOverInputDimensions;
			}
			return sumOverInputDimensions;
		}
	}
}

Formula* createLossFDAGForDatum(Model& model, ModelParams& params, DatumFormulaCache& cache, const Datum& datum, int correctLabel,
		bool doDropout, double dropoutRate) {
	// Create a term for the correct label's score, which is negated in the loss.
	Formula* negCorrect = model.getFormulaObject(NEG);
	// Create a sum node to sum over labels.
	Formula* sum = model.getFormulaObject(PLUS, params.m_numLabels);
	bool separatedTrue = true;
	// Sum over labels.
	for (int label = 0; label < params.m_numLabels; ++label) {
		//cerr << "final sep layer for " << label << ": " << params.m_finalSeparatedLayerIndex.at(label) << endl;
		Formula* score = createLabelScoreFDAGForOutputDimension(model, params, cache, label, 0, datum, doDropout, dropoutRate,
				params.m_finalSeparatedLayerIndex.at(label), separatedTrue);
		// If the current label is the true label, add the score to the negated correct term.
		if (label == correctLabel) negCorrect->addChild(score);
		// Always add the exponentiated label score to the sum over labels.
		Formula* expScore = model.getFormulaObject(EXP);
		expScore->addChild(score);
		sum->addChild(expScore);
	}
	// Put a log in front of the sum of exponentiated label scores.
	Formula* logSum = model.getFormulaObject(LOG);
	logSum->addChild(sum);
	// Add the negated correct score to the log-sum over all exponentiated label scores.
	Formula* negCorrectPlusSum = model.getFormulaObject(PLUS, 2);
	negCorrectPlusSum->addChild(negCorrect);
	negCorrectPlusSum->addChild(logSum);
	// Return the sum.
	return negCorrectPlusSum;
}

void fillInFeatures(DatumFormulaCache& cache, const Datum& datum) {
	// Place the datum's feature values into the constant Formulas in the cache.
	for (unsigned int i = 0; i < cache.m_features.size(); ++i) {
		cache.m_features[i]->setValue(datum.m_features[i]);
	}
}

void doDropoutOnPrecomputedLossFDAG(DatumFormulaCache& cache, double dropoutRate) {
	bool dropout = false;
	/*if (layer != 1 && getRandomDouble() < dropoutRate) dropout = true;
	// More efficiency can be obtained by simply avoiding construction of the units that will be
	// dropped out. Problems, however, can arise when ALL nodes get dropped out. So there should
	// be some code to either ensure that at least one node is not dropped out or to otherwise
	// add a constant child or something (or to debug the case when all are dropped out).
	//if (!dropout) {
		Formula* nodei = NULL;
		nodei = createLabelScoreFDAGForOutputDimension(model, params, cache, label, i, datum, doDropout, dropoutRate, layer - 1, separated);
		// Multiply it by the w vector corresponding to the provided label and the current dimension i.
		Formula* prod = model.getFormulaObject(TIMES);
		prod->addChild(nodei);
		if (separated) {
			prod->addChild(model.getCachedParameterFormula(params.m_separatedLayerWeights[label].at(layer - 1).at(i).at(outputDimension), cache.m_parameters));
		} else {
			prod->addChild(model.getCachedParameterFormula(params.m_sharedLayerWeights.at(layer - 1).at(i).at(outputDimension), cache.m_parameters));
		}
		if (dropout) prod->addChild(model.getConstantFormulaObject(0.0));
		sumOverInputDimensions->addChild(prod);
	//}
	*/
}

// Classifies the given datum by performing an argmax over all labels of the score.
bool classifyDatum(Model& model, ModelParams& params, DatumFormulaCache& cache, const Datum& datum, int exampleNumber) {
	bool doDropoutFalse = false;
	double maxScore = neg_inf;
	int maxLabel = -1;
	bool separatedTrue = true;
	// Perform the argmax over all labels.
	for (int label = 0; label < params.m_numLabels; ++label) {
		Formula* score = createLabelScoreFDAGForOutputDimension(model, params, cache, label, 0, datum, doDropoutFalse, -1.0,
			params.m_finalSeparatedLayerIndex.at(label), separatedTrue);
		double currScore = score->evaluate();
		if (currScore > maxScore) {
			maxScore = currScore;
			maxLabel = label;
		}
	}
	if (maxLabel == -1) {
		cerr << "Error: no label found with score > negative infinity! on line #" << exampleNumber << endl;
	} else {
		if (maxLabel == datum.m_label) return true;
	}
	return false;
}

// Iterates through given data and classifies each datum, then compares predicted label
// to the true label and returns the percent correct.
double classifyData(Model& model, ModelParams& params, const vector<Datum>& data, bool precomputeFormulas,
		vector<Formula*>& precomputedFormulasForLabels, vector<DatumFormulaCache>& precomputedFormulaCachesForLabels) {
	int numCorrect = 0;
	DatumFormulaCache cache;
	for (unsigned int i = 0; i < data.size(); ++i) {
		bool correct;
		//Formula* datumLossFDAG;
		//if (precomputeFormulas) {
			// Retrieve the loss FDAG for the correct label for the current training example.
			//datumLossFDAG = precomputedFormulasForLabels[data[exampleIndex].m_label];
			// Reset the FDAG's values and gradients.
			//datumLossFDAG->changedParamValues();
			// Fill in the correct feature values from the current training example.
			//fillInFeatures(precomputedFormulaCachesForLabels[trainData[exampleIndex].m_label], trainData[exampleIndex]);
			// Do dropout.
		//} else {
			resetForNewExample(model, params, cache);
			correct = classifyDatum(model, params, cache, data[i], i);
		//}
		if (correct) ++numCorrect;
	}
	return (((double)numCorrect) / ((double)(data.size())));
}

double trainAndTest(Model& model, ModelParams& params, const string& algorithm, vector<Datum>& trainData,
		vector<Datum>& testData, double dropoutRate, int numEpochs, bool doFiniteDifferenceChecking, int fold) {
	struct timeval initialTime, endTime, time1, time2, time3, time4,
		timePrebuildStart, timePrebuildEnd, timeBuildStart, timeBuildEnd, timeEvalStart, timeEvalEnd,
		timeDecodeStart, timeDecodeEnd;
	//struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);

	double totalPrebuildTime = 0.0;
	double totalBuildTime = 0.0;
	double totalEvalTime = 0.0;
	double totalDecodeTime = 0.0;

	double stepsize = 0.1;
	int minibatchSize = 1;
	double trainAccuracy = 0.0;
	double testAccuracy = 0.0;

	bool doDropout = true;
	if (dropoutRate <= 0.0) doDropout = false;

	Algorithm alg = model.getAlgorithm(algorithm);
	// Initialize vector of previous updates for SGD w/ momentum.
	vector<double> sgdMomentumUpdates(model.getNumParams(), 0.0);
	double momentumCoefficient = 0.99;
	// Initialize sum of squared gradients for AdaGrad and AdaDelta.
	vector<double> sumOfSquaredGradients(model.getNumParams(), 0.0);
	// Initialize sum of squared updates for AdaDelta.
	vector<double> sumOfSquaredUpdates(model.getNumParams(), 0.0);
	double epsilon = 0.000001;
	double rho = 0.95;

	// Scale feature values: divide each by its max absolute value in the training data.
	scaleFeatures(trainData, testData);

	DatumFormulaCache cache;

	bool precomputeFormulas = false;
	vector<Formula*> precomputedFormulasForLabels;
	vector<DatumFormulaCache> precomputedFormulaCachesForLabels;
	if (precomputeFormulas) {
		gettimeofday(&timePrebuildStart, 0);

		// At the outset, reset Formula data structures in the model.
		resetForNewExample(model, params, cache);
		bool doDropoutFalse = false;
		// Precompute loss FDAG for each possible label.
		for (int label = 0; label < params.m_numLabels; ++label) {
			DatumFormulaCache labelCache;
			resetDatumFormulaCache(model, params, labelCache);

			// Create the loss FDAG for the current label being correct. Just use the first training
			// example arbitrarily. We don't do any dropout when building these generic FDAGs. We'll
			// do dropout later when we instantiate them for particular examples.
			Formula* lossFDAG = createLossFDAGForDatum(model, params, labelCache, trainData.at(0), label, doDropoutFalse, -1.0);
			precomputedFormulasForLabels.push_back(lossFDAG);
			precomputedFormulaCachesForLabels.push_back(labelCache);
		}
		gettimeofday(&timePrebuildEnd, 0);
		totalPrebuildTime += computeTimeDifference(timePrebuildStart, timePrebuildEnd);
	}
	int syncValue = 1;
	// Train for numEpochs iterations of training.
	for (int iter = 1; iter <= numEpochs; ++iter) {
		double iterLoss = 0.0;
		double iterGradientNorm = 0.0;
		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			Formula* datumLossFDAG;
			if (precomputeFormulas) {
				//gettimeofday(&timeBuildStart, 0);
				// Retrieve the loss FDAG for the correct label for the current training example.
				datumLossFDAG = precomputedFormulasForLabels[trainData[exampleIndex].m_label];
				// Old way was to reset the FDAG's values and gradients. But this was too slow, so
				// we now just use a sync value (see call to evalAndComputeDerivatives() below).
				//datumLossFDAG->changedParamValues(true);
				//model.resetFormulaComputedFlags();

				// Fill in the correct feature values from the current training example.
				fillInFeatures(precomputedFormulaCachesForLabels[trainData[exampleIndex].m_label], trainData[exampleIndex]);
				// Do dropout.
				if (doDropout) doDropoutOnPrecomputedLossFDAG(precomputedFormulaCachesForLabels[trainData[exampleIndex].m_label], dropoutRate);
				//gettimeofday(&timeBuildEnd, 0);
				//totalBuildTime += computeTimeDifference(timeBuildStart, timeBuildEnd);

			} else {
				//gettimeofday(&timeBuildStart, 0);
				// We are processing a new example, so reset Formula data structures in the model.
				resetForNewExample(model, params, cache);

				// Create the formula containing the loss for the current example.
				datumLossFDAG = createLossFDAGForDatum(model, params, cache, trainData[exampleIndex], trainData[exampleIndex].m_label,
						doDropout, dropoutRate);
				//gettimeofday(&timeBuildEnd, 0);
				//totalBuildTime += computeTimeDifference(timeBuildStart, timeBuildEnd);
			}

			//gettimeofday(&timeEvalStart, 0);
			double loss = 0.0;
			if (precomputeFormulas) {
				loss = model.evalAndComputeDerivatives(datumLossFDAG,
					precomputedFormulaCachesForLabels[trainData[exampleIndex].m_label].m_parameters,
					syncValue, exampleIndex, doFiniteDifferenceChecking);
				++syncValue;
			} else {
				// Compute gradient and loss for current example.
				loss = model.evalAndComputeDerivatives(datumLossFDAG, cache.m_parameters, 1, exampleIndex, doFiniteDifferenceChecking);
			}
			//datumLossFDAG->printFormula();
			if (loss != loss) {
				cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
			} else if (loss == pos_inf) {
				cerr << "Warning: inf loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
			} else if (loss == neg_inf) {
				cerr << "Warning: -inf loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
			} else {
				iterLoss += loss;
			}
			//gettimeofday(&timeEvalEnd, 0);
			//totalEvalTime += computeTimeDifference(timeEvalStart, timeEvalEnd);

			// If we're at the end of a mini-batch or at the end of the training data, make an update.
			bool doUpdate = false;
			if (exampleIndex % minibatchSize == 0 || exampleIndex == trainData.size() - 1) doUpdate = true;
			if (doUpdate) {
				double gradientSquaredNorm = 0.0;
				if (alg == SGD_MOMENTUM) {
					gradientSquaredNorm = model.doSGDMomentumUpdate(sgdMomentumUpdates, stepsize, momentumCoefficient);
				} else if (alg == ADAGRAD) {
					gradientSquaredNorm = model.doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				} else if (alg == ADADELTA){
					gradientSquaredNorm = model.doAdaDeltaUpdate(sumOfSquaredGradients, sumOfSquaredUpdates, epsilon, rho);
				}
				iterGradientNorm += sqrt(gradientSquaredNorm);
				model.resetGradientVector();
			}
		}
		if (iter % 10 == 0) cerr << "it " << iter << "\tloss\t" << iterLoss <<
				"\tparam_norm\t" << model.computeParameterVectorNorm() <<
				"\tgrad_norm\t" << iterGradientNorm << endl;
		if (0 && iter % 100 == 0 && iter != numEpochs) {
			gettimeofday(&timeDecodeStart, 0);
			testAccuracy = classifyData(model, params, testData, precomputeFormulas,
				precomputedFormulasForLabels, precomputedFormulaCachesForLabels);
			gettimeofday(&timeDecodeEnd, 0);
			totalDecodeTime += computeTimeDifference(timeDecodeStart, timeDecodeEnd);
			cerr << "Partial-fold " << fold << " (iter " << iter << ") test accuracy:\t" << testAccuracy << endl;
		}
	}
	// Compute final train/test accuracy.
	gettimeofday(&timeDecodeStart, 0);
	trainAccuracy = classifyData(model, params, trainData, precomputeFormulas,
			precomputedFormulasForLabels, precomputedFormulaCachesForLabels);
	cerr << "Fold " << fold << " train accuracy:\t" << trainAccuracy << endl;
	testAccuracy = classifyData(model, params, testData, precomputeFormulas,
			precomputedFormulasForLabels, precomputedFormulaCachesForLabels);
	gettimeofday(&timeDecodeEnd, 0);
	totalDecodeTime += computeTimeDifference(timeDecodeStart, timeDecodeEnd);
	cerr << "Fold " << fold << " test accuracy:\t" << testAccuracy << endl;
	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Fold " << fold << " time:\t" << cumulativeTime << endl;

	cerr << "times:\tprebuild\t" << totalPrebuildTime << "\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
		 << "\tdec\t" << totalDecodeTime << endl;

	return testAccuracy;
}

int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("Demonstrates a class-separated deep neural network using fdag; uses AdaGrad to minimize cross-entropy for a multiclass classification task.");
	params.addConstraint("data_file", "each line should contain comma-delimited real-valued features with label in final field", false); // not optional
    params.addConstraint("algorithm", "optimization algorithm; can be \"sgd_momentum\", \"adagrad\", or \"adadelta\" (case insensitive) (default: adadelta)", true); // optional
    params.addConstraint("shared_layer_widths", "widths of class-shared hidden layers, comma-delimited. e.g., \"10,5\", will use one "
    		"hidden layer of width 10 followed by a second hidden layer of width 5; use \"0\" for no class-shared hidden layers "
    		"(default: 20)", true); // optional
    params.addConstraint("separated_layer_widths", "widths of class-separated hidden layers, contains a separate sequence entry for each "
    		"class label separated by colon, as each label can have a different sequence of hidden layers; each sequence is in "
    		"same format as shared_layer_widths; "
    		"use \"0\" for no class-separated hidden layers for a class (default: 20:20)", true); // optional
    params.addConstraint("dropout_rate", "dropout rate; use a value <= 0 for no dropout (default: 0.5)", true); // optional
    params.addConstraint("train_epochs", "number of training epochs (default: 100)", true); // optional
   	if (!params.runConstraints("run_csdnn")) return 0;
	// Read in command-line parameters.
	string dataFile = params.asString("data_file");
	string algorithm = params.asString("algorithm", "adadelta");
	string sharedLayerWidths = params.asString("shared_layer_widths", "20");
	string separatedLayerWidths = params.asString("separated_layer_widths", "20;20");
	double dropoutRate = atof(params.asString("dropout_rate", "0.5").c_str());
	int trainEpochs = atoi(params.asString("train_epochs", "100").c_str());

	// Change following to true to do finite difference checking of gradients.
	bool doFiniteDifferenceChecking = false;
	srand(10);
	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);
	// Load data.
	vector<Datum> data;
	int numLabels;
	int inputDimension = loadData(dataFile, data, &numLabels);

	// Do 5-fold cross-validation.
	double accuracySum = 0.0;
	unsigned int dataSize = data.size();
	int foldSize = dataSize / 5;
	vector<double> accuracies;
	for (int i = 0; i < 5; ++i) {
		// Generate train/test sets for the current fold (approximately based on discretized index cutoffs).
		vector<Datum> trainData, testData;
		for (unsigned int j = 0; j < dataSize; ++j) {
			if (((int)j) >= (i*foldSize) && ((int)j) < (i+1)*foldSize) {
				// If the current datum falls within the indices of the current test fold, add it to testData.
				testData.push_back(data[j]);
			} else {
				// Otherwise, add it to the training set.
				trainData.push_back(data[j]);
			}
		}
		// Create a model for the current fold with given initial Formula capacity.
		Model model(1000, 1000, 1000, 1000);
		// Create and initialize its parameters.
		ModelParams modelParams;
		initializeAllParameters(model, modelParams, inputDimension, numLabels, sharedLayerWidths, separatedLayerWidths);
		// Train and test.
		double currAccuracy = trainAndTest(model, modelParams, algorithm, trainData, testData, dropoutRate, trainEpochs, doFiniteDifferenceChecking, i+1);
		accuracies.push_back(currAccuracy);
		accuracySum += currAccuracy;
	}
	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
	double accuracyMean = accuracySum / 5.0;
	cerr << "5-fold CV mean accuracy: \t" << accuracyMean << endl;
	cerr << "5-fold CV accuracy stddev: \t" << standardDeviation(accuracies) << endl;

    return 0;
}
