#include <iostream>
#include <stdlib.h>
#include <sys/time.h>
#include "../utils/cmdparams.h"
#include "../fdag/model.h"
#include "classifier_utils.h"

struct ModelParams {
	// The input dimension is the number of features.
	unsigned int m_inputDimension;
	// We store separate vectors of the layer input and output dimensions for convenience.
	vector<unsigned int> m_layerInputSizes;
	vector<unsigned int> m_layerOutputSizes;
	// The number of labels.
	int m_numLabels;
	// The number of layers.
	unsigned int m_numLayers;
	// The index to pass as the final layer index when recursively building the FDAGs for all layers.
	unsigned int m_finalLayerIndex;

	// The weight matrices for the layers have the following sizes:
	// inputDimension x m_layerWidths[0]
	// m_layerWidths[0] x m_layerWidths[1]
	// ...
	// m_layerWidths[m_numLayers - 2] x m_layerWidths[m_numLayers - 1]
	// m_layerWidths[m_numLayers - 1] x m_numLabels   <-- this is for the final softmax layer
	vector<vector<vector<PARAM_ID> > > m_weights;
	// The bias vectors for the layers have the following sizes:
	// m_layerWidths[0]
	// m_layerWidths[1]
	// ...
	// m_layerWidths[m_numLayers - 1]
	// m_numLabels   <-- this is for the final softmax layer
	vector<vector<PARAM_ID> > m_bias;

	// Pointers to Formulas for cached nodes for current example, which are NULL if unset (or dropped out).
	vector<vector<Formula*> > m_cachedNodes;
};

void initializeAllParameters(Model& model, ModelParams& params, int inputDimension, int numLabels, const string& layerWidths) {
	params.m_inputDimension = inputDimension;
	vector<string> layerWidthStrings;
	tokenize(layerWidths, ',', layerWidthStrings);
	if (layerWidths == "0") {
		params.m_numLayers = 0;
	} else {
		params.m_numLayers = layerWidthStrings.size();
	}
	params.m_numLabels = numLabels;

	// Fill the vectors of input and output layer widths.
	params.m_layerInputSizes.push_back(params.m_inputDimension);
	for (unsigned int i = 0; i < params.m_numLayers; ++i) {
		int currentLayerWidth = atoi(layerWidthStrings[i].c_str());
		params.m_layerInputSizes.push_back(currentLayerWidth);
		params.m_layerOutputSizes.push_back(currentLayerWidth);
	}
	params.m_layerOutputSizes.push_back(params.m_numLabels);

	params.m_finalLayerIndex = params.m_layerInputSizes.size();

	if (params.m_layerInputSizes.size() != params.m_numLayers + 1) {
		cerr << "Error: number of layer input sizes (" << params.m_layerInputSizes.size()
			 << ") not equal to 1 + number of layers (" << (params.m_numLayers + 1) << ")" << endl;
	}
	if (params.m_layerOutputSizes.size() != params.m_numLayers + 1) {
		cerr << "Error: number of layer output sizes (" << params.m_layerOutputSizes.size()
			 << ") not equal to 1 + number of layers (" << (params.m_numLayers + 1) << ")" << endl;
	}

	//for (unsigned int i = 0; i < params.m_layerInputSizes.size(); ++i) {
	//	cerr << "m_layerInputSizes[" << i << "] = " << params.m_layerInputSizes[i] << endl;
	//	cerr << "m_layerOutputSizes[" << i << "] = " << params.m_layerOutputSizes[i] << endl;
	//}

	params.m_weights.clear();
	params.m_bias.clear();

	// Below we initialize all weights to a random double between -0.5 and 0.5.
	// model.initializeNewParameter() returns the PARAM_ID (parameter index) of the new parameter,
	// which we store in the m_weights and m_bias data structures.

	// We create a weight matrix and bias vector for each transition from one layer to the next.
	for (unsigned int k = 0; k < params.m_layerInputSizes.size(); ++k) {
		// Initialize the current weight matrix to have size params.m_layerInputSizes[k] by params.m_layerOutputSizes[k].
		vector<vector<PARAM_ID> > matk;
		for (unsigned int i = 0; i < params.m_layerInputSizes[k]; ++i) {
			vector<PARAM_ID> rowi;
			for (unsigned int j = 0; j < params.m_layerOutputSizes[k]; ++j) {
				//rowi.push_back(model.initializeNewParameter(0.0));
				rowi.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
			}
			matk.push_back(rowi);
		}
		params.m_weights.push_back(matk);

		// The current bias vector has size params.m_layerOutputSizes[k].
		vector<PARAM_ID> currBias;
		for (unsigned int j = 0; j < params.m_layerOutputSizes[k]; ++j) {
			//currBias.push_back(model.initializeNewParameter(0.0));
			currBias.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
		}
		params.m_bias.push_back(currBias);
	}
	cerr << "Initialized model. # parameters = " << model.getNumParams() << endl;
}

// Resets data structures for processing the next example (whether in training or decoding).
void resetForNewExample(Model& model, ModelParams& params) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	bool resetCachedParameterFormulasTrue = true;
	model.resetFormulaObjects(resetCachedParameterFormulasTrue);

	// Clear all data structures containing cached formulas from the previous datum, and set them up for the new datum.
	if (params.m_cachedNodes.size() == params.m_layerOutputSizes.size()) {

		// Assume that the dimensions of params.m_cachedNodes are correct.
		for (unsigned int i = 0; i < params.m_layerOutputSizes.size(); ++i) {
			if (params.m_cachedNodes[i].size() == params.m_layerOutputSizes[i]) {
				for (unsigned int j = 0; j < params.m_layerOutputSizes[i]; ++j) {
					params.m_cachedNodes[i][j] = NULL;
				}
			} else {
				cerr << "Error: Size of params.m_cachedNodes[" << i << "] (" << params.m_cachedNodes[i].size() << ") != "
					 << "params.m_layerOutputSizes[" << i << "] (" << params.m_layerOutputSizes[i] << ")" << endl;
			}
		}
	} else {
		params.m_cachedNodes.clear();
		params.m_cachedNodes.reserve(params.m_layerOutputSizes.size());
		for (unsigned int i = 0; i < params.m_layerOutputSizes.size(); ++i) {
			vector<Formula*> rowi(params.m_layerOutputSizes.at(i), NULL);
			params.m_cachedNodes.push_back(rowi);
		}
	}
}

Formula* createScoreFDAGForOutputDimension(Model& model, ModelParams& params, unsigned int outputDimension, const Datum& datum,
		bool doDropout, double dropoutRate, int layer) {
	if (layer == 0) {
		return model.getConstantFormulaObject(datum.m_features.at(outputDimension));
	} else {
		if (params.m_cachedNodes.at(layer - 1).at(outputDimension) != NULL) return params.m_cachedNodes.at(layer - 1).at(outputDimension);
		Formula* sumOverInputDimensions = model.getFormulaObject(PLUS);
		unsigned int numInputDimensions = params.m_layerInputSizes.at(layer - 1);
		// Go through all input dimensions to the current layer.
		for (unsigned int i = 0; i < numInputDimensions; ++i) {
			bool dropout = false;
			// Only do dropout if we are not on the lowest layer (layer != 1).
			if (doDropout && layer != 1 && getRandomDouble() < dropoutRate) dropout = true;
			// More efficiency can be obtained by simply avoiding construction of the units that will be
			// dropped out. Problems, however, can arise when ALL nodes get dropped out. So there should
			// be some code to either ensure that at least one node is not dropped out or to otherwise
			// add a constant child or something (or to debug the case when all are dropped out).
			//if (!dropout) {
				Formula* nodei = NULL;
				nodei = createScoreFDAGForOutputDimension(model, params, i, datum, doDropout, dropoutRate, layer - 1);
				// Multiply it by the w vector corresponding to the provided label and the current dimension i.
				Formula* prod = model.getFormulaObject(TIMES);
				prod->addChild(nodei);
				prod->addChild(model.getCachedParameterFormula(params.m_weights.at(layer - 1).at(i).at(outputDimension)));
				if (dropout) prod->addChild(model.getConstantFormulaObject(0.0));
				sumOverInputDimensions->addChild(prod);
			//}
		}
		// Add the bias weight for the current output dimension.
		sumOverInputDimensions->addChild(model.getCachedParameterFormula(params.m_bias.at(layer - 1).at(outputDimension)));
		// Use a nonlinear activation function for all layers except the final one.
		if (layer != ((int) (params.m_layerInputSizes.size()))) {
			Formula* activation = model.getFormulaObject(RELU);
			activation->addChild(sumOverInputDimensions);
			params.m_cachedNodes[layer - 1][outputDimension] = activation;
			return activation;
		} else {
			params.m_cachedNodes[layer - 1][outputDimension] = sumOverInputDimensions;
			return sumOverInputDimensions;
		}
	}
}

Formula* createLossFDAGForDatum(Model& model, ModelParams& params, const Datum& datum, int exampleNumber, bool doDropout, double dropoutRate) {
	// Create a term for the correct label's score, which is negated in the loss.
	Formula* negCorrect = model.getFormulaObject(NEG);
	// Create a sum node to sum over labels.
	Formula* sum = model.getFormulaObject(PLUS);
	// Sum over labels.
	for (int label = 0; label < params.m_numLabels; ++label) {
		Formula* score = createScoreFDAGForOutputDimension(model, params, label, datum, doDropout, dropoutRate, params.m_finalLayerIndex);
		// If the current label is the true label, add the score to the negated correct term.
		if (label == datum.m_label) negCorrect->addChild(score);
		// Always add the exponentiated label score to the sum over labels.
		Formula* expScore = model.getFormulaObject(EXP);
		expScore->addChild(score);
		sum->addChild(expScore);
	}
	// Put a log in front of the sum of exponentiated label scores.
	Formula* logSum = model.getFormulaObject(LOG);
	logSum->addChild(sum);
	// Add the negated correct score to the log-sum over all exponentiated label scores.
	Formula* negCorrectPlusSum = model.getFormulaObject(PLUS);
	negCorrectPlusSum->addChild(negCorrect);
	negCorrectPlusSum->addChild(logSum);
	// Return the sum.
	return negCorrectPlusSum;
}

// Classifies the given datum by performing an argmax over all labels of the score.
bool classifyDatum(Model& model, ModelParams& params, const Datum& datum, int exampleNumber) {
	bool doDropoutFalse = false;
	double maxScore = neg_inf;
	int maxLabel = -1;
	// Perform the argmax over all labels.
	for (int label = 0; label < params.m_numLabels; ++label) {
		Formula* score = createScoreFDAGForOutputDimension(model, params, label, datum, doDropoutFalse, -1.0, params.m_finalLayerIndex);
		double currScore = score->evaluate();
		if (currScore > maxScore) {
			maxScore = currScore;
			maxLabel = label;
		}
	}
	if (maxLabel == -1) {
		cerr << "Error: no label found with score > negative infinity! on line #" << exampleNumber << endl;
	} else {
		if (maxLabel == datum.m_label) return true;
	}
	return false;
}

// Iterates through given data and classifies each datum, then compares predicted label
// to the true label and returns the percent correct.
double classifyData(Model& model, ModelParams& params, const vector<Datum>& data) {
	int numCorrect = 0;
	for (unsigned int i = 0; i < data.size(); ++i) {
		resetForNewExample(model, params);
		bool correct = classifyDatum(model, params, data[i], i);
		if (correct) ++numCorrect;
	}
	return (((double)numCorrect) / ((double)(data.size())));
}

double trainAndTest(Model& model, ModelParams& params, const string& algorithm, vector<Datum>& trainData,
		vector<Datum>& testData, double dropoutRate, int numEpochs, bool doFiniteDifferenceChecking, int fold) {
	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);

	double stepsize = 0.1;
	int minibatchSize = 1;
	//unsigned int numEpochs = 200;
	double trainAccuracy = 0.0;
	double testAccuracy = 0.0;

	bool doDropout = true;
	if (dropoutRate <= 0.0) doDropout = false;

	Algorithm alg = model.getAlgorithm(algorithm);
	// Initialize vector of previous updates for SGD w/ momentum.
	vector<double> sgdMomentumUpdates(model.getNumParams(), 0.0);
	double momentumCoefficient = 0.99;
	// Initialize sum of squared gradients for AdaGrad and AdaDelta.
	vector<double> sumOfSquaredGradients(model.getNumParams(), 0.0);
	// Initialize sum of squared updates for AdaDelta.
	vector<double> sumOfSquaredUpdates(model.getNumParams(), 0.0);
	double epsilon = 0.00001;
	double rho = 0.99;

	// Scale feature values -- divide each by its max value in the training data.
	scaleFeatures(trainData, testData);

	// Train for numEpochs iterations of training.
	for (int iter = 1; iter <= numEpochs; ++iter) {
		double iterLoss = 0.0;
		double iterGradientNorm = 0.0;
		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			// We are processing a new example, so reset Formula data structures in the model.
			resetForNewExample(model, params);

			// Create the formula containing the loss for the current example.
			Formula* datumLossFDAG = createLossFDAGForDatum(model, params, trainData[exampleIndex], exampleIndex, doDropout, dropoutRate);

			// Compute gradient and loss for current example.
			double loss = model.evalAndComputeDerivatives(datumLossFDAG, exampleIndex, doFiniteDifferenceChecking);
			if (loss != loss) {
				cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
			} else if (loss == pos_inf) {
				cerr << "Warning: inf loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
			} else if (loss == neg_inf) {
				cerr << "Warning: -inf loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
			} else {
				iterLoss += loss;
			}

			// If we're at the end of a mini-batch or at the end of the training data, make an update.
			bool doUpdate = false;
			if (exampleIndex % minibatchSize == 0 || exampleIndex == trainData.size() - 1) doUpdate = true;
			if (doUpdate) {
				double gradientSquaredNorm = 0.0;
				if (alg == SGD_MOMENTUM) {
					gradientSquaredNorm = model.doSGDMomentumUpdate(sgdMomentumUpdates, stepsize, momentumCoefficient);
				} else if (alg == ADAGRAD) {
					gradientSquaredNorm = model.doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				} else if (alg == ADADELTA){
					gradientSquaredNorm = model.doAdaDeltaUpdate(sumOfSquaredGradients, sumOfSquaredUpdates, epsilon, rho);
				}
				iterGradientNorm += sqrt(gradientSquaredNorm);
				model.resetGradientVector();
			}
		}
		if (iter % 10 == 0) {
			cerr << "it " << iter << "\tloss\t" << iterLoss <<
				"\tparam_norm\t" << model.computeParameterVectorNorm() <<
				"\tgrad_norm\t" << iterGradientNorm;
			//testAccuracy = classifyData(model, params, testData);
			//cerr << "\ttestacc\t" << testAccuracy;
			cerr << endl;
		}
		if (iter % 100 == 0 && iter != numEpochs) {
			testAccuracy = classifyData(model, params, testData);
			cerr << "Partial-fold " << fold << " (iter " << iter << ") test accuracy:\t" << testAccuracy << endl;
		}
	}
	// Compute final train/test accuracy.
	trainAccuracy = classifyData(model, params, trainData);
	cerr << "Fold " << fold << " train accuracy:\t" << trainAccuracy << endl;
	testAccuracy = classifyData(model, params, testData);
	cerr << "Fold " << fold << " test accuracy:\t" << testAccuracy << endl;
	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Fold " << fold << " time:\t" << cumulativeTime << endl;
	return testAccuracy;
}

int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("Demonstrates a deep neural network using fdag; uses AdaGrad to minimize log loss for a multiclass classification task.");
	params.addConstraint("data_file", "each line should contain comma-delimited real-valued features with label in final field", false); // not optional
    params.addConstraint("algorithm", "optimization algorithm; can be \"sgd_momentum\", \"adagrad\", or \"adadelta\" (case insensitive) (default: adadelta)", true); // optional
    params.addConstraint("layer_widths", "widths of hidden layers, comma-delimited. e.g., \"10,10,5\", or use \"0\" for no hidden layers (default: 10,5)", true); // optional
    params.addConstraint("dropout_rate", "dropout rate; use a value <= 0 for no dropout (default: 0.5)", true); // optional
    params.addConstraint("train_epochs", "number of training epochs (default: 100)", true); // optional
   	if (!params.runConstraints("run_dnn")) return 0;
	// Read in command-line parameters.
	string dataFile = params.asString("data_file");
	string algorithm = params.asString("algorithm", "adadelta");
	string layerWidths = params.asString("layer_widths", "10,5");
	double dropoutRate = atof(params.asString("dropout_rate", "0.5").c_str());
	int trainEpochs = atoi(params.asString("train_epochs", "100").c_str());

	// Change following to true to do finite difference checking of gradients.
	bool doFiniteDifferenceChecking = false;
	srand(10);
	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);
	// Load data.
	vector<Datum> data;
	int numLabels;
	int inputDimension = loadData(dataFile, data, &numLabels);

	// Do 5-fold cross-validation.
	double accuracySum = 0.0;
	unsigned int dataSize = data.size();
	int foldSize = dataSize / 5;
	vector<double> accuracies;
	for (int i = 0; i < 5; ++i) {
		// Generate train/test sets for the current fold (approximately based on discretized index cutoffs).
		vector<Datum> trainData, testData;
		for (unsigned int j = 0; j < dataSize; ++j) {
			if (((int)j) >= (i*foldSize) && ((int)j) < (i+1)*foldSize) {
				// If the current datum falls within the indices of the current test fold, add it to testData.
				testData.push_back(data[j]);
			} else {
				// Otherwise, add it to the training set.
				trainData.push_back(data[j]);
			}
		}
		// Create a model for the current fold with given initial Formula capacity.
		Model model(100000);
		// Create and initialize its parameters.
		ModelParams modelParams;
		initializeAllParameters(model, modelParams, inputDimension, numLabels, layerWidths);
		// Train and test.
		double currAccuracy = trainAndTest(model, modelParams, algorithm, trainData, testData, dropoutRate, trainEpochs, doFiniteDifferenceChecking, i+1);
		accuracies.push_back(currAccuracy);
		accuracySum += currAccuracy;
	}
	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
	double accuracyMean = accuracySum / 5.0;
	cerr << "5-fold CV mean accuracy: \t" << accuracyMean << endl;
	cerr << "5-fold CV accuracy stddev: \t" << standardDeviation(accuracies) << endl;
    return 0;
}
