#ifndef SEGMENT_CLASSIFIER_MODEL_H_
#define SEGMENT_CLASSIFIER_MODEL_H_

#include <fstream>
#include <iostream>
#include <vector>
#include <string>
#include <set>
#include <queue>
#include <list>
#include "../../fdag/model.h"
#include "../classifier_utils.h"
#include "../rnn/lstm.h"
#include "../rnn/gru.h"

using namespace std;

struct CompareIntDoublePairPriorityQueue {
	bool operator() (const pair<int, double>& a, const pair<int, double>& b) const {
		return a.second < b.second;
	}
};

struct CompareStringDoublePairPriorityQueue {
	bool operator() (const pair<string, double>& a, const pair<string, double>& b) const {
		return a.second < b.second;
	}
};

struct CompareStringDoublePairSet {
	bool operator() (const pair<string, double>& a, const pair<string, double>& b) const {
		return a.second > b.second;
	}
};

struct CompareIntDoublePairMap {
	bool operator() (const pair<int, double>& a, const pair<int, double>& b) const {
		return a.second > b.second;
	}
};

class SegmentClassifierModel : public Model {
public:
	//************************************** model.cpp:
	SegmentClassifierModel(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity);
	~SegmentClassifierModel();
	double trainAndTest(const vector<string>& trainLines, const vector<string>& devLines, const vector<string>& testLines,
		const string& devPredFilenamePrefix, const string& testPredFilenamePrefix, const string& phoneSetFilename,
		const string& forwardLSTMs, const string& backwardLSTMs,
		const string& forwardGRUs, const string& backwardGRUs,
		const string& algorithm, double stepsize, int minibatchSize, int numEpochs, int decodeEveryK,
		double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate,
		bool stochasticPooling, double l2NormConstraint, double gradientNormThreshold, bool useCost, bool useCoarseCost, bool doFiniteDifferenceChecking);

private:

	void initializeNetworkFromConfiguration(const string& cfg, bool lstm, bool forward);
	void fillPhoneLabelSet(const string& phoneSetFilename, map<string, int>& labelMap);
	void addLabelMappingToEvaluationLabelMap(const map<string, int>& labelMap, const string& s1, const string& s2);
	void fillEvaluationLabelMap(const map<string, int>& labelMap);

	void resetForNewExample(MFCCSequenceDatum& datum);
	Formula* createLossFDAGForMFCCSequenceDatum(MFCCSequenceDatum& datum, double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate, bool useCost);
	Formula* createSoftmaxScoreForLabel(unsigned int label, MFCCSequenceDatum& datum, double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate, bool usingCost);

	double doDecoding(vector<MFCCSequenceDatum>& data, const string& predFilenamePrefix, const string& filenameSuffix);
	bool decode(MFCCSequenceDatum& datum, int exampleNumber, int& predictedLabel, bool writeToFile, string& posteriorsToWrite);

	//void printNearestNeighbors(vector<MFCCSequenceDatum>& trainData, vector<MFCCSequenceDatum>& devData);

	// The input dimension is the size of each frame.
	int m_inputDimension;
	// The number of class labels.
	unsigned int m_numLabels;

	// The model can contain multiple LSTM RNNs and multiple GRU (gated recurrent unit) RNNs.
	vector<DeepLSTM> m_lstmParams;
	vector<DeepGRU> m_gruParams;

	map<int, int> m_evaluationLabelMap;
	vector<string> m_labelNames;

	bool m_useStochasticPooling;
	int m_maxSequenceLength;
	bool m_useCoarseCost;
};

#endif
