//
// File:   ParamsArray.h
// Author: Shay Cohen
//
// Created on August 27, 2007, 12:22 PM
//

#ifndef PARAMSARRAY_H_
#define	PARAMSARRAY_H_

#include <vector>
#include <map>
#include <string>
#include <iostream>
#include <stdlib.h>
//#include "utils.h"

using namespace std;

class ParamsConstraint {
public:
    ParamsConstraint(const string& p, const string& d, bool o) {
        param = p;
        desc = d;
        optional = o;
    }
    string param;
    string desc;
    bool optional;
};


/**
 * ParamsArray is a class for handling command-line parameters.
 * The parameters in the command line are accepted in the form
 * of <code>name=value name=value name=value</code> ...
 * where name is the parameter's name and value is its value.
 *
 * @author	Shay Cohen, Kevin Gimpel
 */

class ParamsArray {
private:
    map<string, string> paramsHash;
    vector<string> paramsList;
    vector<ParamsConstraint> paramsConstraints;
    string description;
    
    /**
     * A constructor for creating a parameters list.
     *
     * @param args	Array of parameters from the command-line.
     */
public:
    ParamsArray(int argc, char** argv) {
        for (int i = 1; i < argc; ++i) {
            addParam(argv[i]);
        }
    }
    
    /**
     * A constructor for creating an empty parameters list.
     */
    ParamsArray() { }

    void tokenize(const string& str, char delimiter, vector<string>& tokens) {
    		// Skip delimiters at beginning. Find first "non-delimiter".
    		string::size_type lastPos = str.find_first_not_of(delimiter, 0);
    		// Find next instance of the delimiter.
    		string::size_type pos = str.find_first_of(delimiter, lastPos);

    		while (string::npos != pos || string::npos != lastPos) {
    			// Add the token from lastPos to pos.
    			tokens.push_back(str.substr(lastPos, pos - lastPos));

           	// Find the start of the next token.
           	lastPos = str.find_first_not_of(delimiter, pos);

           	// Find the end of the token.
           	pos = str.find_first_of(delimiter, lastPos);
       	}
    }
    
    /**
     * Adds a new parameter to the parameter list.
     *
     * @param param Paramater in the form of name=value
     */
    void addParam(const string& param) {
        vector<string> tokens;
        tokenize(param, '=', tokens);
        
        if (tokens.size() == 2) {
            paramsList.push_back(tokens[0]);
            paramsHash[tokens[0]] = tokens[1];
        } else {
        		cerr << "Error: " << param << " -- expected param=value format for command-line parameters." << endl;
        }
    }
    
    /**
     * Returns a parameter as integer.
     *
     * @param paramName Name of parameter to return
     */
    int asInt(const string& paramName) {
        if (paramsHash.find(paramName) == paramsHash.end()) { return -1; }
        return atoi(paramsHash[paramName].c_str());
    }
    
    /**
     * Returns a parameter as integer with default value.
     *
     * @param paramName Name of parameter to return
     * @param d Default value to return
     */
    int asInt(const string& paramName, int d) {
        if (paramsHash.find(paramName) == paramsHash.end())
            return d;
        
        int i;
        i = atoi(paramsHash[paramName].c_str());
        
        return i;
    }
    
    /**
     * Returns a parameter as integer with default option.
     *
     * @param paramNum Index of parameter to return
     * @param d Default value to return
     */
    int asInt(unsigned int paramNum, int d) {
        if (paramNum >= paramsList.size()) return -1;
        return asInt(paramsList[paramNum], d);
    }
    
    /**
     * Returns a parameter as integer.
     *
     * @param paramNum Index of parameter to return
     */
    int asInt(unsigned int paramNum) {
        if (paramNum >= paramsList.size()) return -1;
        return asInt(paramsList[paramNum]);
    }
    
    /**
     * Returns a parameter as double.
     *
     * @param paramName Name of parameter to return
     */
    double asDouble(const string& paramName) {
        if (paramsHash.find(paramName) == paramsHash.end()) { return -1; }
        
        return atof(paramsHash[paramName].c_str());
    }
    
    /**
     * Returns a parameter as double with default value.
     *
     * @param paramName Name of parameter to return
     * @param d Default value to return
     */
    double asDouble(const string& paramName, double d) {
        if (paramsHash.find(paramName) == paramsHash.end())
            return d;
        return atof(paramsHash[paramName].c_str());
    }
    
    /**
     * Returns a parameter as double with default option.
     *
     * @param paramNum Index of parameter to return
     * @param d Default value to return
     */
    double asDouble(unsigned int paramNum, double d) {
        if (paramNum >= paramsList.size()) return 0;
        return asDouble(paramsList[paramNum], d);
    }
    
    /**
     * Returns a parameter as double.
     *
     * @param paramNum Index of parameter to return
     */
    double asDouble(unsigned int paramNum) {
        if (paramNum >= paramsList.size()) return 0;
        return asDouble(paramsList[paramNum]);
    }
    
    /**
     * Returns a parameter as String with default value.
     *
     * @param paramName Name of parameter to return
     * @param d Default value to return
     */
    string asString(const string& paramName, string d) {
        if (paramsHash.find(paramName) == paramsHash.end())
            return d;
        return paramsHash[paramName];
    }
    
    /**
     * Returns a parameter as string
     *
     * @param paramName Name of parameter to return
     */
    string asString(const string& paramName) {
        if (paramsHash.find(paramName) == paramsHash.end()) { return ""; }
        
        return paramsHash[paramName];
    }
    
    
    /**
     * Returns a parameter as string
     *
     * @param paramNum Number of parameter to return
     */
    string asString(unsigned int paramNum) {
        if (paramNum >= paramsList.size()) return "";
        return asString(paramsList[paramNum]);
    }
    
    /**
     * Returns the name of a parameter
     *
     * @param paramNum Number of parameter to return its name
     */
    string getParamName(unsigned int paramNum) {
        if (paramNum >= paramsList.size()) return "";
        return paramsList[paramNum];
    }
    
    /**
     * Returns the number of parameters in the list
     */
    int length() {
        return paramsList.size();
    }
    
    
    /**
     * Adds a list of parameters to the current list
     *
     * @param otherArray Another list of parameters to add
     */
    void merge(ParamsArray* otherArray) {
        for (int i=0; i<otherArray->length(); i++) {
            string paramName = otherArray->getParamName(i);
            if (paramName.substr(0, 3) == "arg") {
                paramName = "";
            }
            
            addParam(paramName+"="+otherArray->asString(i));
        }
    }
    
    void addConstraint(const string& param, const string& desc, bool optional) {
        paramsConstraints.push_back(ParamsConstraint(param, desc, optional));
    }
    
    bool runConstraints(const string& programName = "", const string& revision = "") {
        bool paramsOk = true;
        
        if ((programName != "") || (revision != "")) {
            cerr<<programName<<" "<<revision<<endl;
        }
        
        for (unsigned int i=0; i<paramsConstraints.size(); i++) {
            ParamsConstraint c = (ParamsConstraint)paramsConstraints[i];
            if ((c.optional == false) && (asString(c.param)=="")) {
                if (paramsOk) {
                    cerr<<description<<endl;
                    cerr<<"Parameters:"<<endl;
                    for (unsigned int j=0; j<paramsConstraints.size(); j++) {
                        ParamsConstraint c2 = (ParamsConstraint)paramsConstraints[j];
                        cerr<<c2.param<<" - "<<c2.desc;
                        if (c2.optional) {
                            cerr<<" (optional)"<<endl;
                        } else {
                            cerr<<endl;
                        }
                    }
                    cerr<<"parameters are passed using param1=value1 param2=value2 ..."<<endl;
                }
                cerr<<c.param<<" not set"<<endl;
                paramsOk = false;
            }
        }
        
        if (paramsOk) {
            cerr<<"Parameters list:"<<endl;
            for (unsigned int i=0; i<paramsConstraints.size(); i++) {
                ParamsConstraint c = (ParamsConstraint)paramsConstraints[i];
                if (asString(c.param) == "") {
                    cerr<<c.param<<" - [default]"<<endl;
                } else {
                    cerr<<c.param<<" - "<<asString(c.param)<<endl;
                }
            }
            cerr<<endl;
        }
        
        return paramsOk;
    }
    
    void setDescription(const string& d) {
        description = d;
    }
};

#endif	/* PARAMSARRAY_H_ */

